#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"

#define VIVY_PRAGMA(x) _Pragma(#x)
#define TODO(x)        VIVY_PRAGMA(message("\"TODO: " #x "\""))

#define VIVY_WIN_EXE_SUFFIX             ".exe"
#define VIVY_ASSERT_STRINGIFY_HELPER(x) #x
#define VIVY_ASSERT_STRINGIFY(x)        VIVY_ASSERT_STRINGIFY_HELPER(x)
#define VIVY_ASSERT_STRING(cond)                                        \
    ::Vivy::Utils::writeAssertLocation("\"" cond "\" in file " __FILE__ \
                                       ", line " VIVY_ASSERT_STRINGIFY(__LINE__))

#define VIVY_GUARD(cond) ((Q_LIKELY(cond)) ? true : (VIVY_ASSERT_STRING(#cond), false))
#define VIVY_UNUSED      __attribute__((unused))
#define VIVY_EXPORT      __attribute__((visibility("default")))
#define VIVY_NO_EXPORT   __attribute__((visibility("hidden")))
#define VIVY_DEPRECATED  __attribute__((__deprecated__))
#define VIVY_CONSTRUCTOR __attribute__((constructor))
#define VIVY_DESTRUCTOR  __attribute__((destructor))
#define VIVY_ASSERT(cond)              \
    {                                  \
        if (!cond)                     \
            VIVY_ASSERT_STRING(#cond); \
    }

// Use chrono instead of std::chrono...
namespace chrono = std::chrono;
using chrono::duration_cast;

// Prety define for OpenMP's parallel for loop with indentation not fucked up
// by clang-format.
#define parallel_for \
    _Pragma("omp parallel for") for

// Disable copy and move assign operators
#define VIVY_DISABLE_ASSIGN_OPERATORS(classname)                        \
    classname &operator=(const classname &) = delete; /* Copy assign */ \
    classname &operator=(classname &&)      = delete; /* Move assign */

#define VIVY_DISABLE_COPY_CTOR(classname) classname(const classname &) = delete; /* Copy */
#define VIVY_DISABLE_MOVE_CTOR(classname) classname(classname &&) = delete;      /* Move */

// Don't move this object, create it in one place and never move it again.
#define VIVY_UNMOVABLE_OBJECT(classname) \
    VIVY_DISABLE_COPY_CTOR(classname)    \
    VIVY_DISABLE_MOVE_CTOR(classname)    \
    VIVY_DISABLE_ASSIGN_OPERATORS(classname)

#define VIVY_DEFAULT_COPY_CTOR(classname) \
    explicit classname(const classname &) noexcept = default; /* Copy */

#define VIVY_DEFAULT_MOVE_CTOR(classname) \
    explicit classname(classname &&) noexcept = default; /* Move  */

namespace Vivy
{
// Concept for classes that can be viewd inside a property view
template <typename T>
concept PropertyConstViewable = requires(T element)
{
    // clang-format off
    { element.getElementName() } -> std::same_as<QString>;
    { element.getProperties() }  -> std::same_as<QJsonDocument>;
    // clang-format on
};

// Concept for mutexes
template <typename T>
concept MutexType = requires(T mtx)
{
    // clang-format off
    { mtx.lock() }     -> std::same_as<void>;
    { mtx.try_lock() } -> std::same_as<bool>;
    { mtx.unlock() }   -> std::same_as<void>;
    // clang-format on
};

// Concept with a better name for inheritence between types
template <class T, class U>
concept Derived = std::is_base_of<U, T>::value;

// Concept for string things
template <typename T>
concept StringType = requires(T str)
{
    // clang-format off
    typename T::iterator;   // Can iterate
    typename T::value_type; // The `Char`
    { str.size() }  -> std::convertible_to<std::size_t>;
    { str.begin() } -> std::same_as<typename T::iterator>;
    { str.end() }   -> std::same_as<typename T::iterator>;
    { str[0] }      -> std::same_as<typename T::const_reference>;
    // clang-format on
};

// Theme type
enum class ThemeType { QtCreator, System, QssFile };

// Prefered collor for a given type
enum class VivyThemeType { Dark, Light, _vv_COUNT };
}

namespace Vivy::Utils
{
// Add more as needed.
enum OsType { OsTypeWindows, OsTypeLinux, OsTypeMac, OsTypeOtherUnix, OsTypeOther };

static const QStringList audioFileSuffix  = { "wave", "wav", "ogg",  "mp3",  "m4a",
                                              "opus", "mp2", "aiff", "flac", "alac" };
static const QStringList videoFileSuffix  = { "mkv", "mp4", "mov", "avi", "av1", "m4v", "flv" };
static const QStringList assFileSuffix    = { "ass" };
static const QStringList vivyFileSuffix   = { "vivy" };
static const QStringList scriptFileSuffix = { "module" };

const QString &getAudioFileSuffixFilter() noexcept;
const QString &getVideoFileSuffixFilter() noexcept;
const QString &getAssFileSuffixFilter() noexcept;
const QString &getScriptFileSuffixFilter() noexcept;
const QString &getVivyDocumentFileSuffixFilter() noexcept;
const QString &getAnyDocumentFileSuffixFilter() noexcept;
const QString &getAnyTopLevelDocumentFileSuffixFilter() noexcept;

static constexpr std::size_t pointerAlignement = alignof(void *);

std::string &ltrim(std::string &s, const char *t = " \t\n\r\f\v");
std::string &rtrim(std::string &s, const char *t = " \t\n\r\f\v");
std::string &trim(std::string &s, const char *t = " \t\n\r\f\v");

template <typename T> inline void
uniqAndSort(std::vector<T> &vec) noexcept
{
    std::sort(std::begin(vec), std::end(vec));
    vec.erase(std::unique(std::begin(vec), std::end(vec)), std::end(vec));
}

// WARN: The V1 and V2 vectors need to be sorted before calling this function.
template <typename T> static inline std::vector<T>
sortedSetDifference(const std::vector<T> &v1, const std::vector<T> &v2)
{
    std::vector<T> res{};

    // Do V1 - V2 -> RES
    std::set_difference(std::begin(v1), std::end(v1), // V1
                        std::begin(v2), std::end(v2), // V2
                        std::inserter(res, std::begin(res)));

    return res;
}

static inline std::string
anyStringToStdString(StringType auto const &str) noexcept
{
    std::string ret;
    std::size_t i;
    const std::size_t size = str.size();
    ret.resize(size);
    for (i = 0; i < size && str[i]; ++i)
        ret[i] = str[i];
    ret.resize(i);
    return ret;
}

enum class DocumentType : quint64 {
    /* Basic types */
    Vivy = (1 << 1),
    ASS  = (1 << 2),

    /* Audio */
    WAVE = (1 << 3),
    WAV  = WAVE,
    OGG  = (1 << 4),
    MP3  = (1 << 5),
    M4A  = (1 << 6),
    OPUS = (1 << 7),
    MP2  = (1 << 8),
    AIFF = (1 << 9),
    FLAC = (1 << 10),
    ALAC = (1 << 11),

    /* Video */
    MKV = (1 << 12),
    MP4 = (1 << 13),
    MOV = (1 << 14),
    AVI = (1 << 14),
    AV1 = (1 << 15),
    M4V = (1 << 16),
    FLV = (1 << 17),

    /* Vivy script */
    VivyScript = (1 << 18),

    /* Meta-types */
    Audio = (WAVE | OGG | MP3 | M4A | OPUS | MP2 | AIFF | FLAC | ALAC),
    Video = (MKV | MP4 | MOV | AVI | AV1 | M4V | FLV)
};

template <typename E> constexpr auto
toUnderlying(E e) noexcept
{
    return static_cast<std::underlying_type_t<E>>(e);
}

// Structure used to represent the time as it is presented in an ASS file.
struct Time final {
    quint64 hour;
    quint64 minute;
    quint64 second;
    quint64 centisecond;

    static Time fromString(const QString &);
    static Time fromUInt(const quint64 &);

    QString toString() const noexcept;
    quint64 toUInt() const noexcept;
};

bool detectDocumentType(const QFileInfo &, DocumentType *const);
bool decodeLineToBoolean(const QString &item, const QString &error);
int decodeLineToInteger(const QString &item, const QString &error);
float decodeLineToFloating(const QString &item, const QString &error);

QString getBaseName(const QString &) noexcept;
[[noreturn]] void writeAssertLocation(const char *msg);

struct OsSpecificAspects final {
private:
    OsSpecificAspects() {}

public:
    static QString withExecutableSuffix(OsType osType, const QString &executable);
    static Qt::CaseSensitivity fileNameCaseSensitivity(OsType osType);
    static Qt::CaseSensitivity envVarCaseSensitivity(OsType osType);
    static QChar pathListSeparator(OsType osType);
    static Qt::KeyboardModifier controlModifier(OsType osType);
    static QString pathWithNativeSeparators(OsType osType, const QString &pathName);
};
}

namespace Vivy
{
// Audio document types
enum class AudioDocumentType : quint64 {
    MP3  = Utils::toUnderlying(Utils::DocumentType::MP3),
    OGG  = Utils::toUnderlying(Utils::DocumentType::OGG),
    M4A  = Utils::toUnderlying(Utils::DocumentType::M4A),
    OPUS = Utils::toUnderlying(Utils::DocumentType::OPUS),
    MP2  = Utils::toUnderlying(Utils::DocumentType::MP2),
    AIFF = Utils::toUnderlying(Utils::DocumentType::AIFF),
    FLAC = Utils::toUnderlying(Utils::DocumentType::FLAC),
    ALAC = Utils::toUnderlying(Utils::DocumentType::ALAC)
};

// Video document types
enum class VideoDocumentType : quint64 {
    MKV = Utils::toUnderlying(Utils::DocumentType::MKV),
    MP4 = Utils::toUnderlying(Utils::DocumentType::MP4),
    MOV = Utils::toUnderlying(Utils::DocumentType::MOV),
    AVI = Utils::toUnderlying(Utils::DocumentType::AVI),
    AV1 = Utils::toUnderlying(Utils::DocumentType::AV1),
    M4V = Utils::toUnderlying(Utils::DocumentType::M4V),
    FLV = Utils::toUnderlying(Utils::DocumentType::FLV)
};

// Ass document types
enum class AssDocumentType : quint64 { ASS = Utils::toUnderlying(Utils::DocumentType::ASS) };
}
