use vvs_utils::*;

/// The color representation
#[derive(Debug, Clone, PartialEq)]
pub enum ASSColor {
    /// The Red Green Blue Alpha representation.
    RGBA { r: u8, g: u8, b: u8, a: u8 },

    /// The Hue Saturation Lightness with Alpha representation. The Hue must be between is in
    /// radian, the lightness and saturation values must be between 0 and 1.
    HSLA { h: f32, s: f32, l: f32, a: u8 },
}

macro_rules! rgb {
    ($NAME: ident => $r: literal $g: literal $b: literal) => {
        pub const $NAME: ASSColor = ASSColor::RGBA { r: $r, g: $g, b: $b, a: 0 };
    };
}

impl ASSColor {
    rgb! { WHITE   => 255 255 255 }
    rgb! { SILVER  => 192 192 192 }
    rgb! { GRAY    => 128 128 128 }
    rgb! { BLACK   =>   0   0   0 }

    rgb! { RED     => 255   0   0 }
    rgb! { MAROON  => 128   0   0 }
    rgb! { YELLOW  => 255 255   0 }
    rgb! { OLIVE   => 128 128   0 }
    rgb! { LIME    =>   0 255   0 }
    rgb! { GREEN   =>   0 128   0 }
    rgb! { AQUA    =>   0 255 255 }
    rgb! { TEAL    =>   0 128 128 }
    rgb! { BLUE    =>   0   0 255 }
    rgb! { NAVY    =>   0   0 128 }
    rgb! { FUCHSIA => 255   0 255 }
    rgb! { PURPLE  => 128   0 128 }

    #[inline]
    fn skip_delimiters(str: &str) -> &str {
        str.trim()
            .trim_start_matches('#')
            .trim_start_matches("&H")
            .trim_start_matches('&')
            .trim_end_matches('&')
    }

    #[inline]
    pub fn try_from_rgba(str: impl AsRef<str>) -> Result<Self, String> {
        let color = Self::skip_delimiters(str.as_ref());
        if !(color.len() == 6 || color.len() == 8) {
            return Err(format!("invalid color string description: {}", str.as_ref()));
        }
        Ok(Self::RGBA {
            r: u8::from_str_radix(&color[0..2], 16).map_err(|err| format!("invalid red description: {err}"))?,
            g: u8::from_str_radix(&color[2..4], 16).map_err(|err| format!("invalid green description: {err}"))?,
            b: u8::from_str_radix(&color[4..6], 16).map_err(|err| format!("invalid blue description: {err}"))?,
            a: (color.len() == 8)
                .then(|| {
                    u8::from_str_radix(&color[6..8], 16).map_err(|err| format!("invalid alpha description: {err}"))
                })
                .unwrap_or(Ok(0))?,
        })
    }

    #[inline]
    pub fn try_from_bgra(str: impl AsRef<str>) -> Result<Self, String> {
        let color = Self::skip_delimiters(str.as_ref());
        if !(color.len() == 6 || color.len() == 8) {
            return Err(format!("invalid color string description: {}", str.as_ref()));
        }
        Ok(Self::RGBA {
            r: u8::from_str_radix(&color[4..6], 16).map_err(|err| format!("invalid red description: {err}"))?,
            g: u8::from_str_radix(&color[2..4], 16).map_err(|err| format!("invalid green description: {err}"))?,
            b: u8::from_str_radix(&color[0..2], 16).map_err(|err| format!("invalid blue description: {err}"))?,
            a: (color.len() == 8)
                .then(|| {
                    u8::from_str_radix(&color[6..8], 16).map_err(|err| format!("invalid alpha description: {err}"))
                })
                .unwrap_or(Ok(0))?,
        })
    }

    #[inline]
    pub fn into_rgba(self) -> Self {
        match self {
            this @ ASSColor::RGBA { .. } => this,
            ASSColor::HSLA { h, s, l, a } => {
                let h = f32_degres_clamp(h);
                let c = (1.0 - (2.0 * l - 1.0).abs()) * s;
                let x = c * (1.0 - ((h / 60.0) % 2.0 - 1.0).abs());
                let m = l - c / 2.0;

                let (r, g, b) = if (0.0..60.0).contains(&h) {
                    (c, x, 0.0)
                } else if (60.0..120.0).contains(&h) {
                    (x, c, 0.0)
                } else if (120.0..180.0).contains(&h) {
                    (0.0, c, x)
                } else if (180.0..240.0).contains(&h) {
                    (0.0, x, c)
                } else if (240.0..300.0).contains(&h) {
                    (x, 0.0, c)
                } else if (300.0..360.0).contains(&h) {
                    (c, 0.0, x)
                } else {
                    unreachable!()
                };

                ASSColor::RGBA {
                    r: ((r + m) * 255.0).trunc() as u8,
                    g: ((g + m) * 255.0).trunc() as u8,
                    b: ((b + m) * 255.0).trunc() as u8,
                    a,
                }
            }
        }
    }

    #[inline]
    pub fn into_hsla(self) -> Self {
        match self {
            this @ ASSColor::HSLA { .. } => this,
            ASSColor::RGBA { r, g, b, a } => {
                const U8_MAX: f32 = u8::MAX as f32;
                let h = self.hue();
                let (r, g, b) = (r as f32 / U8_MAX, g as f32 / U8_MAX, b as f32 / U8_MAX);
                let (min, max) = minmax_partial!(r, g, b);
                let l = (min + max) / 2.0;
                let delta = max - min;
                let s = either!(f32_epsilon_eq(delta, 0.0) => 0.0;
                    delta / (1.0 - (2.0 * l - 1.0).abs())
                );
                ASSColor::HSLA { h, s, l, a }
            }
        }
    }

    /// Returns the HUE from the color (the H in HSL/HSV).
    #[inline]
    fn hue(&self) -> f32 {
        match self {
            ASSColor::HSLA { h, .. } => *h,
            ASSColor::RGBA { r, g, b, .. } => {
                let (r, g, b) = (*r, *g, *b);
                let (min, max) = minmax_partial!(r, g, b);
                let c = max - min;
                ((if c == 0 {
                    0.0
                } else if max == r {
                    let segment = (g as f32 - b as f32) / c as f32;
                    either!(segment < 0.0 => segment + 6.0; segment)
                } else if max == g {
                    let segment = (b as f32 - r as f32) / c as f32;
                    segment + 2.0
                } else if max == b {
                    let segment = (r as f32 - g as f32) / c as f32;
                    segment + 4.0
                } else {
                    panic!()
                } * 60.0)
                    % 360.0)
                    .trunc()
            }
        }
    }
}
