/// The position of an object from the top left corner of the screen. The real position depends on
/// the align of the object.
#[derive(Debug, Default, Clone, Copy, PartialEq, Eq)]
pub enum ASSPosition {
    /// An unspecified position.
    #[default]
    Unspecified,

    /// A static position.
    Pos { x: i64, y: i64 },

    /// A linear movement.
    LinearMove { x1: i64, y1: i64, x2: i64, y2: i64 },

    /// A linear movement with time step specified.
    TimedMove { x1: i64, y1: i64, x2: i64, y2: i64, from_ms: i64, to_ms: i64 },
}

/// The alignement of the object.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASSAlign {
    /// Bottom left
    BL = 1,

    /// Bottom center
    BC = 2,

    /// Bottom right
    BR = 3,

    /// Center left
    CL = 4,

    /// Center center
    CC = 5,

    /// Center right
    CR = 6,

    /// Top left
    TL = 7,

    /// Top center
    TC = 8,

    /// Top right
    TR = 9,
}

/// Pointer used to store a position, to help with mutability with LUA wrappers.
pub type ASSPositionPtr = std::sync::Arc<ASSPosition>;

impl std::str::FromStr for ASSAlign {
    type Err = String;

    #[inline]
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.trim() {
            "1" => Ok(ASSAlign::BL),
            "2" => Ok(ASSAlign::BC),
            "3" => Ok(ASSAlign::BR),
            "4" => Ok(ASSAlign::CL),
            "5" => Ok(ASSAlign::CC),
            "6" => Ok(ASSAlign::CR),
            "7" => Ok(ASSAlign::TL),
            "8" => Ok(ASSAlign::TC),
            "9" => Ok(ASSAlign::TR),
            s => Err(format!("invalid value for ASSAlign, must be in `1..=9`, got: {s}")),
        }
    }
}

impl From<vvs_font::Point> for ASSPosition {
    #[inline]
    fn from(vvs_font::Point { x, y }: vvs_font::Point) -> Self {
        ASSPosition::Pos { x, y }
    }
}

impl ASSPosition {
    #[inline]
    pub fn into_ptr(self) -> ASSPositionPtr {
        std::sync::Arc::new(self)
    }
}
