use crate::*;

mod color {
    use crate::*;

    // https://en.wikipedia.org/wiki/HSL_and_HSV#Hue_and_chroma
    const RGB2HSL_TABLE: [(&str, f32, f32, f32); 19] = [
        ("FFFFFF", 0.000, 0.000, 1.000),
        ("808080", 0.000, 0.000, 0.500),
        ("000000", 0.000, 0.000, 0.000),
        ("FF0000", 0.000, 1.000, 0.500),
        ("BFBF00", 60.00, 1.000, 0.375),
        ("008000", 120.0, 1.000, 0.250),
        ("80FFFF", 180.0, 1.000, 0.750),
        ("8080FF", 240.0, 1.000, 0.750),
        ("BF40BF", 300.0, 0.500, 0.500),
        ("A0A424", 61.80, 0.638, 0.393),
        ("411BEA", 251.1, 0.832, 0.511),
        ("1EAC41", 134.9, 0.707, 0.396),
        ("F0C80E", 49.50, 0.893, 0.497),
        ("B430E5", 283.7, 0.775, 0.542),
        ("ED7651", 14.30, 0.817, 0.624),
        ("FEF888", 56.90, 0.991, 0.765),
        ("19CB97", 162.4, 0.779, 0.447),
        ("362698", 248.3, 0.601, 0.373),
        ("7E7EB8", 240.5, 0.290, 0.607),
    ];

    #[test]
    fn test_from_string() {
        assert!(ASSColor::try_from_rgba("#&AABBCC&").is_ok());
        assert!(ASSColor::try_from_rgba("#AABBCC").is_ok());
        assert!(ASSColor::try_from_rgba("&AABBCC&").is_ok());
        assert!(ASSColor::try_from_rgba("AABBCC").is_ok());
        assert!(ASSColor::try_from_rgba("AABBCCAA").is_ok());

        assert_eq!(ASSColor::try_from_rgba("AABBCC").unwrap(), ASSColor::try_from_bgra("CCBBAA").unwrap());
    }

    #[test]
    fn test_rgb2hsl() {
        for (rgb, h_target, s1_target, l1_target) in RGB2HSL_TABLE {
            const EPSILON_DEG: f32 = 1.0;
            const EPSILON_0_1: f32 = 10E-3;
            let ASSColor::HSLA { h, s, l, .. } = ASSColor::try_from_rgba(rgb).unwrap().into_hsla() else {
                unreachable!()
            };
            assert!((h_target - h).abs() <= EPSILON_DEG, "invalid convertion for color #{rgb}, hue {h_target} != {h}");
            assert!((s - s1_target).abs() <= EPSILON_0_1, "invalid convertion for color #{rgb}, s {s1_target} != {s}");
            assert!((l - l1_target).abs() <= EPSILON_0_1, "invalid convertion for color #{rgb}, l {l1_target} != {l}");
        }
    }

    #[test]
    fn test_hsl2rgb() {
        for (rgb, h, s, l) in RGB2HSL_TABLE {
            macro_rules! eq {
                ($a: expr, $b: expr, $($msg: expr),+) => {
                    assert!(($a as f32 - $b as f32).abs() <= 1.0, $($msg),+);
                };
            }
            let ASSColor::RGBA { r, g, b, .. } = ASSColor::HSLA { h, s, l, a: 0 }.into_rgba() else {
                unreachable!()
            };
            let ASSColor::RGBA { r: r_target, g: g_target, b: b_target, .. } = ASSColor::try_from_rgba(rgb).unwrap()
            else {
                unreachable!()
            };
            eq! { r_target, r, "invalid convertion on red for #{rgb}: {r_target} != {r}"}
            eq! { g_target, g, "invalid convertion on green for #{rgb}: {g_target} != {g}"}
            eq! { b_target, b, "invalid convertion on blue for #{rgb}: {b_target} != {b}"}
        }
    }
}

#[test]
fn test_parse_empty_ass() {
    #[derive(Default, Clone, Copy, PartialEq, Eq)]
    struct AT();
    impl AuxTable for AT {
        fn deep_clone(&self) -> Self {
            *self
        }
    }

    if let Err(err) =
        reader::ass_container_from_str::<AT>(reader::ContainerFileType::ASS, include_str!("../utils/empty.ass"))
    {
        panic!("{err}")
    }
}
