//! The VivyScript cli

use crate::args::Args;
use anyhow::{anyhow, Context, Result};
use clap::CommandFactory;
use std::{fs, io};
use vvs_ass::{ass_container_from_file, ASSContainer};
use vvs_cli::{
    args,
    config::{Config, ConfigFile},
    logger,
};
use vvs_parser::prelude::*;
use vvs_runtime::{types::VVRTTable, JIT};
use vvs_utils::xdg::*;

fn print_face_info(name: &str, font: &[u8]) -> Result<()> {
    let font = vvs_font::Font::try_from(font).with_context(|| format!("failed to parse font {name}"))?;

    let font_types = font
        .font_types()
        .into_iter()
        .map(|ty| format!("{ty:?}"))
        .collect::<Vec<_>>()
        .join(", ");

    println!("###");
    println!("# Family Name:        {}", font.name().context("failed to get the font name")?);
    println!("# Name(s):            {}", font.family_names().join(", "));
    println!("# Font Type(s):       {font_types}");
    println!("# Number of glyph(s): {}", font.number_of_glyphs());

    Ok(())
}

fn main() -> Result<()> {
    logger::init(None).context("failed to init logger")?;
    let mut config = Config::default();
    config.extend([
        XDGConfig::<ConfigFile, XDGConfigMergedSilent>::new("vvcc", ConfigFile::deserialize)
            .file("vvcc.toml")
            .read_or_default(ConfigFile::serialize),
    ]);
    config.extend([<Args as clap::Parser>::parse()]);
    let Config {
        script,
        ass_file: ass_file_path,
        options: options_path,
        info,
        manpage,
        shell,
        font_info,
        verbose,
        include_folders: _,
        kara_maker,
    } = config;
    logger::level(verbose);

    if manpage {
        log::debug!(target: "vvcc", "generate the manpage for vvcc");
        return clap_mangen::Man::new(Args::command())
            .render(&mut io::stdout())
            .context("failed to render manpage for vvcc");
    }

    if let Some(shell) = shell {
        log::debug!(target: "vvcc", "generate shell completion script for shell {shell}");
        let mut command = Args::command();
        let command_name = command.get_name().to_string();
        clap_complete::generate(shell, &mut command, command_name, &mut io::stdout());
        return Ok(());
    }

    if let Some(font_info) = font_info {
        return match font_info {
            Some(path) => print_face_info(
                &path.to_string_lossy(),
                &fs::read(&path).with_context(|| format!("failed to read font: {}", path.to_string_lossy()))?,
            ),
            None => vvs_font::embeded_fonts()
                .iter()
                .try_for_each(|(n, f)| print_face_info(n, f)),
        };
    }

    match script {
        None => Args::command()
            .print_help()
            .context("failed to print help message for vvcc"),

        Some(script) => {
            let script = SourceCode::from_path(script)?;
            log::info!("run as `{kara_maker}` script `{}`", script.name());
            if let Some(ref options) = options_path {
                log::info!("use option file `{}`", options.display())
            }

            let output = FrontendPipeline::new(&script)
                .passes(&[])
                .options(options_path.as_ref())
                .context("failed to parse option file")?
                .run()
                .context("failed to parse script files")?;

            if info {
                match ass_file_path {
                    Some(ref path) => println!("ass file ........... {}", path.display()),
                    None => println!("ass file ........... ø"),
                }
                match options_path {
                    Some(path) => println!("options file ....... {}", path.display()),
                    None => println!("options file ....... ø"),
                }
                println!("{output}");
            }

            // let FrontendOutput { ast, imports, main, options, library } = output;

            let ass: ASSContainer<VVRTTable> = ass_container_from_file(
                ass_file_path.ok_or_else(|| anyhow!("no subtitle file to run the script `{}` on", script.name()))?,
            )
            .context("failed to parse the subtitle file")?;

            // let jit = JIT::new()?;
            // let module = unsafe { LowererBuilder::new(script.name(), jit.ctx()) }.build().lower();
            // unsafe { jit.add_module(module)? };

            log::error!("use the ASS container and insert it into the JIT: {ass:#?}");

            unimplemented!()
        }
    }
}
