use clap::{
    builder::TypedValueParser,
    error::{ContextKind, ContextValue, ErrorKind},
};
use std::path::PathBuf;

#[derive(Debug, Clone, Copy)]
pub(crate) struct FileTypeValueParser {
    extension: &'static str,
}

impl FileTypeValueParser {
    pub fn new(extension: &'static str) -> Self {
        Self { extension }
    }
}

impl TypedValueParser for FileTypeValueParser {
    type Value = PathBuf;

    fn parse_ref(
        &self,
        cmd: &clap::Command,
        arg: Option<&clap::Arg>,
        value: &std::ffi::OsStr,
    ) -> Result<Self::Value, clap::Error> {
        let mut err = clap::Error::new(ErrorKind::ValueValidation).with_cmd(cmd);
        if let Some(arg) = arg {
            err.insert(ContextKind::InvalidArg, ContextValue::String(arg.to_string()));
        }
        match value.to_ascii_lowercase().to_str() {
            Some(value) => match value.trim().split('.').last() {
                Some(file_ext) => (file_ext == self.extension)
                    .then_some(PathBuf::from(value))
                    .ok_or_else(|| {
                        err.insert(
                            ContextKind::InvalidValue,
                            ContextValue::String(format!("invalid extension {file_ext}, expected {}", self.extension)),
                        );
                        err
                    }),
                None => {
                    err.insert(
                        ContextKind::InvalidValue,
                        ContextValue::String(format!("file has no extension, expected extension {}", self.extension)),
                    );
                    Err(err)
                }
            },
            None => {
                err.insert(ContextKind::InvalidValue, ContextValue::String("invalid utf8 string".to_string()));
                Err(err)
            }
        }
    }
}
