use crate::Value;
use hashbrown::HashMap;
use vvs_llvm::sys::*;
use vvs_runtime_types::vvll::LLVMTypeIsManaged;

/// Just a hashmap used to store values in scope.
#[derive(Default, Clone)]
pub(crate) struct ValueContext {
    /// The values in context.
    values: HashMap<Box<str>, Value>,

    /// The values that are to be droped, in the order of their définition! They must be dropped in
    /// the reverse order in which they are stored!
    to_drop: Vec<Value>,

    /// The number of anon values present in this context.
    anon_count: u64,
}

impl ValueContext {
    /// Create a new sub-context. We clone the available values but not the destructors to run as
    /// we don't want to run them too early.
    pub fn sub_context(&self) -> Self {
        let Self { values, anon_count: i, .. } = self;
        Self { values: values.clone(), anon_count: *i, to_drop: vec![] }
    }

    /// Register a new value in the context. We don't give a shit about its name, see if we need a
    /// destructor for this value.
    pub fn register_anon(&mut self, value: Value) {
        if let Some(name) = value.as_str() {
            let name = format!(".anon.{}:{}", self.anon_count, name);
            self.anon_count += 1;
            self.register(name, value)
        }
    }

    /// Clone the context and register a new value into it. Used to reduce a bit calls to clone
    /// everywhere...
    pub fn to_register(&self, name: impl AsRef<str>, value: Value) -> Self {
        let mut this = self.clone();
        this.register(name, value);
        this
    }

    /// Clone the context and register a new value into it. Used to reduce a bit calls to clone
    /// everywhere...
    pub fn to_register_anon(&self, value: Value) -> Self {
        let mut this = self.clone();
        this.register_anon(value);
        this
    }

    /// Register a new value in the context.
    pub fn register(&mut self, name: impl AsRef<str>, value: Value) {
        fn inner(this: &mut ValueContext, name: &str, value: Value) {
            log::error!("take into account destructors for {name}");
            if Value::Nil == value {
                panic!("can't register nil value `{name}`")
            } else if this.values.insert(name.into(), value).is_some() {
                panic!("multiple definitions of `{name}`")
            }
        }
        inner(self, name.as_ref(), value)
    }

    /// Register a multiple values in the context.
    pub fn register_all(&mut self, iter: impl IntoIterator<Item = (impl AsRef<str>, Value)>) {
        iter.into_iter().for_each(|(name, value)| self.register(name, value));
    }

    /// Get the values to drop in this context in the order of their definition. Also remove them
    /// from this context.
    pub fn get_values_to_drop(&mut self) -> Vec<Value> {
        std::mem::take(&mut self.to_drop)
    }

    /// Forget a value to drop by its llvm value ref. This way we can also forget anon values.
    pub fn forget_by_llvm_value(&mut self, value: LLVMValueRef) {
        use Value::*;
        if let Some(idx) = self.to_drop.iter().enumerate().find_map(|(idx, val)| {
            matches!(*val, Const(_, val) | Alloca(_, _, val) | Function(_, _, val) if val == value).then_some(idx)
        }) {
            self.to_drop.remove(idx);
        }
    }

    /// Add values to drop from another context to extend their lifetime. We can't fail that thing
    /// because we ignore values that are re-registered.
    pub fn add_values_to_drop(&mut self, mut values: Vec<Value>) {
        values.retain(|v| {
            v.llvm_value()
                .is_some_and(|v| self.to_drop.iter().any(|u| u.llvm_value().is_some_and(|u| u == v)))
                && v.llvm_type().is_some_and(|t| unsafe { LLVMTypeIsManaged(t) })
        });
        self.to_drop.append(&mut values);
    }

    /// Get the value by its name in the context, or panic!
    pub fn get_value(&self, name: impl AsRef<str>) -> &Value {
        self.values
            .get(name.as_ref())
            .unwrap_or_else(|| panic!("failed to find value `{}`", name.as_ref()))
    }
}
