use crate::lowerer::*;
use vvs_lang::ast::*;
use vvs_llvm::*;

/// Lower a single instruction.
pub(super) struct CallLowerer<'a> {
    llvm: LLVMLowerContext,
    ctx: &'a ValueContext,
}

impl<'a> CallLowerer<'a> {
    pub(super) fn new(llvm: LLVMLowerContext, ctx: &'a ValueContext) -> Self {
        Self { llvm, ctx }
    }

    pub(super) unsafe fn call_intrinsic<const N: usize>(
        name: &'static str,
        (m, _, b): LLVMLowerContext,
        ret: LLVMTypeRef,
        mut args: [LLVMValueRef; N],
    ) -> LLVMValueRef {
        let id = LLVMLookupIntrinsicID(name.as_ptr() as *const _, name.len());
        let mut args_ty: Vec<_> = args.iter().map(|val| LLVMTypeOf(*val)).collect();
        let (args_ty, args_len, args) = (args_ty.as_mut_ptr(), args.len().try_into().unwrap(), args.as_mut_ptr());
        let (functy, func) = (
            LLVMFunctionType(ret, args_ty, args_len, 0),
            LLVMGetIntrinsicDeclaration(m, id, args_ty, args_len as usize),
        );
        LLVMBuildCall2(b, functy, func, args, args_len, cstr!())
    }

    pub(super) fn lower_method_invok(
        self,
        obj: &ASTExpr,
        func: impl AsRef<str>,
        args: &[ASTExpr],
    ) -> (ValueContext, Value) {
        let Self { llvm: llvm @ (_, c, b), ctx } = self;

        let (ctx, object) = ExpressionLowerer::with_assumed_type(llvm, obj, ctx).lower();
        let (oty, obj) = (object.get_ast_type(), object.loaded(c, b));

        let (aargs, rty, lty, lfn) = match ctx.get_value(VVRTSymbol::method(oty, func.as_ref()).unwrap().to_string()) {
            Value::Function(ASTType::Function(aargs, rty), lty, lfn) if args.len() == aargs.len() => {
                (aargs.clone(), rty.clone(), *lty, *lfn)
            }
            val => unreachable!("unexpected not a function value: {val:?}"),
        };

        let mut largs = Vec::with_capacity(args.len());
        let largc = 1 + args.len() as u32;
        largs.push(obj);
        let ctx = args.iter().zip(aargs).fold(ctx, |ctx, (arg, ty)| {
            let (ctx, arg) = ExpressionLowerer::with_expected_type(llvm, arg, &ty, &ctx).lower();
            largs.push(arg.loaded(c, b));
            ctx
        });

        match *rty {
            ASTType::Nil => unsafe {
                LLVMBuildCall2(b, lty, lfn, largs.as_mut_ptr(), largc, c"".as_ptr());
                (ctx, Value::Nil)
            },
            ty => unsafe {
                let ret_ty = TypeLowerer::new(c, &ty).lower();
                let ptr = LLVMBuildAlloca(b, ret_ty, cstr!());
                let val = LLVMBuildCall2(b, lty, lfn, largs.as_mut_ptr(), largc, cstr!());
                LLVMBuildStore(b, val, ptr);
                (ctx, Value::Alloca(ty, ret_ty, ptr))
            },
        }
    }

    pub(super) fn lower_func_call(self, func: &ASTExpr, args: &[ASTExpr]) -> (ValueContext, Value) {
        let Self { llvm: llvm @ (_, c, b), ctx } = self;

        let (ctx, aty, rty_ast, lty, lfn) =
            match ExpressionLowerer::with_expected_type(llvm, func, &ASTType::Nil, ctx).lower() {
                (ctx, Value::Function(ASTType::Function(aty, rty), lty, lfn)) if args.len() == aty.len() => {
                    (ctx, aty, rty, lty, lfn)
                }
                (_, what) => unreachable!("unexpected not-a-function-thingy: {what:?}"),
            };

        let mut largs = Vec::with_capacity(args.len());
        let args_len = args.len().try_into().expect("too many arguments");
        let mut ctx = args.iter().zip(aty).fold(ctx, |ctx, (arg, ty)| {
            let (ctx, expr) = ExpressionLowerer::with_expected_type(llvm, arg, &ty, &ctx).lower();
            largs.push(expr.loaded(c, b));
            ctx
        });

        match unsafe { LLVMGetReturnType(lty) } {
            rty if rty == unsafe { LLVMVoidTypeInContext(c) } => unsafe {
                LLVMBuildCall2(b, lty, lfn, largs.as_mut_ptr(), args_len, c"".as_ptr());
                (ctx, Value::Nil)
            },
            rty => unsafe {
                let val = LLVMBuildCall2(b, lty, lfn, largs.as_mut_ptr(), args_len, cstr!());
                let ptr = LLVMBuildAlloca(b, rty, c".ignored".as_ptr());
                LLVMBuildStore(b, val, ptr);
                let value = Value::Alloca(*rty_ast, rty, ptr);
                ctx.register_anon(value.clone());
                (ctx, value)
            },
        }
    }
}
