use crate::lowerer::*;
use vvs_lang::ast::*;
use vvs_llvm::*;
use vvs_runtime_types::{types::*, vvll::LLVMExported};

pub(super) struct CastLowerer<'a> {
    llvm: LLVMLowerContext,

    source: &'a Value,
    from_type: &'a ASTType,
    as_type: &'a ASTType,

    ctx: &'a ValueContext,
}

/// Represents the value that was casted by the [CastLowerer] lowerer.
#[derive(Debug)]
pub(super) enum CastedValue {
    /// A value can be [CastedValue::Unchanged] which means that we did nothing to it. you don't
    /// need to register it again to trop it as it's nothing new...
    Unchanged { value: Value },

    /// A value can be [CastedValue::Dynamic] which means that we generated code to try to convert
    /// the thing at runtime. The first element is the flag which is set to true if we did succeed,
    /// false otherwise. This is an `i8` and must be checked at runtime. Because we created a
    /// value, we muste register it to be dropped latter if needed.
    Dynamic { flag: LLVMValueRef, value: Value },

    /// A value can be [CastedValue::Static], which means that we did generate the code to cast the
    /// thing and that we will always succeed. We still need to register this value to be dropped
    /// if necessary...
    Static { value: Value },
}

impl<'a> CastLowerer<'a> {
    /// Create a new [CastLowerer] from an AST value.
    pub fn from_ast(llvm: LLVMLowerContext, source: &'a ASTVar, as_type: &'a ASTType, ctx: &'a ValueContext) -> Self {
        Self::from_llvm(llvm, ctx.get_value(source.name()), as_type, ctx)
    }

    /// Create a new [CastLowerer] from a LLVM value.
    pub fn from_llvm(llvm: LLVMLowerContext, source: &'a Value, as_type: &'a ASTType, ctx: &'a ValueContext) -> Self {
        Self { llvm, source, as_type, from_type: source.get_ast_type(), ctx }
    }

    /// Tells if the types are the same and we don't need to do the casting.
    fn is_same_type(from: &ASTType, to: &ASTType) -> bool {
        (to == from)
            || (to.is_table() && from.is_table())
            || (to.is_sequence() && from.is_sequence())
            || (to.is_variant() && from.is_variant())
    }

    /// Lower in a static context, meaning that we assume that the casting process must succeed and
    /// must not require a runtime check.
    pub fn lower_static(self) -> CastedValue {
        use ASTType::*;
        let Self { llvm: llvm @ (_, c, b), source, from_type, as_type, ctx } = self;

        if matches!(as_type, Nil) {
            return CastedValue::Unchanged { value: ConstantLowerer::new(c, &ASTConst::Nil).lower() };
        } else if Self::is_same_type(from_type, as_type) {
            return CastedValue::Unchanged { value: source.clone() };
        }

        let loaded = source.loaded(c, b);
        let store_into_ptr = |ty: ASTType, value: LLVMValueRef| unsafe {
            let ptr = LLVMBuildAlloca(b, LLVMTypeOf(value), cstr!());
            LLVMBuildStore(b, value, ptr);
            CastedValue::Static { value: Value::Alloca(ty, LLVMTypeOf(value), ptr) }
        };

        macro_rules! build_into {
            ($ty: ident: $conv: ident) => {
                unsafe {
                    let llvm_type = TypeLowerer::new(c, &$ty).lower();
                    let ptr = LLVMBuildAlloca(b, llvm_type, cstr!());
                    LLVMBuildStore(b, $conv(b, loaded, llvm_type, cstr!()), ptr);
                    CastedValue::Static { value: Value::Alloca($ty, llvm_type, ptr) }
                }
            };
        }

        match from_type {
            Nil => match as_type {
                Integer => CastedValue::Static {
                    value: Value::Const(Integer, unsafe {
                        LLVMConstInt(TypeLowerer::new(c, &Integer).lower(), Default::default(), 0)
                    }),
                },
                Floating => CastedValue::Static {
                    value: Value::Const(Floating, unsafe {
                        LLVMConstReal(TypeLowerer::new(c, &Floating).lower(), Default::default())
                    }),
                },
                Boolean => CastedValue::Static {
                    value: Value::Const(Boolean, unsafe {
                        LLVMConstInt(TypeLowerer::new(c, &Boolean).lower(), Default::default(), 0)
                    }),
                },
                ty if ty.is_table() => store_into_ptr(AnyTable, ctx.get_value("VVRTTable_new").call(b, [])),
                ty if ty.is_sequence() => store_into_ptr(AnySequence, ctx.get_value("VVRTSeq_new").call(b, [])),
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_nil").call(b, [])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            Integer => match as_type {
                Integer => CastedValue::Unchanged { value: source.clone() },
                Boolean => build_into! { Boolean: LLVMBuildIntCast },
                Floating => build_into! { Floating: LLVMBuildSIToFP },
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_from_integer").call(b, [loaded])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            Floating => match as_type {
                Floating => CastedValue::Unchanged { value: source.clone() },
                Integer => build_into! { Integer: LLVMBuildFPToSI },
                Boolean => build_into! { Boolean: LLVMBuildFPToSI },
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_from_floating").call(b, [loaded])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            Boolean => match as_type {
                Boolean => CastedValue::Unchanged { value: source.clone() },
                Integer => build_into! { Integer: LLVMBuildIntCast },
                Floating => build_into! { Floating: LLVMBuildSIToFP },
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_from_boolean").call(b, [loaded])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            ty if ty.is_sequence() => match as_type {
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_from_seq").call(b, [source.loaded(c, b)])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            Tuple(tuple) => match as_type {
                Tuple(as_tuple) if tuple.len() == as_tuple.len() => unsafe {
                    let as_llvm = TypeLowerer::new(c, as_type).lower();
                    let from_tuple = TypeLowerer::new(c, from_type).lower();
                    let ptr = LLVMBuildAlloca(b, as_llvm, cstr!());

                    tuple.iter().enumerate().for_each(|(idx, from)| {
                        // Get the value that we must coerce.
                        let from_ptr = LLVMBuildStructGEP2(b, from_tuple, loaded, idx as u32, cstr!());
                        let from_val = Value::Alloca(from.clone(), TypeLowerer::new(c, from).lower(), from_ptr);

                        // Because of how we define tuples, we don't have to register the temp
                        // values here, they should only be integers, floats or booleans.
                        let value = match CastLowerer::from_llvm(llvm, &from_val, as_type, ctx).lower_static() {
                            CastedValue::Unchanged { value } | CastedValue::Static { value } => value.loaded(c, b),
                            casted => unreachable!("should have statically casted, got: {casted:?}"),
                        };
                        let mut idxs = [
                            LLVMConstInt(LLVMInt32TypeInContext(c), 0, 0),
                            LLVMConstInt(LLVMInt32TypeInContext(c), idx as u64, 0),
                        ];
                        let ptr = LLVMBuildGEP2(b, as_llvm, ptr, idxs.as_mut_ptr(), idxs.len() as u32, cstr!());
                        LLVMBuildStore(b, value, ptr);
                    });

                    CastedValue::Static { value: Value::Alloca(as_type.clone(), as_llvm, ptr) }
                },

                ty if ty.is_numeric() && tuple.len() == 1 => unsafe {
                    let from_val = Value::Alloca(
                        tuple[0].clone(),
                        TypeLowerer::new(c, &tuple[0]).lower(),
                        LLVMBuildStructGEP2(b, TypeLowerer::new(c, from_type).lower(), loaded, 0, cstr!()),
                    );
                    CastLowerer::from_llvm(llvm, &from_val, ty, ctx).lower_static()
                },

                ty if ty.is_table() => unsafe {
                    let table = ctx.get_value("VVRTTable_new").call(b, []);
                    let from_tuple = TypeLowerer::new(c, from_type).lower();

                    tuple.iter().enumerate().for_each(|(idx, from)| {
                        let from_ptr = LLVMBuildStructGEP2(b, from_tuple, loaded, idx as u32, cstr!());
                        let from_val = Value::Alloca(from.clone(), TypeLowerer::new(c, from).lower(), from_ptr);

                        // Here we don't need to clone because we know that all types whithing a
                        // tuple are trivially copiable.
                        let value = match CastLowerer::from_llvm(llvm, &from_val, &Any, ctx).lower_static() {
                            CastedValue::Unchanged { value } | CastedValue::Static { value } => value.loaded(c, b),
                            casted => unreachable!("should have statically casted, got: {casted:?}"),
                        };

                        let key = idx.to_string();
                        let (key, len) = (
                            LLVMConstStringInContext(c, key.as_ptr() as *const _, key.len() as u32, 1),
                            LLVMConstInt(LLVMInt32TypeInContext(c), key.len() as u64, 0),
                        );
                        ctx.get_value("VVRTTable_insert_from_raw_key")
                            .call(b, [table, key, len, value]);
                    });

                    let ptr = LLVMBuildAlloca(b, VVRTTable::llvm_type(c), cstr!());
                    LLVMBuildStore(b, table, ptr);
                    CastedValue::Static { value: Value::Alloca(AnyTable, VVRTTable::llvm_type(c), ptr) }
                },

                ty if ty.is_sequence() => unsafe {
                    let seq = ctx.get_value("VVRTSeq_new").call(b, []);
                    let from_tuple = TypeLowerer::new(c, from_type).lower();

                    tuple.iter().enumerate().for_each(|(idx, from)| {
                        let from_ptr = LLVMBuildStructGEP2(b, from_tuple, loaded, idx as u32, cstr!());
                        let from_val = Value::Alloca(from.clone(), TypeLowerer::new(c, from).lower(), from_ptr);

                        // Here we don't need to clone because we know that all types whithing a
                        // tuple are trivially copiable.
                        let value = match CastLowerer::from_llvm(llvm, &from_val, &Any, ctx).lower_static() {
                            CastedValue::Unchanged { value } | CastedValue::Static { value } => value.loaded(c, b),
                            casted => unreachable!("should have statically casted, got: {casted:?}"),
                        };

                        ctx.get_value("VVRTSeq_push").call(b, [seq, value]);
                    });

                    let ptr = LLVMBuildAlloca(b, VVRTSeq::llvm_type(c), cstr!());
                    LLVMBuildStore(b, seq, ptr);
                    CastedValue::Static { value: Value::Alloca(AnySequence, VVRTSeq::llvm_type(c), ptr) }
                },

                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            ty if ty.is_variant() => match as_type {
                as_type if as_type.is_table() => todo!(),
                String => store_into_ptr(String, ctx.get_value("VVRTVariant_to_string").call(b, [loaded])),
                Any => store_into_ptr(Any, ctx.get_value("VVRTAny_from_variant").call(b, [loaded])),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            // Forbiden casts or Dynamic casts
            ty => panic!("can't cast a value of type `{ty}` or this type need a runtime check to do the cast"),
        }
    }

    /// Try to do the lowering at runtime, if it failed we guarenty that the value will be nil and
    /// that no values will need to be dropped. In case of success we calling function must
    /// register the value in the context.
    pub fn lower_dynamic(self) -> CastedValue {
        use ASTType::*;
        let Self { llvm: (_, c, b), source, from_type, as_type, ctx } = self;

        if matches!(as_type, Nil) && !matches!(from_type, ASTType::Any) {
            return CastedValue::Unchanged { value: ConstantLowerer::new(c, &ASTConst::Nil).lower() };
        } else if Self::is_same_type(from_type, as_type) {
            return CastedValue::Unchanged { value: source.clone() };
        }

        let loaded = source.loaded(c, b);
        let store_into_ptr = |ty: ASTType, value: LLVMValueRef| unsafe {
            let ptr = LLVMBuildAlloca(b, LLVMTypeOf(value), cstr!());
            LLVMBuildStore(b, value, ptr);
            Value::Alloca(ty, LLVMTypeOf(value), ptr)
        };
        let any_is_ty = |ty: VVRTType| {
            ctx.get_value("VVRTAny_is_ty")
                .call(b, [loaded, unsafe { ty.as_llvm_const(c) }])
        };

        match from_type {
            ty if ty.is_table() => match as_type {
                Any => CastedValue::Static {
                    value: store_into_ptr(Any, ctx.get_value("VVRTAny_from_table").call(b, [loaded])),
                },
                ty if ty.is_variant() => todo!(),
                ty if ty.is_sequence() => todo!(),
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            // Get value out of any.
            Any => match as_type {
                Nil => CastedValue::Dynamic {
                    flag: any_is_ty(VVRTType::Nil),
                    value: ConstantLowerer::new(c, &ASTConst::Nil).lower(),
                },
                Integer => CastedValue::Dynamic {
                    flag: any_is_ty(VVRTType::Number),
                    value: store_into_ptr(Integer, ctx.get_value("VVRTAny_maybe_as_integer").call(b, [loaded])),
                },
                Boolean => CastedValue::Dynamic {
                    flag: any_is_ty(VVRTType::Number),
                    value: store_into_ptr(Boolean, ctx.get_value("VVRTAny_maybe_as_boolean").call(b, [loaded])),
                },
                Floating => CastedValue::Dynamic {
                    flag: any_is_ty(VVRTType::Number),
                    value: store_into_ptr(Floating, ctx.get_value("VVRTAny_maybe_as_floating").call(b, [loaded])),
                },
                String => {
                    let val = ctx.get_value("VVRTAny_maybe_as_string").call(b, [loaded]);
                    let val = ctx.get_value("VVRTString_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::String), value: store_into_ptr(String, val) }
                }
                Syllabe => {
                    let val = ctx.get_value("VVRTAny_maybe_as_syllabe").call(b, [loaded]);
                    let val = ctx.get_value("VVRTSyllabe_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::ASSSyllabe), value: store_into_ptr(Syllabe, val) }
                }
                Line => {
                    let val = ctx.get_value("VVRTAny_maybe_as_line").call(b, [loaded]);
                    let val = ctx.get_value("VVRTLine_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::ASSLine), value: store_into_ptr(Line, val) }
                }
                ty if ty.is_variant() => {
                    let val = ctx.get_value("VVRTAny_maybe_as_variant").call(b, [loaded]);
                    let val = ctx.get_value("VVRTVariant_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::Variant), value: store_into_ptr(Variant, val) }
                }
                ty if ty.is_table() => {
                    let val = ctx.get_value("VVRTAny_maybe_as_table").call(b, [loaded]);
                    let val = ctx.get_value("VVRTTable_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::Table), value: store_into_ptr(AnyTable, val) }
                }
                ty if ty.is_sequence() => {
                    let val = ctx.get_value("VVRTAny_maybe_as_seq").call(b, [loaded]);
                    let val = ctx.get_value("VVRTSeq_clone").call(b, [val]);
                    CastedValue::Dynamic { flag: any_is_ty(VVRTType::Seq), value: store_into_ptr(AnySequence, val) }
                }
                as_type => panic!("can't convert `{from_type} into {as_type}`"),
            },

            // Don't need to perform a runtime check to do the cast, use the lower function
            _ => self.lower_static(),
        }
    }
}
