use crate::Value;
use std::mem;
use vvs_lang::ast::*;
use vvs_llvm::*;

/// Lower a single constant. We only need the LLVM context what should be the type...
pub(crate) struct ConstantLowerer<'a> {
    /// The context.
    c: LLVMContextRef,

    /// We expression we want to lower.
    constant: &'a ASTConst,
}

impl<'a> ConstantLowerer<'a> {
    /// Create a new constant expression lowerer.
    pub fn new(c: LLVMContextRef, constant: &'a ASTConst) -> Self {
        Self { c, constant }
    }

    /// Create a sub-lowerer, for constants that need a recursive descent.
    fn sub_lowerer(&self, constant: &'a ASTConst) -> Self {
        Self { c: self.c, constant }
    }

    /// Lower the constant.
    pub fn lower(self) -> Value {
        let Self { c, constant } = self;
        let ty = constant.get_const_type();
        use {ASTConst::*, Value::*};
        unsafe {
            match constant {
                ASTConst::Nil => Const(ty, LLVMConstPointerNull(LLVMInt64TypeInContext(c))),
                False => Const(ty, LLVMConstInt(LLVMInt1TypeInContext(c), 0, 0)),
                True => Const(ty, LLVMConstInt(LLVMInt1TypeInContext(c), 1, 0)),
                Floating(f) => Const(ty, LLVMConstReal(LLVMFloatTypeInContext(c), *f as f64)),
                String(s) => {
                    let (ptr, len) = (s.as_bytes().as_ptr() as *const _, s.len().try_into().unwrap());
                    Const(ty, LLVMConstString(ptr, len, 1))
                }
                Integer(i) => {
                    let int = mem::transmute::<i64, u64>(*i as i64);
                    Const(ty, LLVMConstInt(LLVMInt32TypeInContext(c), int, 1))
                }
                Tuple(inner) => {
                    let i = inner.iter().map(|cnst| match self.sub_lowerer(cnst).lower() {
                        Value::Const(_, val) => val,
                        val => unreachable!("expected a constant, got {val:?}"),
                    });
                    let mut i = i.collect::<Vec<_>>();
                    let (ptr, len) = (i.as_mut_ptr(), i.len().try_into().unwrap());
                    Const(ty, LLVMConstStructInContext(c, ptr, len, 0))
                }
                Color(_) => panic!("need to implement colors"),
                Table(_) => panic!("need to implement tables"),
            }
        }
    }
}
