use crate::{value::Value, Graph, ValueContext};
use std::collections::BTreeSet;
use vvs_llvm::sys::*;

/// Drop all the values in the correct order at the end of a given basic block. We take into
/// account any terminator of said basic block.
pub(super) struct DropLowerer<'a> {
    /// The context.
    c: LLVMContextRef,

    /// Our own builder.
    b: LLVMBuilderRef,

    /// The values to drop, they are in the order of their declaration. We drop them in the reverse
    /// order!
    values: &'a [Value],

    /// The function from which the values must be dropped.
    func: LLVMValueRef,

    /// The basic block to use to drop things. We will create our own builder to do the thing, so
    /// we don't need others things to pass their builder.
    bb: LLVMBasicBlockRef,

    /// The value context.
    ctx: &'a ValueContext,
}

impl<'a> DropLowerer<'a> {
    /// Create a new lowerer to drop the specified values from a given function at the end of a
    /// given basic block.
    ///
    /// # Safety
    /// All the values must have been declared inside the passed the function. The basic block must
    /// be from the passed function.
    pub unsafe fn new(
        c: LLVMContextRef,
        func: LLVMValueRef,
        values: &'a [Value],
        bb: LLVMBasicBlockRef,
        ctx: &'a ValueContext,
    ) -> Self {
        let b = unsafe { LLVMCreateBuilderInContext(c) };
        Self { c, b, values, func, bb, ctx }
    }

    pub fn lower(self) {
        use LLVMOpcode::*;
        let Self { c, b, values, bb, func, ctx } = self;

        // Position the builder correcly.

        unsafe {
            match LLVMGetLastInstruction(bb) {
                last if last.is_null() => LLVMPositionBuilderAtEnd(b, bb),
                last => match LLVMGetInstructionOpcode(last) {
                    LLVMRet | LLVMBr | LLVMSwitch | LLVMIndirectBr | LLVMCallBr => LLVMPositionBuilderBefore(b, last),
                    _ => LLVMPositionBuilderAtEnd(b, bb),
                },
            };
        };

        // Get the dominance graph to filter the values that we must really drop here!

        let bbs_iter = unsafe { LLVMFunctionIter::new(func) };
        let bbs = Graph::with_capacity(bbs_iter.count()).add_nodes(bbs_iter);
        let bbs = bbs_iter.fold(bbs, |bbs, bb| unsafe {
            match LLVMBasicBlockIter::new(bb).last() {
                Some(instr) => match LLVMGetInstructionOpcode(instr) {
                    LLVMBr => match LLVMGetNumOperands(instr) {
                        1 => bbs.add_edge(bb, LLVMValueAsBasicBlock(LLVMGetOperand(instr, 0))),
                        3 => bbs
                            .add_edge(bb, LLVMValueAsBasicBlock(LLVMGetOperand(instr, 1)))
                            .add_edge(bb, LLVMValueAsBasicBlock(LLVMGetOperand(instr, 2))),
                        n => unreachable!("got br instruction with {n} operands"),
                    },
                    LLVMSwitch => todo!("handle switch, got {} operands", LLVMGetNumOperands(instr)),
                    LLVMIndirectBr => todo!("handle indirectbr, got {} operands", LLVMGetNumOperands(instr)),
                    _ => bbs,
                },
                None => bbs,
            }
        });

        // Filter the values that we must really delete and call the destructors. We can destroy
        // the values in any order because things are reference-counted. Also, just drop the values

        let declared_values: BTreeSet<LLVMValueRef> = unsafe {
            bbs.any_path_content(LLVMGetEntryBasicBlock(func), bb)
                .into_iter()
                .flat_map(|bb| LLVMBasicBlockIter::new(bb))
                .collect()
        };

        values
            .iter()
            .filter(|val| val.llvm_value().map_or(false, |val| declared_values.contains(&val)))
            .flat_map(|val| val.get_dropper_name().map(|drop| (val, drop)))
            .for_each(|(val, drop)| {
                ctx.get_value(drop).call(b, [val.loaded(c, b)]);
            });
    }
}

impl<'a> Drop for DropLowerer<'a> {
    fn drop(&mut self) {
        unsafe { LLVMDisposeBuilder(self.b) };
    }
}
