use crate::{lowerer::*, Value, ValueContext};
use vvs_lang::ast::*;
use vvs_llvm::sys::{LLVMIntPredicate::*, LLVMRealPredicate::*, *};
use vvs_runtime_types::{
    types::{VVRTString, VVRTTable},
    vvll::LLVMExported,
};
use vvs_utils::either;

/// Lower a single expression.
pub(super) struct ExpressionLowerer<'a> {
    llvm: LLVMLowerContext,

    /// We expression we want to lower.
    expr: &'a ASTExpr,

    /// The expected type, to see if we need to generate coertion.
    as_type: Option<&'a ASTType>,

    /// The value context, this stores available already-lowered LLVM values.
    ctx: &'a ValueContext,
}

impl<'a> ExpressionLowerer<'a> {
    /// Create a new expression lowerer.
    pub(super) fn with_assumed_type(llvm: LLVMLowerContext, expr: &'a ASTExpr, ctx: &'a ValueContext) -> Self {
        Self { llvm, expr, as_type: None, ctx }
    }

    /// Create a new expression lowerer with an expected type for the lowered expression. If needed
    /// the lower process will call [CastLowerer].
    pub(super) fn with_expected_type(
        llvm: LLVMLowerContext,
        expr: &'a ASTExpr,
        as_type: &'a ASTType,
        ctx: &'a ValueContext,
    ) -> Self {
        Self { llvm, expr, as_type: Some(as_type), ctx }
    }

    /// Overwrite the used [ValueContext] used to do the lowering.
    pub(super) fn with_ctx(mut self, ctx: &'a ValueContext) -> Self {
        self.ctx = ctx;
        self
    }

    /// Set the type that is expected by the expression, calling the [CastLowerer] if needed!
    pub(super) fn into_type(mut self, ty: &'a ASTType) -> Self {
        self.as_type = Some(ty);
        self
    }

    /// Create a new sub-expression lowerer. Used for recusive descent to lower complexe
    /// expressions, thus private.
    fn sub_lowerer(&self, expr: &'a ASTExpr) -> Self {
        let Self { llvm, ctx, .. } = *self;
        Self { llvm, expr, as_type: None, ctx }
    }

    /// Lower an equality.
    unsafe fn lower_equality(
        llvm @ (_, c, b): LLVMLowerContext,
        ctx: &ValueContext,
        (l, lty): (LLVMValueRef, &ASTType),
        (r, rty): (LLVMValueRef, &ASTType),
    ) -> LLVMValueRef {
        use ASTType::*;
        match lty {
            Nil => LLVMConstInt(LLVMInt1TypeInContext(c), 1, 0),
            Boolean | Integer => LLVMBuildICmp(b, LLVMIntEQ, l, r, cstr!()),
            Floating => {
                let flt = LLVMFloatTypeInContext(c);
                let diff = LLVMBuildFSub(b, l, r, c"".as_ptr());
                let diff = CallLowerer::call_intrinsic("llvm.fabs.f32", llvm, flt, [diff]);
                LLVMBuildFCmp(b, LLVMRealOLE, diff, LLVMConstReal(flt, 10e-5), c"".as_ptr())
            }

            Tuple(lt) => match rty {
                Tuple(rt) if lt.len() == rt.len() => lt.iter().zip(rt).enumerate().fold(
                    LLVMConstInt(LLVMInt1TypeInContext(c), 1, 0),
                    |acc, (idx, (lt, rt))| {
                        let idx = idx.try_into().expect("too many elements");
                        let l = LLVMBuildStructGEP2(b, LLVMTypeOf(l), l, idx, cstr!());
                        let r = LLVMBuildStructGEP2(b, LLVMTypeOf(r), r, idx, cstr!());
                        let ret = Self::lower_equality(llvm, ctx, (l, lt), (r, rt));
                        LLVMBuildAnd(b, ret, acc, cstr!())
                    },
                ),
                _ => LLVMConstInt(LLVMInt1TypeInContext(c), 0, 0),
            },

            Any => ctx.get_value("VVRTAny_eq").call(b, [l, r]),
            Line => ctx.get_value("VVRTLine_eq").call(b, [l, r]),
            String => ctx.get_value("VVRTString_eq").call(b, [l, r]),
            Syllabe => ctx.get_value("VVRTSyllabe_eq").call(b, [l, r]),
            ty if ty.is_sequence() => ctx.get_value("VVRTSeq_eq").call(b, [l, r]),
            ty if ty.is_variant() => ctx.get_value("VVRTVariant_eq").call(b, [l, r]),
            ty if ty.is_table() => ctx.get_value("VVRTTable_eq").call(b, [l, r]),

            _ => panic!("can't apply `{:?}` to expressions of type `{lty}` and `{rty}`", ASTBinop::CmpEQ),
        }
    }

    /// Lower the value into LLVM-IR at the position specified by the builder passed at
    /// construction time. If any value is created it will be added in the returned [ValueContext].
    pub(super) fn lower(self) -> (ValueContext, Value) {
        let Self { llvm: llvm @ (_, c, b), expr: ASTExpr { content, .. }, as_type, ctx } = self;
        let (bln, int, flt) = (unsafe { LLVMInt1TypeInContext(c) }, unsafe { LLVMInt32TypeInContext(c) }, unsafe {
            LLVMFloatTypeInContext(c)
        });
        let (false_, true_, zero_, one_) = (
            unsafe { LLVMConstInt(bln, 0, 0) },
            unsafe { LLVMConstInt(bln, 1, 0) },
            unsafe { LLVMConstInt(int, 0, 0) },
            unsafe { LLVMConstInt(int, 1, 0) },
        );
        let store_scalar = |ctx: ValueContext, val: LLVMValueRef, ty: LLVMTypeRef, ast: ASTType| unsafe {
            let ptr = LLVMBuildAlloca(b, ty, cstr!());
            LLVMBuildStore(b, val, ptr);
            (ctx, Value::Alloca(ast, ty, ptr))
        };

        match content {
            ASTExprVariant::Var(var) => (ctx.clone(), ctx.get_value(var.name()).clone()),

            ASTExprVariant::Const(value) => {
                let value = ConstantLowerer::new(c, value).lower();
                match as_type {
                    Some(expected_type) if expected_type != value.get_ast_type() => {
                        match CastLowerer::from_llvm(llvm, &value, expected_type, ctx).lower_static() {
                            CastedValue::Unchanged { value } => (ctx.clone(), value),
                            CastedValue::Static { value } => (ctx.to_register_anon(value.clone()), value),
                            CastedValue::Dynamic { .. } => unreachable!(),
                        }
                    }
                    _ => (ctx.clone(), value),
                }
            }

            ASTExprVariant::Tuple(tuple) => unsafe {
                let (mut lowered_tuple, mut types, mut tuple_ty) =
                    (Vec::with_capacity(tuple.len()), Vec::with_capacity(tuple.len()), Vec::with_capacity(tuple.len()));

                let ctx = tuple.iter().fold(self.ctx.clone(), |ctx, item| {
                    let (ctx, expr) = self.sub_lowerer(item).with_ctx(&ctx).lower();
                    let (ast_type, expr) = (expr.get_ast_type(), expr.loaded(c, b));
                    lowered_tuple.push(expr);
                    types.push(LLVMTypeOf(expr));
                    tuple_ty.push(ast_type.clone());
                    ctx
                });

                let tuple_len = types.len().try_into().expect("too many elements");
                let types = LLVMStructTypeInContext(c, types.as_mut_ptr(), tuple_len, 0);
                let tuple_ptr = LLVMBuildAlloca(b, types, c"".as_ptr());

                for (idx, val) in lowered_tuple.into_iter().enumerate() {
                    let idx = idx.try_into().expect("too many elements");
                    let ptr = LLVMBuildStructGEP2(b, types, tuple_ptr, idx, c"".as_ptr());
                    LLVMBuildStore(b, val, ptr);
                }

                (ctx, Value::Alloca(ASTType::Tuple(tuple_ty), types, tuple_ptr))
            },

            ASTExprVariant::Table(table) => unsafe {
                let ty = VVRTTable::llvm_type(c);
                let ptr = LLVMBuildAlloca(b, ty, c"".as_ptr());
                LLVMBuildStore(b, ctx.get_value("VVRTTable_new").call(b, []), ptr);
                let mut ctx = table.iter().fold(self.ctx.clone(), |_, (key, value)| {
                    // Static key! The table will create it's own string.
                    let key_len: u32 = key.len().try_into().expect("too many elements");
                    let key = LLVMConstStringInContext(c, key.as_ptr() as *const _, key_len, 1);
                    let key_len = LLVMConstInt(LLVMInt32TypeInContext(c), key_len as u64, 0);

                    // Lower the value and cast it into an Any.
                    let (mut ctx, value) = self.sub_lowerer(value).into_type(&ASTType::Any).lower();
                    let value = value.loaded(c, b);
                    ctx.forget_by_llvm_value(value);

                    // Insert the value with the correct key.
                    ctx.get_value("VVRTTable_insert_from_raw_key")
                        .call(b, [ptr, key, key_len, value]);
                    ctx
                });
                let value = Value::Alloca(ASTType::AnyTable, ty, ptr);
                ctx.register_anon(value.clone());
                (ctx, value)
            },

            ASTExprVariant::FuncCall(func, args) => {
                let (ctx, value) = CallLowerer::new(llvm, self.ctx).lower_func_call(func, args);
                (ctx.to_register_anon(value.clone()), value)
            }
            ASTExprVariant::MethodInvok(obj, method, args) => {
                let (ctx, value) = CallLowerer::new(llvm, self.ctx).lower_method_invok(obj, method, args);
                (ctx.to_register_anon(value.clone()), value)
            }

            ASTExprVariant::Unop(op, inner) => unsafe {
                use vvs_lang::ast::{ASTType::*, ASTUnop::*};

                let (ctx, inner) = match as_type {
                    Some(ty) => self.sub_lowerer(inner).into_type(ty).lower(),
                    None => self.sub_lowerer(inner).lower(),
                };
                let (ty, inner) = (inner.get_ast_type(), inner.loaded(c, b));
                let call_len = |ctx: ValueContext, func: &str| {
                    let value = ctx.get_value(func).call(b, [inner]);
                    store_scalar(ctx, value, int, Integer)
                };

                match op {
                    LogicNot | BitNot | Neg if matches!(ty, Boolean) => {
                        let cmp = LLVMBuildICmp(b, LLVMIntEQ, inner, false_, cstr!());
                        let res = LLVMBuildSelect(b, cmp, true_, false_, cstr!());
                        store_scalar(ctx, res, bln, Boolean)
                    }

                    LogicNot if matches!(ty, Integer) => {
                        let cmp = LLVMBuildICmp(b, LLVMIntEQ, inner, zero_, cstr!());
                        let res = LLVMBuildSelect(b, cmp, one_, zero_, cstr!());
                        store_scalar(ctx, res, int, Integer)
                    }

                    BitNot if matches!(ty, Integer) => {
                        let res = CallLowerer::call_intrinsic("llvm.bitreverse.i32", llvm, int, [inner]);
                        store_scalar(ctx, res, int, Integer)
                    }

                    #[rustfmt::skip] Neg if matches!(ty, Integer ) => store_scalar(ctx, LLVMBuildNeg (b, inner, cstr!()), int, Integer ),
                    #[rustfmt::skip] Neg if matches!(ty, Floating) => store_scalar(ctx, LLVMBuildFNeg(b, inner, cstr!()), flt, Floating),

                    #[rustfmt::skip] Len => match ty {
                        Line                   => call_len(ctx, "VVRTLine_len"),
                        String                 => call_len(ctx, "VVRTString_len"),
                        Syllabe                => call_len(ctx, "VVRTSyllabe_len"),
                        ty if ty.is_sequence() => call_len(ctx, "VVRTSeq_len"),
                        ty if ty.is_table()    => call_len(ctx, "VVRTTable_len"),
                        Tuple(len) => {
                            let len = TryInto::<i32>::try_into(len.len()).expect("too many elements") as u64;
                            (ctx, Value::Const(Integer, LLVMConstInt(int, len, 0)))
                        }
                        ty => panic!("can't apply `{op:?}` to expression of type `{ty}`"),
                    },

                    op => panic!("can't apply `{op:?}` to expression of type `{ty}`"),
                }
            },

            ASTExprVariant::Binop(left, op, right) => unsafe {
                use vvs_lang::ast::{ASTBinop::*, ASTType::*};

                let (ctx, left) = match as_type {
                    Some(ty) => self.sub_lowerer(left).into_type(ty).lower(),
                    None => self.sub_lowerer(left).lower(),
                };
                let lty = left.get_ast_type();
                let (ctx, right) = self
                    .sub_lowerer(right)
                    .into_type(either!(*op == Power => &Integer; lty))
                    .with_ctx(&ctx)
                    .lower();
                let (l, r, rty) = (left.loaded(c, b), right.loaded(c, b), right.get_ast_type());
                let (is_int, is_flt, is_bln) =
                    (matches!(lty, Integer), matches!(lty, Floating), matches!(lty, Boolean));

                let (res, ty, ast) = match op {
                    Power if is_int => todo!(),
                    Power if is_flt => todo!(),

                    #[rustfmt::skip] Mul if is_int => (LLVMBuildMul (b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] Div if is_int => (LLVMBuildSDiv(b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] Add if is_int => (LLVMBuildAdd (b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] Sub if is_int => (LLVMBuildSub (b, l, r, cstr!()), int, Integer),

                    #[rustfmt::skip] Mul if is_flt => (LLVMBuildFMul(b, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] Div if is_flt => (LLVMBuildFDiv(b, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] Add if is_flt => (LLVMBuildFAdd(b, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] Sub if is_flt => (LLVMBuildFSub(b, l, r, cstr!()), flt, Floating),

                    #[rustfmt::skip] Mod if is_int => (LLVMBuildSRem(b, l, r, cstr!()), int, Integer ),
                    #[rustfmt::skip] Mod if is_flt => (LLVMBuildFRem(b, l, r, cstr!()), flt, Floating),

                    #[rustfmt::skip] StrCat if matches!(lty, String) => {
                        (ctx.get_value("VVRTString_cat").call(b, [l, r]), VVRTString::llvm_type(c), String)
                    },

                    #[rustfmt::skip] CmpLE if is_int => (LLVMBuildICmp(b, LLVMIntSLE,  l, r, cstr!()), bln, Boolean),
                    #[rustfmt::skip] CmpLT if is_int => (LLVMBuildICmp(b, LLVMIntSLT,  l, r, cstr!()), bln, Boolean),
                    #[rustfmt::skip] CmpGE if is_int => (LLVMBuildICmp(b, LLVMIntSGE,  l, r, cstr!()), bln, Boolean),
                    #[rustfmt::skip] CmpGT if is_int => (LLVMBuildICmp(b, LLVMIntSGT,  l, r, cstr!()), bln, Boolean),

                    #[rustfmt::skip] CmpLE if is_flt => (LLVMBuildFCmp(b, LLVMRealOLE, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] CmpLT if is_flt => (LLVMBuildFCmp(b, LLVMRealOLT, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] CmpGE if is_flt => (LLVMBuildFCmp(b, LLVMRealOGE, l, r, cstr!()), flt, Floating),
                    #[rustfmt::skip] CmpGT if is_flt => (LLVMBuildFCmp(b, LLVMRealOGT, l, r, cstr!()), flt, Floating),

                    #[rustfmt::skip] BitAnd | LogicAnd if is_bln => (LLVMBuildAnd(b, l, r, cstr!()), bln, Boolean),
                    #[rustfmt::skip] BitXor | LogicXor if is_bln => (LLVMBuildXor(b, l, r, cstr!()), bln, Boolean),
                    #[rustfmt::skip] BitOr  | LogicOr  if is_bln => (LLVMBuildOr (b, l, r, cstr!()), bln, Boolean),

                    #[rustfmt::skip] BitAnd   if is_int => (LLVMBuildAnd(b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] BitXor   if is_int => (LLVMBuildXor(b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] BitOr    if is_int => (LLVMBuildOr (b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] LogicAnd if is_int => (LLVMBuildICmp(b, LLVMIntEQ, LLVMBuildAnd(b, l, r, cstr!()), one_, cstr!()), bln, Boolean),
                    #[rustfmt::skip] LogicXor if is_int => (LLVMBuildICmp(b, LLVMIntEQ, LLVMBuildXor(b, l, r, cstr!()), one_, cstr!()), bln, Boolean),
                    #[rustfmt::skip] LogicOr  if is_int => (LLVMBuildICmp(b, LLVMIntEQ, LLVMBuildOr (b, l, r, cstr!()), one_, cstr!()), bln, Boolean),

                    #[rustfmt::skip] BitShiftLeft  if is_int => (LLVMBuildShl (b, l, r, cstr!()), int, Integer),
                    #[rustfmt::skip] BitShiftRight if is_int => (LLVMBuildAShr(b, l, r, cstr!()), int, Integer),

                    CmpEQ => (Self::lower_equality(llvm, &ctx, (l, lty), (r, rty)), bln, Boolean),
                    CmpNE => {
                        let not = Self::lower_equality(llvm, &ctx, (l, lty), (r, rty));
                        (LLVMBuildNot(b, not, cstr!()), bln, Boolean)
                    }

                    op => panic!("can't apply `{op:?}` to expressions of type `{lty}` and `{rty}`"),
                };
                store_scalar(ctx, res, ty, ast)
            },

            ASTExprVariant::Suffixed(table, _fields) => {
                let (_ctx, _table) = self.sub_lowerer(table).lower();
                todo!()
            }

            ASTExprVariant::Color(variant) | ASTExprVariant::Movement(variant) => {
                let mut arguments = Vec::with_capacity(variant.args_len());
                let _ctx = variant.args().iter().fold(ctx.clone(), |ctx, expr| {
                    let (ctx, expr) = self.sub_lowerer(expr).with_ctx(&ctx).lower();
                    arguments.push(expr);
                    ctx
                });
                let _variant = variant.variant();

                todo!()
            }

            ASTExprVariant::Default(_) => todo!(),

            // See later for clozures
            ASTExprVariant::FuncBind(_, _) => unreachable!(),
        }
    }
}
