use crate::lowerer::*;
use std::sync::OnceLock;
use vvs_lang::{anon_expression, anon_instruction, ast::*};
use vvs_llvm::*;
use vvs_utils::either;

#[derive(Clone, Copy)]
struct BBLoop {
    body_entry: LLVMBasicBlockRef,
    loop_exit: LLVMBasicBlockRef,
}

/// Lower a single instruction.
pub(super) struct InstructionLowerer<'a> {
    llvm: LLVMLowerContext,

    /// The function containing the instruction to lower.
    func: LLVMValueRef,

    /// Remeber the last loop to be able to break or continue.
    last_loop: Option<BBLoop>,

    /// The instruction we want to lower.
    instr: &'a ASTInstr,

    /// The return type of the function, to see if we need to generetate coertion.
    func_return_type: ASTType,

    /// The value context, this stores available already-lowered LLVM values.
    ctx: ValueContext,
}

impl<'a> InstructionLowerer<'a> {
    /// Create a new expression lowerer.
    pub(super) fn new(
        llvm: LLVMLowerContext,
        func: LLVMValueRef,
        instruction: &'a ASTInstr,
        func_return_type: ASTType,
        ctx: ValueContext,
    ) -> Self {
        let last_loop = Default::default();
        Self { llvm, func, instr: instruction, func_return_type, ctx, last_loop }
    }

    /// Create a new sub-expression lowerer. Used for recusive descent to lower complexe
    /// expressions, thus private.
    fn sub_lowerer(&self, instruction: &'a ASTInstr) -> Self {
        let Self { llvm, func, last_loop, .. } = *self;
        let func_return_type = self.func_return_type.clone();
        let ctx = self.ctx.clone();
        Self { llvm, instr: instruction, func, func_return_type, ctx, last_loop }
    }

    /// Short-hand to create an expression lowerer by passing some arguments correctly.
    fn expr(&'a self, expr: &'a ASTExpr) -> ExpressionLowerer<'a> {
        ExpressionLowerer::with_assumed_type(self.llvm, expr, &self.ctx)
    }

    /// Get a new sub-expression lowerer. We need to know the exit block of this scope so that we
    /// can build the calls to drop for this scope.
    fn scoped_lowerer(&mut self, exit_bb: LLVMBasicBlockRef, cb: impl FnOnce(Self) -> ValueContext) {
        static mut NOOP: OnceLock<ASTInstr> = OnceLock::new();

        // Save context.
        let pre_ctx = self.ctx.clone();
        self.ctx = pre_ctx.sub_context();

        // Build an empty instruction to give as parent lowerer in scope.
        let Self { llvm: llvm @ (_, c, _), func, last_loop, .. } = *self;
        let instruction = unsafe { NOOP.get_or_init(|| anon_instruction!(Noop)) };
        let func_return_type = self.func_return_type.clone();
        let ctx = self.ctx.sub_context();
        let this = Self { llvm, func, last_loop, instr: instruction, func_return_type, ctx };

        // Drop what was declared in the scope
        let mut ctx = cb(this);
        unsafe { DropLowerer::new(c, func, &ctx.get_values_to_drop(), exit_bb, &ctx) }.lower();

        // Restore context.
        self.ctx = pre_ctx;
    }

    /// Set the last loop to lower a given instruction.
    fn with_loop(mut self, r#loop: BBLoop) -> Self {
        self.last_loop = Some(r#loop);
        self
    }

    /// Set the value context.
    fn with_ctx(mut self, ctx: ValueContext) -> Self {
        self.ctx = ctx;
        self
    }

    pub(super) fn lower(mut self) -> ValueContext {
        let Self { llvm: llvm @ (_, c, b), func, instr: ASTInstr { content, span }, .. } = self;
        match content {
            // Nop
            ASTInstrVariant::Noop => self.ctx,

            // A simple list of instructions
            ASTInstrVariant::Block(instructions) => unsafe {
                let exit = LLVMAppendBasicBlockInContext(c, self.func, c".block".as_ptr());
                self.scoped_lowerer(exit, |this| {
                    instructions
                        .iter()
                        .fold(this.ctx.clone(), |ctx, instruction| this.sub_lowerer(instruction).with_ctx(ctx).lower())
                });
                LLVMBuildBr(b, exit);
                LLVMPositionBuilderAtEnd(b, exit);
                self.ctx
            },

            // Breaks/Continue
            ASTInstrVariant::Break | ASTInstrVariant::Continue if self.last_loop.is_none() => {
                panic!("try to lower break/continue while not being inside a loop")
            }
            ASTInstrVariant::Break => unsafe {
                LLVMBuildBr(b, self.last_loop.expect("internal error").loop_exit);
                self.ctx
            },
            ASTInstrVariant::Continue => unsafe {
                LLVMBuildBr(b, self.last_loop.expect("internal error").body_entry);
                self.ctx
            },

            // Return from function
            ASTInstrVariant::Return(expr) => unsafe {
                let (ctx, value) = self.expr(expr).lower();
                match value {
                    Value::Nil => LLVMBuildRetVoid(b),
                    Value::Function(..) => panic!("error at `{span}`: can't return a function for now"),
                    value => LLVMBuildRet(b, value.loaded(c, b)),
                };
                ctx
            },

            // Assignation
            ASTInstrVariant::Assign(dests, srcs) if dests.len() != srcs.len() => {
                unreachable!("internal error: verifier failed for assignations at `{span}`")
            }
            ASTInstrVariant::Assign(dests, srcs) => dests.iter().zip(srcs).fold(self.ctx, |ctx, (dest, src)| {
                let (dest, dest_ty) = Self::compute_dest_location(c, b, &ctx, dest);
                let (ctx, src) = ExpressionLowerer::with_expected_type(llvm, src, dest_ty, &ctx).lower();
                Self::store_into(c, b, dest, &src);
                ctx
            }),

            // Declarations
            ASTInstrVariant::Decl(dests, srcs) if dests.len() != srcs.len() => {
                unreachable!("internal error: verifier failed for declarations at `{span}`")
            }
            ASTInstrVariant::Decl(dests, srcs) => {
                let mut declarations = Vec::with_capacity(dests.len());
                let ctx = dests.iter().zip(srcs).fold(self.ctx.clone(), |ctx, (dest, src)| {
                    let dest_ast_ty = dest.get_specified_type().expect("internal error");
                    let dest_llvm_ty = TypeLowerer::new(c, dest_ast_ty).lower();
                    let val = unsafe { LLVMBuildAlloca(b, dest_llvm_ty, cstr!("{}", dest.name())) };
                    let (ctx, expr) = self.expr(src).with_ctx(&ctx).into_type(dest_ast_ty).lower();
                    Self::store_into(c, b, val, &expr);
                    declarations.push((dest.name(), Value::Alloca(dest_ast_ty.clone(), dest_llvm_ty, val)));
                    ctx
                });
                declarations.into_iter().fold(ctx, |mut ctx, (name, val)| {
                    ctx.register(name, val);
                    ctx
                })
            }

            // Call a function, discard the result
            ASTInstrVariant::FuncCall(func, args) => {
                let (ctx, value) = CallLowerer::new(llvm, &self.ctx).lower_func_call(func, args);
                ctx.to_register_anon(value)
            }
            ASTInstrVariant::MethodInvok(obj, method, args) => {
                let (ctx, value) = CallLowerer::new(llvm, &self.ctx).lower_method_invok(obj, method, args);
                ctx.to_register_anon(value)
            }

            // If/Else/Elseif blocks
            ASTInstrVariant::Cond { if_blocks, else_block } => unsafe {
                let exit = LLVMAppendBasicBlockInContext(c, self.func, c".if.exit".as_ptr());
                let pre_ctx = self.ctx.clone();
                let true_cond = ASTCondition::BooleanTrue(anon_expression!(Const(ASTConst::True)));

                let get_bb = move |idx: usize, what: &str| {
                    LLVMAppendBasicBlockInContext(c, self.func, cstr!(".if.{}.{}", idx, what))
                };
                let mut bbs: HashMap<usize, LLVMBasicBlockRef> = Default::default();
                let mut get_bb_cond = move |idx| *bbs.entry(idx).or_insert_with(move || get_bb(idx, "condition"));

                /// Use a macro to not write the same code twice by forgeting something...
                macro_rules! build_cond_block {
                    (($if_num: expr, $if_cond: expr, $if_inner: expr), $else_bb: expr, $exit: expr) => {{
                        let (if_bb_body, if_bb_body_exit) = (get_bb($if_num, "body"), get_bb($if_num, "body.exit"));
                        self.scoped_lowerer(if_bb_body_exit, |mut scoped| {
                            // CONDITION: We don't need to drop things even if we didn't succeded,
                            //            add the new ctx to lower the conditional body. The head
                            //            BB of the conditional is cached by it's number.
                            LLVMPositionBuilderAtEnd(b, get_bb_cond($if_num));
                            let ctx = scoped.lower_condition($if_cond, if_bb_body, $else_bb);

                            // BODY: We need to drop things after the exit of the loop, we also
                            //       drop the conditional stuff
                            LLVMPositionBuilderAtEnd(b, if_bb_body);
                            let ctx = scoped.sub_lowerer($if_inner).with_ctx(ctx).lower();
                            LLVMBuildBr(b, if_bb_body_exit);
                            LLVMPositionBuilderAtEnd(b, if_bb_body_exit);
                            LLVMBuildBr(b, $exit);

                            // WARN: scoped.ctx -> contains also the temporary values! We must drop
                            //       all of those if we succeded or not.
                            DropLowerer::new(c, func, &scoped.ctx.get_values_to_drop(), $else_bb, &ctx).lower();
                            DropLowerer::new(c, func, &scoped.ctx.get_values_to_drop(), if_bb_body_exit, &ctx).lower();

                            ctx
                        })
                    }};
                }

                let else_block = else_block.as_ref().map(|instr| (&true_cond, instr.as_ref()));
                let if_blocks = if_blocks.iter().map(|(cond, instr): &_| (cond, instr));
                let blocks = if_blocks.chain(else_block).enumerate().collect::<Vec<(_, (&_, &_))>>();
                assert!(!blocks.is_empty(), "empty conditionals not allowed");

                // IF -> Need to jump to the first condition checker
                LLVMBuildBr(b, get_bb_cond(0));

                // Build IF, for ELSE -> will be handled at next iteration
                blocks.windows(2).for_each(|slice| match slice {
                    [(if_num, (if_cond, if_inner)), (else_num, _)] => {
                        build_cond_block!((*if_num, if_cond, if_inner), get_bb_cond(*else_num), exit)
                    }
                    _ => unreachable!(),
                });

                // ELSE / IF in the case where we have only one conditional -> Handle last item
                let (else_num, (else_cond, else_inner)) = blocks.last().expect("internal error");
                build_cond_block!((*else_num, else_cond, else_inner), exit, exit);

                LLVMPositionBuilderAtEnd(b, exit);
                pre_ctx
            },

            // While loops
            ASTInstrVariant::WhileDo(condition, inner) => unsafe {
                let (cond, exit, body, body_exit) = (
                    LLVMAppendBasicBlockInContext(c, self.func, c".while.cond".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".while.exit".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".while.body".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".while.body.exit".as_ptr()),
                );

                self.scoped_lowerer(body_exit, |mut scoped| {
                    // For a WhileDo the condition binding variables are only visible in the body
                    // of the loop. Here we ensure variables are only available inside the loop.
                    LLVMBuildBr(b, cond);
                    LLVMPositionBuilderAtEnd(b, cond);
                    let ctx = scoped.lower_condition(condition, body, exit);

                    LLVMPositionBuilderAtEnd(b, body);
                    let ctx = scoped
                        .sub_lowerer(inner)
                        .with_ctx(ctx)
                        .with_loop(BBLoop { body_entry: body, loop_exit: exit })
                        .lower();

                    // Drop what was
                    LLVMBuildBr(b, body_exit);
                    LLVMPositionBuilderAtEnd(b, body_exit);
                    DropLowerer::new(c, func, &scoped.ctx.get_values_to_drop(), body_exit, &ctx).lower();
                    LLVMBuildBr(b, cond);
                    ctx
                });

                LLVMPositionBuilderAtEnd(b, exit);
                self.ctx
            },

            // For loop
            ASTInstrVariant::ForLoop { var, lower: l, upper: u, step, body: inner } => unsafe {
                let (cond, exit, body, body_exit) = (
                    LLVMAppendBasicBlockInContext(c, self.func, c".forloop.cond".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".forloop.exit".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".forloop.body".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".forloop.body.exit".as_ptr()),
                );

                let (ctx, lower) = self.expr(l).into_type(&ASTType::Integer).lower();
                let (mut ctx, upper) = self.expr(u).with_ctx(&ctx).into_type(&ASTType::Integer).lower();
                let (lower, upper) = (lower.loaded(c, b), upper.loaded(c, b));

                // Setup the index
                let t_i32 = LLVMInt32TypeInContext(c);
                let step = step.unwrap_or(1) as i64;
                assert!(step > 0, "the for-loop step must be strictly positive");
                let step = LLVMConstInt(t_i32, std::mem::transmute_copy(&step), 0);

                let index = LLVMBuildAlloca(b, t_i32, cstr!(".index.{}", var));
                ctx.register(var.name(), Value::Alloca(ASTType::Integer, t_i32, index));
                LLVMBuildStore(b, lower, index);

                // Build the condition check
                LLVMBuildBr(b, cond);
                LLVMPositionBuilderAtEnd(b, cond);
                const PREDICATE: LLVMIntPredicate = LLVMIntPredicate::LLVMIntSLT;
                let loaded = LLVMBuildLoad2(b, t_i32, index, cstr!(".index.{}.loaded", var));
                let cont = LLVMBuildICmp(b, PREDICATE, loaded, upper, cstr!(".index.{}.flag", var));
                LLVMBuildCondBr(b, cont, body, exit);

                // Lower the body, we need a scope for that.
                self.scoped_lowerer(body_exit, |scoped| {
                    LLVMPositionBuilderAtEnd(b, body);
                    let ctx = scoped
                        .sub_lowerer(inner)
                        .with_loop(BBLoop { body_entry: body, loop_exit: exit })
                        .lower();
                    let loaded = LLVMBuildLoad2(b, t_i32, index, cstr!(".index.{}.loaded", var));
                    let loaded = LLVMBuildAdd(b, loaded, step, cstr!(".index.{}.incremented", var));
                    LLVMBuildStore(b, loaded, index);
                    LLVMBuildBr(b, body_exit);
                    LLVMPositionBuilderAtEnd(b, body_exit);
                    LLVMBuildBr(b, cond);
                    ctx
                });

                // End of loop
                LLVMPositionBuilderAtEnd(b, exit);
                ctx
            },
            ASTInstrVariant::ForInto { .. } => unsafe {
                let (_cond, _exit, _body, _body_exit) = (
                    LLVMAppendBasicBlockInContext(c, self.func, c".foreach.cond".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".foreach.exit".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".foreach.body".as_ptr()),
                    LLVMAppendBasicBlockInContext(c, self.func, c".foreach.body.exit".as_ptr()),
                );
                todo!()
            },
        }
    }

    /// Store the content of a value into a location in memory.
    ///
    /// TODO: Do that in the Value enum
    fn store_into(c: LLVMContextRef, b: LLVMBuilderRef, dest: LLVMValueRef, src: &Value) {
        match src {
            Value::Function(_, _, _) => panic!("we don't handle function pointers for now"),
            Value::Const(_, v) => unsafe { LLVMBuildStore(b, *v, dest) },
            Value::Nil => match ConstantLowerer::new(c, &ASTConst::Nil).lower() {
                Value::Const(_, v) => unsafe { LLVMBuildStore(b, v, dest) },
                _ => unreachable!(),
            },
            Value::Alloca(_, ty, v) => unsafe {
                let v = LLVMBuildLoad2(b, *ty, *v, cstr!());
                LLVMBuildStore(b, v, dest)
            },
        };
    }

    /// Get the address and the type of a variable to store into it.
    fn compute_dest_location<'b>(
        _c: LLVMContextRef,
        _b: LLVMBuilderRef,
        _ctx: &'b ValueContext,
        _expr: &'b ASTExpr,
    ) -> (LLVMValueRef, &'b ASTType) {
        todo!()
    }

    /// Lower the condition. Returns the new context with the added binded values if needed. If the
    /// condition is not checked we don't need to drop anything from the returned value context.
    /// This means that we must drop the returned context only on the then condition path.
    ///
    /// If things where allocated to check the condition, then we add them to the current context,
    /// but not to the returned context! This means that may the condition succeed or not, you must
    /// drop the context of the calling lowerer at the end of the `then` and `else` paths.
    ///
    /// Most of the time the context is already scopped, so it shouldn't be too difficult trop
    /// things at the right moment.
    fn lower_condition(
        &mut self,
        condition: &ASTCondition,
        bb_then: LLVMBasicBlockRef,
        bb_else: LLVMBasicBlockRef,
    ) -> ValueContext {
        let Self { llvm: llvm @ (_, c, b), .. } = *self;
        match condition {
            ASTCondition::BooleanTrue(cond) => unsafe {
                let (mut ctx, cond) = self.expr(cond).into_type(&ASTType::Boolean).lower();
                LLVMBuildCondBr(b, cond.loaded(c, b), bb_then, bb_else);

                let ret_ctx = self.ctx.clone();
                self.ctx.add_values_to_drop(ctx.get_values_to_drop());
                ret_ctx
            },

            ASTCondition::CoercibleInto { new, as_type, source: src }
                if src
                    .get_specified_type()
                    .expect("should be setted")
                    .coercible_to(as_type) =>
            unsafe {
                let llvm_true = LLVMConstInt(LLVMInt1TypeInContext(c), 1, 0);
                let (register, flag, src) = match CastLowerer::from_ast(llvm, src, as_type, &self.ctx).lower_dynamic() {
                    CastedValue::Dynamic { flag, value } => (true, flag, value),
                    CastedValue::Static { value } => (true, llvm_true, value),
                    CastedValue::Unchanged { value } => (false, llvm_true, value),
                };
                LLVMBuildCondBr(b, flag, bb_then, bb_else);

                LLVMPositionBuilderAtEnd(b, bb_then);
                let ty = TypeLowerer::new(c, as_type).lower();
                let dest = LLVMBuildAlloca(b, ty, cstr!(".{}.alloca", new));
                LLVMBuildStore(b, src.loaded(c, b), dest);

                either!(!register => self.ctx.clone();
                    self.ctx.to_register(new.name(), Value::Alloca(as_type.clone(), ty, dest))
                )
            },

            ASTCondition::CoercibleInto { .. } => unsafe {
                LLVMBuildBr(b, bb_else);
                self.ctx.clone()
            },

            ASTCondition::BindVariant { .. } => todo!(),
        }
    }
}
