mod calls;
mod cast;
mod constant;
mod drops;
mod expression;
mod function;
mod instruction;
mod types;

use crate::{
    cstr,
    lowerer::{calls::*, cast::*, drops::*, expression::*, function::*, instruction::*, types::*},
    value::Value,
    LLVMLowerContext, ValueContext,
};
use hashbrown::HashMap;
use std::ffi::{CStr, CString};
use vvs_lang::ast::*;
use vvs_llvm::*;
use vvs_runtime_types::{VVRTSymbol, VVRTSymbolType};

// Re-export the constant lowerer because it can be usefull in other parts of the codegen.
pub(crate) use self::constant::ConstantLowerer;

pub struct Lowerer<'a> {
    llvm: LLVMLowerContext,

    functions: Vec<(ASTString, ASTString, &'a ASTFunction)>,
    globals: HashMap<(ASTString, ASTString), &'a ASTConst>,

    ctx: ValueContext,
}

pub struct LowererBuilder<'a> {
    name: CString,
    c: LLVMContextRef,
    m: Option<LLVMModuleRef>,
    b: LLVMBuilderRef,
    programs: Option<&'a ASTProgram>,
    modules: Vec<&'a ASTModule>,
}

impl<'a> LowererBuilder<'a> {
    /// Create a new lowerer builder. We need a name and the context from the JIT thing.
    ///
    /// # Safety
    /// This function is unsafe because we use the LLVM context pointer passed from anywhere. This
    /// pointer must be not null and be a valid LLVM context.
    pub unsafe fn new(name: impl AsRef<str>, c: LLVMContextRef) -> Self {
        Self {
            c,
            name: CString::new(name.as_ref()).expect("invalid module name"),
            b: LLVMCreateBuilderInContext(c),
            m: Default::default(),
            modules: Default::default(),
            programs: Default::default(),
        }
    }

    /// Add a complete program to lower. If any other program was present, overwrite it.
    pub fn program(mut self, program: &'a ASTProgram) -> Self {
        self.programs = Some(program);
        self
    }

    /// Add a module to lower to the list of things to lower.
    pub fn module(mut self, module: &'a ASTModule) -> Self {
        self.modules.push(module);
        self
    }

    /// Build the lowerer from this builder.
    pub fn build(self) -> Lowerer<'a> {
        let Self { c, m, b, programs, modules, name } = self;

        // Get all the modules that will be lowered inside the final LLVM module. By doing this we
        // have a similar thing to LTO for Vivy Script.

        let init = HashMap::<ASTString, &'a ASTModule>::default();
        let modules = modules
            .into_iter()
            .chain(programs.iter().flat_map(|prog| prog.modules()))
            .fold(init, |mut map, module| match map.entry(module.name()) {
                hashbrown::hash_map::Entry::Occupied(_) => map,
                hashbrown::hash_map::Entry::Vacant(entry) => {
                    entry.insert(module);
                    map
                }
            });

        // Get the globals that needs to be lowered / needs to be known by the lowerer. We first
        // collect the setted options from the program as those will overwrite values from the
        // modules, then we get what is defined as default in the module. To ensure we doesn't
        // overwrite things, we only insert in the final hashmap if the thing was not already
        // defined in the hashmap. Other than setted overwriting defaults, there should not be any
        // other collisions.

        let init: HashMap<(ASTString, ASTString), &'a ASTConst> = programs
            .iter()
            .flat_map(|prog| {
                prog.setted_options()
                    .iter()
                    .map(|((m, n), v)| ((m.clone(), n.clone()), v))
            })
            .collect();
        let globals = modules.iter().flat_map(|(mname, m)| {
            let hdl = |(var, val): (&'a ASTVar, &'a _)| (mname.clone(), var.name(), val);
            m.consts(ASTVisibilityRule::any()).map(hdl).chain(m.options().map(hdl))
        });
        let globals = globals.fold(init, |mut globals, (mn, n, val)| match globals.entry((mn, n)) {
            hashbrown::hash_map::Entry::Occupied(_) => globals,
            hashbrown::hash_map::Entry::Vacant(entry) => {
                entry.insert(val);
                globals
            }
        });

        // Get the functions that are available in all the module to lower them. Should not be any
        // collision.

        let functions = modules.iter().flat_map(|(mname, m)| {
            m.functions(ASTVisibilityRule::any())
                .map(|(fname, f)| (mname.clone(), fname, f))
        });
        let functions = functions.collect();

        // Now we can create the lowerer with the specified LLVM module or by creating a new one.

        let m = m.unwrap_or_else(|| unsafe { LLVMModuleCreateWithNameInContext(name.as_ptr(), c) });
        let mut ctx = ValueContext::default();
        ctx.register_all(
            unsafe { vvs_runtime_types::vvll::LLVMRegisterExportedIntoModule(c, m) }
                .into_iter()
                .map(|(name, asttype, val, ty)| (name, Value::Function(asttype, ty, val))),
        );

        Lowerer { llvm: (m, c, b), functions, globals, ctx }
    }
}

impl<'a> Lowerer<'a> {
    /// Lower the content of the programs and modules, then returns the LLVM module containing
    /// everything...
    pub fn lower(self) -> LLVMModuleRef {
        let Self { llvm: llvm @ (m, c, b), functions, globals, mut ctx } = self;

        // Register all the globals in the context, they will be directly inlined in the lowering
        // process.

        ctx.register_all(
            globals
                .into_iter()
                .map(|((m, n), cnst)| (format!("{m}.{n}"), ConstantLowerer::new(c, cnst).lower())),
        );

        // Lower all the functions. First declare the things and register them with the mangled
        // name and the unmagnled name. As we collect the iterator we ensure that all the functions
        // are declared before we generate them. Doing so allows us to use recursion and we don't
        // need to find a good order to define-use the functions like in C. When we latter lower
        // the content of the functions we try to find the first error and return it.

        functions.into_iter().for_each(|(module, name, function)| {
            // Declare the thing
            let mangled = VVRTSymbol::new(VVRTSymbolType::Function, &module, &name)
                .expect("invalid function names")
                .mangle();
            let ast_function_type = function.function_type();
            let ty = TypeLowerer::new(c, &ast_function_type).lower();
            let value = unsafe { LLVMAddFunction(m, mangled.as_ptr(), ty) };

            // Register it. When we collect the iterator.
            let to_register = Value::Function(ast_function_type, ty, value);
            ctx.register(mangled.to_str().expect("invalid function names"), to_register.clone());
            ctx.register(format!("{module}.{name}"), to_register);

            // We will generate it latter
            FunctionLowerer::new(llvm, &ctx, value, function).lower();
        });

        // Cleanup and optimize the code.

        unsafe {
            LLVMDisposeBuilder(b);
            Self::verify_and_apply_optimizations(m, 3)
        }
    }

    /// Run the optimization passes on a module.
    unsafe fn verify_and_apply_optimizations(m: LLVMModuleRef, lvl: u8) -> LLVMModuleRef {
        let mut cmsg = std::ptr::null_mut();
        if 0 != LLVMVerifyModule(m, LLVMVerifierFailureAction::LLVMPrintMessageAction, &mut cmsg) {
            let msg = CStr::from_ptr(cmsg).to_string_lossy().to_string();
            LLVMDisposeMessage(cmsg);
            panic!("invalid module: {msg}");
        }

        let pbo = LLVMCreatePassBuilderOptions();
        let pm = LLVMCreatePassManager();
        let tm = LLVMCreateTargetMachine(
            LLVMGetTargetFromName(LLVMGetTarget(m)),
            LLVMGetDefaultTargetTriple(),
            LLVMGetHostCPUName(),
            LLVMGetHostCPUFeatures(),
            LLVMCodeGenOptLevel::LLVMCodeGenLevelAggressive,
            LLVMRelocMode::LLVMRelocPIC,
            LLVMCodeModel::LLVMCodeModelDefault,
        );

        let err = LLVMRunPasses(m, cstr!("default<O{}>", lvl), tm, pbo);
        if !err.is_null() {
            let msg = CStr::from_ptr(LLVMGetErrorMessage(err)).to_string_lossy().to_string();
            LLVMConsumeError(err);
            panic!("failed to run optimizer with level O{lvl} on module: {msg}")
        }

        LLVMDisposePassBuilderOptions(pbo);
        LLVMDisposePassManager(pm);
        LLVMDisposeTargetMachine(tm);
        m
    }
}
