use crate::{cstr /*lowerer*/};
use vvs_lang::ast::{ASTConst, ASTType, Typed};
use vvs_llvm::sys::*;

#[derive(Debug, Default, Clone, PartialEq, Eq)]
pub(crate) enum Value {
    /// Just the Nil/Null/None/nullptr value.
    #[default]
    Nil,

    /// A constant.
    Const(ASTType, LLVMValueRef),

    /// A pointer to the stack plus the underlying type.
    Alloca(ASTType, LLVMTypeRef, LLVMValueRef),

    /// A function with its function type.
    Function(ASTType, LLVMTypeRef, LLVMValueRef),
}

impl Value {
    /// Try to get the name of the value. Only [Value::Nil] have no name.
    pub fn as_str(&self) -> Option<&str> {
        use Value::*;
        match *self {
            Nil => None,
            Function(.., val) | Const(_, val) | Alloca(.., val) => unsafe {
                let mut len = 0;
                let str = LLVMGetValueName2(val, &mut len) as *const _;
                Some(std::str::from_utf8_unchecked(std::slice::from_raw_parts(str, len)))
            },
        }
    }

    pub fn llvm_type(&self) -> Option<LLVMTypeRef> {
        match self {
            Value::Const(..) | Value::Nil => None,
            Value::Alloca(_, t, _) | Value::Function(_, t, _) => Some(*t),
        }
    }

    pub fn llvm_value(&self) -> Option<LLVMValueRef> {
        match self {
            Value::Nil => None,
            Value::Const(_, v) | Value::Alloca(_, _, v) | Value::Function(_, _, v) => Some(*v),
        }
    }

    pub fn loaded(&self, c: LLVMContextRef, b: LLVMBuilderRef) -> LLVMValueRef {
        match self {
            Value::Nil => todo!(), // lowerer::ConstantLowerer::new(c, &ASTConst::Nil).lower().loaded(c, b),
            Value::Const(_, val) => *val,
            Value::Alloca(_, ty, ptr) => unsafe { LLVMBuildLoad2(b, *ty, *ptr, cstr!()) },
            Value::Function(_, _, func_ptr) => *func_ptr,
        }
    }

    /// Get the name of the dropper function, if any!
    pub fn get_dropper_name(&self) -> Option<&'static str> {
        match self.get_ast_type() {
            ty if ty.is_variant() => Some("VVRTVariant_drop"),
            ty if ty.is_table() => Some("VVRTTable_drop"),
            ty if ty.is_sequence() => Some("VVRTSeq_drop"),
            ASTType::Syllabe => Some("VVRTSyllabe_drop"),
            ASTType::String => Some("VVRTString_drop"),
            ASTType::Line => Some("VVRTLine_drop"),
            ASTType::Any => Some("VVRTAny_drop"),

            ASTType::Function(_, _) => todo!("handle functions and clozures..."),
            _ => None,
        }
    }

    /// Call a function with some values. Note that this function won't create memory manage points
    /// for the passed values (see types that are managed and the ones that are not.)
    pub fn call<const N: usize>(&self, b: LLVMBuilderRef, mut args: [LLVMValueRef; N]) -> LLVMValueRef {
        match self {
            Value::Function(_, func_ty, func_val) => unsafe {
                let args_ptr = args.as_mut_ptr();
                let args_len = args.len().try_into().expect("too much arguments");
                LLVMBuildCall2(b, *func_ty, *func_val, args_ptr, args_len, c"".as_ptr())
            },
            val => panic!("can't call a non-function value: {val:?}"),
        }
    }

    pub fn get_ast_type(&self) -> &ASTType {
        const NIL: ASTType = ASTType::Nil;
        match self {
            Value::Nil => &NIL,
            Value::Const(ty, _) | Value::Alloca(ty, _, _) | Value::Function(ty, _, _) => ty,
        }
    }
}

impl Typed for Value {
    fn get_type(&self, _: &vvs_lang::ast::ASTTypeContext) -> ASTType {
        self.get_ast_type().clone()
    }
}
