use crate::ast::*;
use anyhow::{anyhow, ensure, Context};

/// A constant expr.
#[derive(Debug, PartialEq, Clone)]
pub enum ASTConst {
    Nil,
    False,
    True,
    Tuple(Vec<ASTConst>),
    Color(ASTVariant<ASTConst>),
    String(ASTString),
    Integer(ASTInteger),
    Floating(ASTFloating),
    Table(ASTTable<ASTConst>),
}

/// Get the type for a const table, returns the more specialized type ([ASTType::UniformTable], etc)
fn get_const_table_type(table: &ASTTable<ASTConst>) -> ASTType {
    let integer_keys = table.keys().all(|key| key.parse::<i64>().is_ok());
    let mut uniform_values = table.values().map(|val| val.get_const_type()).collect::<Vec<_>>();
    let uniform_values = uniform_values
        .pop()
        .map(|ty| uniform_values.into_iter().all(|a| ty.eq(&a)).then_some(ty));
    match (integer_keys, uniform_values) {
        (true, Some(Some(ty))) => ASTType::Sequence(Box::new(ty)),
        _ => ASTType::Table(
            table
                .iter()
                .map(|(key, value)| (key.clone(), value.get_const_type()))
                .collect(),
        ),
    }
}

impl Typed for ASTConst {
    #[inline]
    fn get_type(&self, _: &ASTTypeContext) -> ASTType {
        self.get_const_type()
    }
}

impl std::fmt::Display for ASTConst {
    #[inline]
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ASTConst::Nil => f.write_str("nil"),
            ASTConst::False => f.write_str("false"),
            ASTConst::True => f.write_str("true"),
            ASTConst::Color(ASTVariant { variant, args }) => write!(f, "color#{variant}{args:?}"),
            ASTConst::String(str) => write!(f, "{str:?}"),
            ASTConst::Integer(int) => write!(f, "{int}"),
            ASTConst::Floating(flt) => write!(f, "{flt}"),
            ASTConst::Table(table) => write_ast_table(f, table),
            ASTConst::Tuple(tuple) => write_ast_tuple(f, tuple),
        }
    }
}

impl MaybeConstExpr for ASTConst {
    #[inline]
    fn is_const_expr(&self) -> bool {
        true
    }

    #[inline]
    fn eval_const_expr(&self) -> Option<ASTConst> {
        Some(self.clone())
    }
}

impl ASTConst {
    #[inline]
    pub fn get_const_type(&self) -> ASTType {
        match self {
            ASTConst::Nil => ASTType::Nil,
            ASTConst::True | ASTConst::False => ASTType::Boolean,
            ASTConst::Color(..) => ASTType::Color,
            ASTConst::String(_) => ASTType::String,
            ASTConst::Integer(_) => ASTType::Integer,
            ASTConst::Floating(_) => ASTType::Floating,
            ASTConst::Tuple(inner) => ASTType::Tuple(inner.iter().map(|expr| expr.get_const_type()).collect()),
            ASTConst::Table(content) => get_const_table_type(content),
        }
    }

    /// Try to create an [ASTConst] from a [toml::Value]. We verify that the value is valid, i.e.
    /// we don't have nested tables and such.
    ///
    /// NOTE: For now we can't construct colors and/or movements from toml values...
    pub fn from_toml_value(cache: &ASTStringCacheHandle, value: toml::Value) -> anyhow::Result<Self> {
        use toml::Value as TomlValue;

        fn table_from_toml_value(
            cache: &ASTStringCacheHandle,
            mut values: impl Iterator<Item = (ASTString, TomlValue)>,
        ) -> anyhow::Result<ASTConst> {
            let (lower, upper) = values.size_hint();
            let table = ASTTable::with_capacity(upper.unwrap_or(lower));
            let table = values.try_fold(table, |mut table, (key, value)| {
                let value = ASTConst::from_toml_value(cache, value)?;
                ensure!(!matches!(value, ASTConst::Table(_)), "forbiden nested tables");
                table
                    .insert(key, value)
                    .map(|_| Err(anyhow!("redefinition of a value in the table")))
                    .unwrap_or(Ok(table))
            })?;
            Ok(ASTConst::Table(table))
        }

        match value {
            TomlValue::Datetime(date) => Ok(ASTConst::String(cache.get(date.to_string()))),
            TomlValue::String(str) => Ok(ASTConst::String(cache.get(str))),

            TomlValue::Float(num) => Ok(ASTConst::Floating(num as f32)),
            TomlValue::Integer(num) => Ok(ASTConst::Integer(num.try_into().context("integer overflow")?)),
            TomlValue::Boolean(true) => Ok(ASTConst::True),
            TomlValue::Boolean(false) => Ok(ASTConst::False),

            TomlValue::Array(values) => {
                ensure!(values.len() < i32::MAX as usize, "too many elements in array");
                let values = values
                    .into_iter()
                    .enumerate()
                    .map(|(k, v)| (cache.get(k.to_string()), v));
                table_from_toml_value(cache, values)
            }
            TomlValue::Table(values) => {
                ensure!(values.len() < i32::MAX as usize, "too many elements in table");
                table_from_toml_value(cache, values.into_iter().map(|(k, v)| (cache.get(k), v)))
            }
        }
    }
}
