use crate::ast::*;
use std::ops::{Deref, DerefMut};

#[macro_export]
macro_rules! anon_expression {
    (const     $variant: ident $($args: tt)?) => { $crate::anon_expression!(Const (ASTConst::$variant $($args)?)) };
    (box const $variant: ident $($args: tt)?) => { Box::new($crate::anon_expression!(Const (ASTConst::$variant $($args)?))) };
    (box       $variant: ident $($args: tt)?) => { Box::new($crate::anon_expression!($variant $($args)?)) };

    ($variant: ident $($args: tt)?) => {
        ASTExpr {
            span: Default::default(),
            content: ASTExprVariant::$variant $($args)?,
        }
    };
}

#[macro_export]
macro_rules! expression {
    ($span: expr, $variant: ident $($args: tt)?) => {
        ASTExpr {
            span: $span.into(),
            content: ASTExprVariant::$variant $($args)?,
        }
    };
}

/// Binops, sorted by precedence.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASTBinop {
    /// Assoc: right
    Power,

    Mul,
    Div,
    Mod,
    Add,
    Sub,

    /// Assoc: right
    StrCat,

    CmpLE,
    CmpLT,
    CmpGE,
    CmpGT,
    CmpEQ,
    CmpNE,

    LogicAnd,
    LogicXor,
    LogicOr,

    BitAnd,
    BitXor,
    BitOr,
    BitShiftLeft,
    BitShiftRight,
}

/// Unops.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASTUnop {
    LogicNot,
    BitNot,
    Len,
    Neg,
}

/// Expressions. For the partial equality we skip the span field to be able to test efficiently the
/// parsing.
#[derive(Debug, Clone, Default)]
pub struct ASTExpr {
    pub content: ASTExprVariant,
    pub span: ASTSpan,
}

/// Expressions.
#[derive(Debug, PartialEq, Clone)]
pub enum ASTExprVariant {
    /// A table, like in Lua
    Table(ASTTable<ASTExpr>),

    /// A binary operation, like in most of languages.
    Binop(Box<ASTExpr>, ASTBinop, Box<ASTExpr>),

    /// A unary operation, like in most of languages.
    Unop(ASTUnop, Box<ASTExpr>),

    /// A call to a function.
    FuncCall(Box<ASTExpr>, Vec<ASTExpr>),

    /// A function where the first arguments are binded.
    FuncBind(Box<ASTExpr>, Vec<ASTExpr>),

    /// Invocation of a method. To call a function from a table, we don't use the
    /// [ASTExprVariant::MethodInvok] thing, but a [ASTExprVariant::FuncCall] where the first
    /// element is the dereferencing of the table. Here the name of the method must be known and be
    /// like: `{ty}::{method}`.
    MethodInvok(Box<ASTExpr>, ASTString, Vec<ASTExpr>),

    /// We access fields from an expression.
    Suffixed(Box<ASTExpr>, Vec<ASTField>),

    /// A non-constant expression color.
    Color(ASTVariant<ASTExpr>),

    /// Represents a movement for a line or syllabe. We don't have a constant variant for this
    /// representation because movements are very dynamic (it depends on the resolution of the
    /// video...) and thus doesn't make much sense to have a constant movement.
    ///
    /// NOTE: See if other people agree to that, there can be some arguments for having a constant
    ///       movement or an optional one.
    Movement(ASTVariant<ASTExpr>),

    /// A tuple, Lua don't have ones but because we won't be stack based we can't do the multiple
    /// return thing without a tuple...
    Tuple(Vec<ASTExpr>),

    /// A variable load.
    Var(ASTVar),

    /// A constant expression.
    Const(ASTConst),

    /// The default value for a type.
    Default(ASTType),
}

/// Fields indexes can be expressions or identifiers
#[derive(Debug, PartialEq, Clone)]
pub enum ASTField {
    /// We index by a string or an integer...
    Expr(ASTSpan, ASTExprVariant),

    /// A name field.
    Identifier(ASTSpan, ASTString),
}

impl PartialEq for ASTExpr {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        self.content == other.content
    }
}

/// Get the type for a table, returns the more specialized type ([ASTType::UniformTable], etc)
fn get_table_type<T: Typed>(table: &ASTTable<T>, ctx: &ASTTypeContext) -> ASTType {
    let integer_keys = table.keys().all(|key| key.parse::<i64>().is_ok());
    let mut uniform_values = table.values().map(|val| val.get_type(ctx)).collect::<Vec<_>>();
    let uniform_values = uniform_values
        .pop()
        .map(|ty| uniform_values.into_iter().all(|a| ty.eq(&a)).then_some(ty));
    match (integer_keys, uniform_values) {
        (true, Some(Some(ty))) => ASTType::Sequence(Box::new(ty)),
        _ => ASTType::Table(
            table
                .iter()
                .map(|(key, value)| (key.clone(), value.get_type(ctx)))
                .collect(),
        ),
    }
}

fn get_table_deref_type(ctx: &ASTTypeContext, table: ASTType, fields: &[ASTField]) -> ASTType {
    let (first, tail) = match fields {
        [] => return table,
        [first, tail @ ..] => (first, tail),
    };

    match table {
        ASTType::Tuple(_) => todo!(),

        ASTType::AnyTable => ASTType::Any,
        ASTType::Table(_) => todo!(),
        ASTType::UniformTable(_) => todo!(),

        ty @ ASTType::Syllabe | ty @ ASTType::Line => match first {
            ASTField::Expr(..) => ASTType::Nil,
            ASTField::Identifier(_, field) => {
                let ty = crate::ast::get_field_extensions(ctx, ty)
                    .find_map(|(key, ty)| key.eq(&field.as_ref()).then_some(ty.clone()))
                    .unwrap_or(ASTType::Nil);
                get_table_deref_type(ctx, ty, tail)
            }
        },

        _ => ASTType::Nil,
    }
}

impl Typed for ASTExprVariant {
    fn get_type(&self, ctx: &ASTTypeContext) -> ASTType {
        use crate::ast::{ASTBinop::*, ASTUnop::*};
        match self {
            ASTExprVariant::FuncBind(_, _) => todo!(),

            ASTExprVariant::MethodInvok(_, _, _) => todo!(),

            ASTExprVariant::FuncCall(_, _) => todo!(),

            ASTExprVariant::Default(ty) => ty.clone(),
            ASTExprVariant::Color(_) => ASTType::Color,
            ASTExprVariant::Movement(_) => ASTType::Movement,
            ASTExprVariant::Const(r#const) => r#const.get_type(ctx),

            ASTExprVariant::Var(var) => match var.get_specified_type() {
                Some(ty) => ty.clone(),
                None => ctx.get(var),
            },

            ASTExprVariant::Tuple(tuple) => ASTType::Tuple(tuple.iter().map(|expr| expr.get_type(ctx)).collect()),
            ASTExprVariant::Table(table) => get_table_type(table, ctx),
            ASTExprVariant::Suffixed(expr, fields) => get_table_deref_type(ctx, expr.get_type(ctx), fields),

            ASTExprVariant::Unop(op, inner) => {
                use ASTType::*;
                match (op, inner.get_type(ctx)) {
                    (BitNot | LogicNot, ty @ Integer | ty @ Boolean) => ty,
                    (Neg, ty @ Floating | ty @ Integer | ty @ Boolean) => ty,
                    (
                        Len,
                        String | Tuple(_) | Syllabe | Line | AnyTable | UniformTable(_) | AnySequence | Table(_)
                        | Sequence(_),
                    ) => Integer,
                    _ => Nil,
                }
            }

            ASTExprVariant::Binop(left, op, right) => {
                use ASTType::*;
                let (left, right) = (left.get_type(ctx), right.get_type(ctx));
                match op {
                    Power => match (left, right) {
                        (ty @ Integer | ty @ Floating, Integer) => ty,
                        _ => Nil,
                    },

                    Mul | Div | Add | Sub => match (left, right) {
                        (ty @ Integer, Integer) | (ty @ Floating, Floating) => ty,
                        _ => Nil,
                    },

                    Mod => match (left, right) {
                        (Integer, Integer) => Integer,
                        _ => Nil,
                    },

                    StrCat => match (left, right) {
                        (String, String) => String,
                        _ => Nil,
                    },

                    CmpLE | CmpLT | CmpGE | CmpGT => match (left, right) {
                        (Integer, Integer) | (Floating, Floating) => Boolean,
                        _ => Nil,
                    },

                    CmpEQ | CmpNE => match right.coercible_to(&left) {
                        true => Boolean,
                        false => Nil,
                    },

                    LogicAnd | LogicXor | LogicOr => match (left, right) {
                        (Boolean, Boolean) => Boolean,
                        _ => Nil,
                    },

                    BitAnd | BitXor | BitOr => match (left, right) {
                        (Integer, Integer) => Integer,
                        (Boolean, Boolean) => Boolean,
                        _ => Nil,
                    },

                    BitShiftLeft | BitShiftRight => match (left, right) {
                        (Integer, Integer) => Integer,
                        _ => Nil,
                    },
                }
            }
        }
    }
}

impl Deref for ASTExpr {
    type Target = ASTExprVariant;
    #[inline]
    fn deref(&self) -> &Self::Target {
        &self.content
    }
}

impl DerefMut for ASTExpr {
    #[inline]
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.content
    }
}

impl VariantName for ASTExpr {
    #[inline]
    fn variant_name(&self) -> &'static str {
        self.content.variant_name()
    }
}

impl VariantName for ASTExprVariant {
    #[inline]
    fn variant_name(&self) -> &'static str {
        match self {
            ASTExprVariant::Movement(_) => "movement",
            ASTExprVariant::Color(_) => "color",
            ASTExprVariant::Table(_) => "table",
            ASTExprVariant::Binop(_, _, _) => "binary operation",
            ASTExprVariant::Unop(_, _) => "unary operation",
            ASTExprVariant::FuncCall(_, _) => "function call",
            ASTExprVariant::FuncBind(_, _) => "function bind",
            ASTExprVariant::MethodInvok(_, _, _) => "method invokation",
            ASTExprVariant::Suffixed(_, _) => "prefixed expression",
            ASTExprVariant::Tuple(_) => "tuple",
            ASTExprVariant::Var(_) => "variable",
            ASTExprVariant::Const(_) => "constant",
            ASTExprVariant::Default(_) => "default",
        }
    }
}

impl ASTField {
    #[inline]
    pub fn span(&self) -> ASTSpan {
        match self {
            ASTField::Expr(span, _) | ASTField::Identifier(span, _) => *span,
        }
    }
}

impl MaybeConstExpr for ASTExprVariant {
    fn is_const_expr(&self) -> bool {
        use ASTExprVariant::*;
        match self {
            // We don't do constant method invokation for now.
            FuncCall(_, _) | MethodInvok(_, _, _) | FuncBind(_, _) => false,

            // Depends on the content.
            Binop(l, _, r) => l.is_const_expr() && r.is_const_expr(),
            Unop(_, inner) => inner.is_const_expr(),
            Tuple(inner) => inner.iter().all(|expr| expr.is_const_expr()),
            Color(variant) | Movement(variant) => variant.is_const_expr(),
            Suffixed(table, suffixes) => table.is_const_expr() && suffixes.iter().all(|field| field.is_const_expr()),
            Default(ty) => ty.is_const_constructible(),

            // Well, it depend, for now we will say that it's not compile time evaluable.
            Table(_) => false,
            Var(_) => false,

            // Obviously
            Const(_) => true,
        }
    }

    fn eval_const_expr(&self) -> Option<ASTConst> {
        use ASTExprVariant::*;
        match self {
            Const(r#const) => Some(r#const.clone()),

            Binop(_, _, _) => todo!(),
            Unop(_, _) => todo!(),

            Default(ty) if ty.is_const_constructible() => Some(match ty {
                ASTType::Nil | ASTType::Any => ASTConst::Nil,
                ASTType::Integer => ASTConst::Integer(0),
                ASTType::Floating => ASTConst::Floating(0.),
                ASTType::Boolean => ASTConst::False,
                _ => unreachable!(),
            }),

            Color(ASTVariant { variant, args }) => Some(ASTConst::Color(ASTVariant {
                variant: variant.clone(),
                args: args
                    .iter()
                    .map(|expr| expr.eval_const_expr().ok_or(()))
                    .collect::<Result<Vec<_>, _>>()
                    .ok()?,
            })),

            Tuple(inner) => Some(ASTConst::Tuple(
                inner
                    .iter()
                    .map(|expr| expr.eval_const_expr().ok_or(()))
                    .collect::<Result<Vec<_>, _>>()
                    .ok()?,
            )),

            Suffixed(_, _)
            | FuncCall(_, _)
            | FuncBind(_, _)
            | MethodInvok(_, _, _)
            | Table(_)
            | Movement(_)
            | Default(_)
            | Var(_) => None,
        }
    }
}

impl Typed for ASTField {
    #[inline]
    fn get_type(&self, ctx: &ASTTypeContext) -> ASTType {
        match self {
            ASTField::Expr(_, expr) => expr.get_type(ctx),
            ASTField::Identifier(sp, name) => {
                let var = ASTVar::new(ctx.strings()).span(sp).name(name.clone()).build();
                ctx.get(&var)
            }
        }
    }
}

impl MaybeConstExpr for ASTField {
    #[inline]
    fn is_const_expr(&self) -> bool {
        match self {
            ASTField::Expr(_, expr) => expr.is_const_expr(),
            ASTField::Identifier(_, _) => false,
        }
    }

    #[inline]
    fn eval_const_expr(&self) -> Option<ASTConst> {
        match self {
            ASTField::Identifier(..) => None,
            ASTField::Expr(_, expr) => expr.eval_const_expr(),
        }
    }
}

impl Default for ASTExprVariant {
    fn default() -> Self {
        ASTExprVariant::Const(ASTConst::Nil)
    }
}
