use crate::ast::*;

#[derive(Debug, Default, PartialEq, Eq, Clone)]
pub enum ASTIdentifier {
    /// We ignore the value, like in ocaml and rust, this is the '_'.
    #[default]
    Joker,

    /// A normal identifier, one that can be used by the user.
    String(ASTString),

    /// If an identifier begins by '_' and is followed other things then it is reserved by the
    /// parser, the generator, and other things. Those identifiers can't be used by a user and they
    /// should never see one.
    Reserved(ASTString),
}

impl std::fmt::Display for ASTIdentifier {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_ref())
    }
}

impl AsRef<str> for ASTIdentifier {
    fn as_ref(&self) -> &str {
        use ASTIdentifier::*;
        match self {
            Joker => "_",
            String(str) | Reserved(str) => str.as_ref(),
        }
    }
}

impl ASTIdentifier {
    /// Create a new identifier. The passed string will be trimed. If the string is empty or
    /// contains non ascii-alphanumeric-underscored characters then an error will be raised.
    pub fn new(cache: ASTStringCacheHandle, identifier: impl AsRef<str>) -> Result<Self, String> {
        let identifier = identifier.as_ref().trim();
        match identifier
            .char_indices()
            .find(|(_, c)| '_'.ne(c) && !c.is_ascii_alphanumeric())
        {
            Some((idx, char)) => Err(format!("found invalid character '{char}' in identifier at index {idx}")),
            None => {
                if identifier.is_empty() {
                    Err("an identifier can't be empty".to_string())
                } else if identifier.starts_with('_') && identifier.len() == 1 {
                    Ok(Self::Joker)
                } else if identifier.starts_with('_') {
                    Ok(Self::Reserved(cache.get(identifier.trim_start_matches('_'))))
                } else {
                    Ok(Self::Reserved(cache.get(identifier)))
                }
            }
        }
    }
}
