use crate::ast::{ASTModule, ASTModulePtr, ASTString};
use hashbrown::HashMap;
use std::{
    cell::RefCell,
    path::{Path, PathBuf},
    rc::Rc,
};

/// Structure used to resolve module imports.
///
/// By default, includes the current working dir.
#[derive(Debug)]
pub struct ImportResolver {
    path: Vec<PathBuf>,
    cache: Rc<RefCell<HashMap<ASTString, ASTModulePtr>>>,
}

pub(crate) enum ResolveResult {
    Path(PathBuf),
    Module(ASTModulePtr),
}

impl FromIterator<PathBuf> for ImportResolver {
    #[inline]
    fn from_iter<T: IntoIterator<Item = PathBuf>>(iter: T) -> Self {
        Self { path: iter.into_iter().collect(), cache: Default::default() }
    }
}

impl Default for ImportResolver {
    /// Get the default import resolver, always includes the current working dir...
    #[inline]
    fn default() -> Self {
        Self { path: vec![std::env::current_dir().expect("failed to get the CWD")], cache: Default::default() }
    }
}

impl ImportResolver {
    /// Create a new empty resolver. Note that it's different from the default resolver that uses
    /// the current working directory (CWD, PWD, etc...)
    #[inline]
    pub fn empty() -> Self {
        Self { path: vec![], cache: Default::default() }
    }

    pub(crate) fn resolve(&self, module: impl AsRef<str>) -> Option<ResolveResult> {
        let find = |prefix| Self::get_file_path(prefix, module.as_ref()).map(ResolveResult::Path);
        self.cache
            .borrow()
            .get(module.as_ref())
            .cloned()
            .map(ResolveResult::Module)
            .or_else(|| self.path.iter().find_map(find))
    }

    #[inline]
    pub(crate) fn cache_module(&self, name: ASTString, module: ASTModulePtr) {
        if let Some(module) = self.cache.borrow_mut().insert(name, module) {
            log::error!("re-cache module `{}`", ASTModule::get(&module).name())
        }
    }

    fn get_file_path(prefix: impl AsRef<Path>, module: impl AsRef<str>) -> Option<PathBuf> {
        let mut file = prefix.as_ref().join(module.as_ref());
        file.set_extension("vvl");
        file.canonicalize()
            .map_err(|err| match err.kind() {
                std::io::ErrorKind::NotFound => (),
                err => log::error!(
                    "failed to lookup for `{}` in `{}`: {err}",
                    module.as_ref(),
                    prefix.as_ref().to_string_lossy()
                ),
            })
            .ok()
    }
}
