use crate::ast::*;
use std::ops::{Deref, DerefMut};

#[macro_export]
macro_rules! anon_instruction {
    ($variant: ident $($args: tt)?) => {
        ASTInstr {
            span: Default::default(),
            content: ASTInstrVariant::$variant $($args)?,
        }
    };
}

#[macro_export]
macro_rules! instruction {
    ($span: expr, $variant: ident $($args: tt)?) => {
        ASTInstr {
            span: $span.into(),
            content: ASTInstrVariant::$variant $($args)?,
        }
    };
}

/// Instructions.
#[derive(Debug, Clone)]
pub struct ASTInstr {
    pub content: ASTInstrVariant,
    pub span: ASTSpan,
}

/// Instructions.
#[derive(Debug, PartialEq, Clone)]
pub enum ASTInstrVariant {
    /// Declare variables with something inside.
    Decl(Vec<ASTVar>, Vec<ASTExpr>),

    /// Assign into variables.
    Assign(Vec<ASTExpr>, Vec<ASTExpr>),

    /// A function call.
    FuncCall(Box<ASTExpr>, Vec<ASTExpr>),

    /// A method invokation.
    MethodInvok(Box<ASTExpr>, ASTString, Vec<ASTExpr>),

    /// Begin a block.
    Block(Vec<ASTInstr>),

    /// A WhileDo instruction.
    WhileDo(ASTCondition, Box<ASTInstr>),

    /// Conditionals, contract the elseid blocks.
    Cond { if_blocks: Vec<(ASTCondition, ASTInstr)>, else_block: Option<Box<ASTInstr>> },

    /// For loop, the classic one:
    /// ```vvs
    /// for elem = 1, 3, 1 do print(elem) end
    /// ```
    ForLoop { var: ASTVar, lower: ASTExpr, upper: ASTExpr, step: Option<ASTInteger>, body: Box<ASTInstr> },

    /// For loop with an iterable expression (table):
    /// ```vvs
    /// for elem in { 1, 2, 3 } do print(elem) end
    /// for elem in ( 1, 2, 3 ) do print(elem) end
    /// for elem in my_table    do print(elem) end
    /// ```
    ForInto { var: ASTVar, list: ASTExpr, body: Box<ASTInstr> },

    /// Just a noop.
    Noop,

    /// Final thing: break from block.
    Break,

    /// Final thing: continue to next iteration.
    Continue,

    /// Final thing: return something.
    Return(ASTExpr),
}

impl PartialEq for ASTInstr {
    fn eq(&self, other: &Self) -> bool {
        self.content == other.content
    }
}

fn get_instruction_slice_type(ctx: &ASTTypeContext, block: &[ASTInstr]) -> ASTType {
    let Some((last, instructions)) = block.split_last() else {
        return ASTType::Nil;
    };
    let coerce_type = |from: ASTType, var: &ASTVar| {
        let (span, ty, name) = (var.span(), var.get_specified_type().unwrap_or(&from), var.name());
        let strings = ctx.strings();
        let var = ASTVar::new(strings).span(span).name(name.clone());
        if from.coercible_to(ty) {
            var.with_type(from).build()
        } else {
            log::error!(target: "cc", "{span}; try to assign value of type `{from}` into `{name}: {ty}`");
            var.with_type(ASTType::Nil).build()
        }
    };
    let fold_declaration = |mut ctx, (var, from): (&ASTVar, &ASTExpr)| {
        let var = coerce_type(from.get_type(&ctx), var);
        let (_, var_ty) = (var.span(), var.get_type(&ctx));
        ctx.declare(var, var_ty);
        ctx
    };
    let fold_on_declaration = |ctx, instruction: &ASTInstr| match &instruction.content {
        ASTInstrVariant::Decl(vars, froms) => vars.iter().zip(froms.iter()).fold(ctx, fold_declaration),
        _ => ctx,
    };
    last.get_type(&instructions.iter().fold(ctx.for_scope(), fold_on_declaration))
}

impl Typed for ASTInstrVariant {
    fn get_type(&self, ctx: &ASTTypeContext) -> ASTType {
        match self {
            ASTInstrVariant::Decl(_, _)
            | ASTInstrVariant::Noop
            | ASTInstrVariant::Assign(_, _)
            | ASTInstrVariant::Break
            | ASTInstrVariant::Continue => ASTType::Nil,

            ASTInstrVariant::WhileDo(_, _) | ASTInstrVariant::ForLoop { .. } | ASTInstrVariant::ForInto { .. } => {
                ASTType::Nil
            }

            ASTInstrVariant::Return(expr) => expr.get_type(ctx),
            ASTInstrVariant::Block(block) => get_instruction_slice_type(ctx, block),
            ASTInstrVariant::Cond { if_blocks, else_block } => {
                let mut types = else_block
                    .as_ref()
                    .map(|i| i.get_type(ctx))
                    .into_iter()
                    .chain(if_blocks.iter().map(|(_, i)| i.get_type(ctx)));
                types
                    .next()
                    .and_then(|ty| types.all(|other| ty.eq(&other)).then_some(ty))
                    .unwrap_or_default()
            }

            ASTInstrVariant::MethodInvok(_, _, _) | ASTInstrVariant::FuncCall(_, _) => todo!("implement clozures"),
        }
    }
}

impl Deref for ASTInstr {
    type Target = ASTInstrVariant;

    fn deref(&self) -> &Self::Target {
        &self.content
    }
}

impl DerefMut for ASTInstr {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.content
    }
}

impl VariantName for ASTInstr {
    fn variant_name(&self) -> &'static str {
        self.content.variant_name()
    }
}

impl VariantName for ASTInstrVariant {
    fn variant_name(&self) -> &'static str {
        match self {
            ASTInstrVariant::Decl(_, _) => "variable declaration",
            ASTInstrVariant::Assign(_, _) => "variable assignation",
            ASTInstrVariant::FuncCall(_, _) => "function call",
            ASTInstrVariant::MethodInvok(_, _, _) => "method invokation",
            ASTInstrVariant::Block(_) => "instruction block",
            ASTInstrVariant::WhileDo(_, _) => "while loop",
            ASTInstrVariant::Cond { .. } => "conditionals",
            ASTInstrVariant::ForLoop { step: Some(_), .. } => "for loop with step",
            ASTInstrVariant::ForLoop { step: None, .. } => "for loop",
            ASTInstrVariant::ForInto { .. } => "for into loop",
            ASTInstrVariant::Noop => "noop",
            ASTInstrVariant::Break => "break statement",
            ASTInstrVariant::Continue => "continue statement",
            ASTInstrVariant::Return(_) => "return statement",
        }
    }
}
