//! Module used to store all the definition of the AST used for parsing VVS/VVL files.

mod constant;
mod expression;
mod extension;
mod function;
mod identifier;
mod import;
mod instruction;
mod maybe_const_expr;
mod module;
mod options;
mod pattern;
mod program;
mod span;
mod string;
mod type_context;
mod typed;
mod types;
mod variable;
mod variant;
mod visibility;

use hashbrown::HashMap;
use std::rc::Rc;

pub use self::{
    constant::*, expression::*, extension::*, function::*, identifier::*, import::*, instruction::*,
    maybe_const_expr::*, module::*, options::*, pattern::*, program::*, span::*, string::*, type_context::*, typed::*,
    types::*, variable::*, variant::*, visibility::*,
};

pub type ASTString = Rc<str>;
pub type ASTFloating = f32;
pub type ASTInteger = i32;
pub type ASTTable<T> = HashMap<ASTString, T>;

pub fn write_ast_table<T: std::fmt::Display>(f: &mut std::fmt::Formatter<'_>, table: &ASTTable<T>) -> std::fmt::Result {
    f.write_str("{")?;
    if let Some((key, value)) = table.iter().next() {
        write!(f, " {key} = {value}")?;
    }
    for (key, value) in table.iter().skip(1) {
        write!(f, ", {key} = {value}")?;
    }
    f.write_str(" }")
}

pub fn write_ast_tuple<T: std::fmt::Display>(f: &mut std::fmt::Formatter<'_>, tuple: &[T]) -> std::fmt::Result {
    f.write_str("(")?;
    if let Some(value) = tuple.iter().next() {
        write!(f, " {value}")?;
    }
    for value in tuple.iter().skip(1) {
        write!(f, ", {value}")?;
    }
    f.write_str(" )")
}

/// Trait that allows to get a variant name out of an enum or a struct containing an enum.
pub(crate) trait VariantName {
    /// Get the name out of the value.
    fn variant_name(&self) -> &'static str;
}
