use crate::ast::*;
use hashbrown::{HashMap, HashSet};
use std::{
    cell::UnsafeCell,
    rc::{Rc, Weak},
};

/// A VVL module of a Vivy Script
#[derive(Debug, Clone)]
pub struct ASTModule {
    /// The name of the module, should be a valid identifier.
    name: ASTString,

    /// Checks if we want to put it into the type context or not...
    variants: HashMap<ASTString, ASTVariantRules>,

    functions: HashMap<ASTString, (ASTVisibility, ASTFunction)>,
    jobs: HashMap<ASTString, (ASTVisibility, ASTFunction)>,
    consts: HashMap<ASTString, (ASTVisibility, ASTVar, ASTConst)>,
    options: HashMap<ASTString, (ASTVar, ASTConst)>,
    imports: HashMap<ASTString, (ASTSpan, ASTModulePtr)>,

    /// Contains all the symbols available in the module, to be sure that options, constants,
    /// functions and jobs doesn't enter in collision.
    symbols: HashSet<ASTString>,

    /// Caching for strings, identified by their hash and reuse the same memory location to reduce
    /// memory footprint. Use the .finish function from the hasher to get the key.
    strings: ASTStringCacheHandle,

    /// Type context intern to the module. This is self-referencing. Under the hood we use a weak
    /// pointer and a lazy initialize thingy for the type context.
    tyctx: ASTTypeContext,
}

pub type ASTModulePtr = Rc<UnsafeCell<ASTModule>>;
pub type ASTModuleWeakPtr = Weak<UnsafeCell<ASTModule>>;

impl ASTModule {
    /// Create a new empty module.
    #[inline]
    pub fn new(name: ASTString, strings: ASTStringCacheHandle) -> ASTModulePtr {
        Rc::new_cyclic(|ptr| {
            let colors_rules = ASTVariantRules::new_colors(strings.clone());
            let movement_rules = ASTVariantRules::new_movements(strings.clone());
            UnsafeCell::new(Self {
                name,
                strings,
                tyctx: ASTTypeContext::new(ptr.clone()),
                variants: HashMap::from_iter([
                    (colors_rules.name(), colors_rules),
                    (movement_rules.name(), movement_rules),
                ]),
                symbols: Default::default(),
                functions: Default::default(),
                jobs: Default::default(),
                consts: Default::default(),
                options: Default::default(),
                imports: Default::default(),
            })
        })
    }

    /// Get all the variant rules.
    #[inline]
    pub fn all_variant_rules(&self) -> &HashMap<ASTString, ASTVariantRules> {
        &self.variants
    }

    /// Get a set of rule for a specified variant if it exists in a mutable way. We only expose
    /// this function to this crate because it's the only place where we want to be able to mutate
    /// the module.
    #[inline]
    pub(crate) fn variant_rules_mut(&mut self, rule: impl AsRef<str>) -> Option<&mut ASTVariantRules> {
        self.variants.get_mut(rule.as_ref())
    }

    /// Get the type context in a constant way.
    #[inline]
    pub fn tyctx(&self) -> &ASTTypeContext {
        &self.tyctx
    }

    /// Get the type context in a mutable way. Only possible if we have a mutable reference to the
    /// module anyway. If there are multiple strong references to the module it won't be correct...
    #[inline]
    pub fn tyctx_mut(&mut self) -> &mut ASTTypeContext {
        &mut self.tyctx
    }

    /// Get the name of the module.
    #[inline]
    pub fn name(&self) -> ASTString {
        self.name.clone()
    }

    /// Get a specific option if it exists.
    #[inline]
    pub fn option(&self, option: impl AsRef<str>) -> Option<&(ASTVar, ASTConst)> {
        self.options.get(option.as_ref())
    }

    /// Get a job by its name, returns [None] if the job was not defined.
    #[inline]
    pub fn job(&self, name: impl AsRef<str>) -> Option<&ASTFunction> {
        self.jobs.get(name.as_ref()).map(|(_, job)| job)
    }

    /// Get a function by its name, returns [None] if the function was not defined.
    #[inline]
    pub fn function(&self, name: impl AsRef<str>) -> Option<&ASTFunction> {
        self.functions.get(name.as_ref()).map(|(_, function)| function)
    }

    /// Get all the options... All the options are public by design so no visibility rule is
    /// necessary to query them.
    #[inline]
    pub fn options(&self) -> impl Iterator<Item = (&ASTVar, &ASTConst)> {
        self.options.values().map(|(a, b): &(_, _)| (a, b))
    }

    /// Get all the constants with the specified visibility rule.
    #[inline]
    pub fn consts(&self, rule: ASTVisibilityRule) -> impl Iterator<Item = (&ASTVar, &ASTConst)> {
        self.consts
            .values()
            .filter_map(move |(v, var, value)| (rule.allows(*v)).then_some((var, value)))
    }

    /// Get all the functions from the module with the specified visibility rule.
    #[inline]
    pub fn functions(&self, rule: ASTVisibilityRule) -> impl Iterator<Item = (ASTString, &ASTFunction)> {
        self.functions
            .values()
            .filter_map(move |(v, f)| rule.allows(*v).then_some((f.name.clone(), f)))
    }

    /// Get all the jobs from the module with the specified visibility rule.
    #[inline]
    pub fn jobs(&self, rule: ASTVisibilityRule) -> impl Iterator<Item = (ASTString, &ASTFunction)> {
        self.jobs
            .values()
            .filter_map(move |(v, f)| rule.allows(*v).then_some((f.name.clone(), f)))
    }

    /// Get all the callable items from the module with the specified visibility rule. This
    /// includes jobs and functions.
    pub(crate) fn callables(&self, rule: ASTVisibilityRule) -> impl Iterator<Item = (ASTString, &ASTFunction)> {
        self.functions(rule).chain(self.jobs(rule))
    }

    /// Get all the jobs from a module in a mutable way, with the specified visibility rule.
    pub(crate) fn jobs_mut(&mut self, rule: ASTVisibilityRule) -> impl Iterator<Item = (ASTString, &mut ASTFunction)> {
        self.jobs
            .values_mut()
            .filter_map(move |(v, f)| rule.allows(*v).then_some((f.name.clone(), f)))
    }

    /// Get all the fonctions from a module in a mutable way, with the specified visibility rule.
    pub(crate) fn functions_mut(
        &mut self,
        rule: ASTVisibilityRule,
    ) -> impl Iterator<Item = (ASTString, &mut ASTFunction)> {
        self.functions
            .values_mut()
            .filter_map(move |(v, f)| rule.allows(*v).then_some((f.name.clone(), f)))
    }

    /// Get the imported modules, required by the current module.
    #[inline]
    pub fn imports(&self) -> impl Iterator<Item = &ASTModule> + '_ {
        self.imports.iter().map(|(_, (_, module))| ASTModule::get(module))
    }

    /// Get the import location of a module, for debug purpose.
    #[inline]
    pub fn get_import_location(&self, import: &ASTString) -> Option<ASTSpan> {
        self.imports
            .iter()
            .find_map(|(module, (span, _))| import.eq(module).then_some(*span))
    }

    /// Get a handle to the string cache.
    #[inline]
    pub fn strings(&self) -> &ASTStringCacheHandle {
        &self.strings
    }

    /// Get a const reference to the module. As long as you don't use the [ASTModule::get_mut]
    /// function, this one should be Ok.
    #[inline]
    pub(crate) fn get(this: &ASTModulePtr) -> &ASTModule {
        unsafe { &*this.as_ref().get() }
    }

    /// Get a mutable reference to the module... Unsafe!
    #[inline]
    pub(crate) fn get_mut(this: &ASTModulePtr) -> *mut ASTModule {
        this.as_ref().get()
    }
}
