use crate::ast::*;
use anyhow::{bail, Context};
use hashbrown::HashMap;

/// Options for a program, represent the thing that is red from a file.
pub struct OptionTable {
    modules: HashMap<ASTString, ASTTable<ASTConst>>,
}

impl OptionTable {
    pub fn new(cache: &ASTStringCacheHandle, values: toml::Table) -> anyhow::Result<Self> {
        let modules = HashMap::with_capacity(values.len());
        let modules = values.into_iter().try_fold(modules, |mut modules, (name, values)| {
            use hashbrown::hash_map::EntryRef::*;
            let toml::Value::Table(values) = values else {
                bail!("expected a table for `{name}`")
            };
            let table = modules
                .entry(cache.get(&name))
                .or_insert(ASTTable::with_capacity(values.len()));
            values
                .into_iter()
                .try_for_each(|(option, value)| match table.entry_ref(option.as_str()) {
                    Occupied(_) => bail!("redefinition of option `{option}` in module `{name}`"),
                    Vacant(entry) => {
                        entry.insert(ASTConst::from_toml_value(cache, value)?);
                        Ok(())
                    }
                })?;
            if table.is_empty() {
                modules.remove(name.as_str());
            }
            Ok(modules)
        });
        Ok(Self { modules: modules.context("invalid toml file")? })
    }
}

impl Iterator for OptionTable {
    type Item = ((ASTString, ASTString), ASTConst);

    fn next(&mut self) -> Option<Self::Item> {
        let Self { modules } = self;
        let module = modules.keys().next()?.clone();

        let values = match modules.get_mut(&module) {
            Some(values) => values,
            None => {
                modules.remove(&module);
                return None;
            }
        };

        let (option, value) = match values.keys().next().cloned() {
            Some(option) => values.remove_entry(&option)?,
            None => {
                modules.remove(&module);
                return None;
            }
        };

        Some(((module, option), value))
    }
}
