use crate::ast::*;
use hashbrown::HashMap;

/// The first element of the tuple is the destination variable, the second is the tuple to describe
/// the job picked from a module, the last is the list of variables to use as input for this job.
pub type ProgramOperation = (ASTSpan, ASTString, (ASTString, ASTString), Option<Vec<ASTConst>>, Vec<ASTString>);

/// The main VVS file/program.
#[derive(Debug)]
pub struct ASTProgram {
    /// The list of all the modules used in the program, with their import location.
    modules: HashMap<ASTString, (ASTSpan, ASTModulePtr)>,

    /// The setted options, from the VVS file.
    setted: Vec<((ASTString, ASTString), ASTConst)>,

    /// The operations to do in order.
    operations: Vec<ProgramOperation>,

    /// What the program writes.
    writes: (ASTSpan, Vec<ASTString>),

    /// Initial variable.
    initial_var: ASTString,

    /// Caching for strings, identified by theyr hash and reuse the same memory location to reduce
    /// memory footprint. Use the .finish function from the hasher to get the key.
    strings: ASTStringCacheHandle,
}

impl ASTProgram {
    pub fn new(strings: ASTStringCacheHandle) -> Self {
        Self {
            initial_var: strings.get("INIT"),
            setted: Default::default(),
            writes: Default::default(),
            modules: Default::default(),
            operations: Default::default(),
            strings,
        }
    }

    /// Iterate over all the modules in the program.
    #[inline]
    pub fn modules(&self) -> impl Iterator<Item = &ASTModule> + '_ {
        self.modules.values().map(|(_, module)| ASTModule::get(module))
    }

    /// Get a handle to the string cache.
    #[inline]
    pub fn strings(&self) -> &ASTStringCacheHandle {
        &self.strings
    }

    /// Get a const reference to a module.
    pub fn module(&self, name: impl AsRef<str>) -> Option<&ASTModule> {
        self.modules
            .get(name.as_ref())
            .map(|(_, module)| ASTModule::get(module))
    }

    /// Get the pointer to a module. Returns [None] if the module was not present in the program.
    ///
    /// # Safety
    /// With this function we create the possibility to make multiple mutable references to the
    /// same module down the line, so we mark it as an unsafe operation for now...
    pub(crate) unsafe fn module_ptr(&self, name: impl AsRef<str>) -> Option<ASTModulePtr> {
        self.modules.get(name.as_ref()).map(|(_, ptr)| ptr.clone())
    }

    /// Get all the available options for this program, with their default value.
    pub fn available_options(&self) -> impl Iterator<Item = (ASTString, &ASTVar, &ASTConst)> {
        self.modules()
            .flat_map(|module| module.options().map(|(var, def)| (module.name(), var, def)))
    }

    /// Get all the setted options from the program.
    #[inline]
    pub fn setted_options(&self) -> &[((ASTString, ASTString), ASTConst)] {
        self.setted.as_ref()
    }

    /// Get the value for an option. For the resolution: we first take the value from the setted
    /// list before the default value. Note that if we set an option that was never declared we
    /// don't return it here, it should have raised a warning and should not be used in the code
    /// anyway.
    pub fn option(&self, module: &ASTString, name: &ASTString) -> Option<ASTConst> {
        let default = self
            .module(module)?
            .options()
            .find_map(|(var, val)| var.name().eq(name).then_some(val))?
            .clone();
        let setted = self
            .setted
            .iter()
            .find_map(|((mm, var), val)| (*module == *mm && *name == *var).then_some(val));
        Some(setted.cloned().unwrap_or(default))
    }

    /// Get the operations and the variables to write. Note that no check is done here.
    pub fn as_operations(&self) -> (ASTString, Vec<ProgramOperation>, Vec<ASTString>) {
        (self.initial_var.clone(), self.operations.clone(), self.writes.1.clone())
    }

    #[inline]
    pub(crate) fn writes_span(&self) -> ASTSpan {
        self.writes.0
    }

    /// Set an option in the [Program], if the option was already set or was not declared in a
    /// [Module] we raise a warning.
    pub fn set_option(mut self, module: ASTString, name: ASTString, value: ASTConst) -> Self {
        if self
            .module(&module)
            .map(|module| module.options().any(|(var, _)| var.name().eq(&name)))
            .unwrap_or_default()
        {
            log::warn!(target: "cc", ";set option {module}.{name} which was not declared");
        } else if self.setted.iter().any(|((mm, nn), _)| *mm == module && name == *nn) {
            log::warn!(target: "cc", ";re-set option {module}.{name}");
        }
        self.setted.push(((module, name), value));
        self
    }

    /// Import a module into the program.
    #[inline]
    pub(crate) fn import_module(mut self, name: ASTString, span: ASTSpan, module: ASTModulePtr) -> Self {
        if let Some((_, module)) = self.modules.insert(name, (span, module)) {
            log::warn!(target: "cc", ";re-import of module `{}`", ASTModule::get(&module).name())
        }
        self
    }

    /// Use options from a toml file to set things in the program and simplify a bit said
    /// program...
    pub fn with_options(self, _options: OptionTable) -> Self {
        self
    }
}
