use crate::ast::*;
use core::cell::OnceCell;

/// Match any item, private or public.
const RULE_ANY: ASTVisibilityRule = ASTVisibilityRule::AtLeast(ASTVisibility::Private);
const RULE_PUB: ASTVisibilityRule = ASTVisibilityRule::Only(ASTVisibility::Public);

/// The ast context for storing types. Any use of the module must be delayed because when we create
/// the type context for a module, we do it with a cyclic thingy, so the [Rc] is not already
/// constructed (no strong reference), so all we have is a [std::rc::Weak].
#[derive(Debug, Clone)]
pub struct ASTTypeContext {
    module: ASTModuleWeakPtr,
    content: OnceCell<ASTTypeContextContent>,
}

/// The content of the [ASTTypeContext] must be lazyly initialized because of cyclic references
/// with the module and it's type context.
type ASTTypeContextContent = (
    HashMap<ASTString, ASTType>,         // Variables
    HashMap<ASTString, ASTVariantRules>, // Variants
    ASTStringCacheHandle,                // String cache
    Vec<ASTString>,                      // Name of imported modules
);

impl ASTTypeContext {
    fn init_content(&self) -> ASTTypeContextContent {
        let ptr = self.module.upgrade().expect("no more strong references to module...");
        let module = ASTModule::get(&ptr);

        // We need to import things from the current module and from the imported modules. But for
        // imported modules we must prepend the constant/option/callable name by the module name
        // with a dot, to call those things like in lua.
        let modules = [(None, module)]
            .into_iter()
            .chain(module.imports().map(|module| (Some(module.name()), module)));

        let scope = modules.flat_map(|(mname, module)| {
            let rule = if mname.is_some() { RULE_PUB } else { RULE_ANY };
            let (mname1, mname2, mname3) = (mname.clone(), mname.clone(), mname);

            let options = module.options().map(move |(var, _)| {
                let name = mname1
                    .as_ref()
                    .map(|n| module.strings().get(format!("{n}.{}", var.name())))
                    .unwrap_or_else(|| var.name());
                (name, var.get_specified_type_or_nil())
            });

            let consts = module.consts(rule).map(move |(var, _)| {
                let name = mname2
                    .as_ref()
                    .map(|n| module.strings().get(format!("{n}.{}", var.name())))
                    .unwrap_or_else(|| var.name());
                (name, var.get_specified_type_or_nil())
            });

            let funcs = module.callables(rule).map(move |(name, func)| {
                let name = mname3
                    .as_ref()
                    .map(|n| module.strings().get(format!("{n}.{name}")))
                    .unwrap_or(name);

                let arguments = func.arguments.iter().map(|arg| arg.get_specified_type_or_nil());
                let func_type = ASTType::Function(arguments.collect(), Box::new(func.returns.clone()));

                (name, func_type)
            });

            options.chain(consts).chain(funcs)
        });

        let strings = module.strings().clone();
        let variants = module.all_variant_rules().clone();
        let imports = module.imports().map(|module| module.name()).collect();
        (scope.collect(), variants, strings, imports)
    }

    fn content(&self) -> &ASTTypeContextContent {
        self.content.get_or_init(|| self.init_content())
    }

    fn content_mut(&mut self) -> &mut ASTTypeContextContent {
        self.content();
        self.content.get_mut().expect("should be initialized")
    }

    #[inline]
    pub fn new(module: ASTModuleWeakPtr) -> Self {
        Self { module, content: OnceCell::new() }
    }

    /// Try to map rules to a pattern. Returns the matching [Some] variables with their destination
    /// types or [None] if the matching is not possible to do. Because all rules must be kown at
    /// compile time we can now statically if a variant mapping is correct or not!
    pub fn variant_try_map_rule(
        &self,
        rule: impl AsRef<str>,
        to: &[ASTPatternElement],
    ) -> Option<Vec<(ASTVar, ASTType)>> {
        // Get the rule-set.
        let rule = self.content().1.get(rule.as_ref())?;

        // The first thing must be a constant string or identifier with the name of the sub-rule.
        use ASTPatternElement::*;
        let (rule, content) = match to {
            [Const(ASTConst::String(sub_rule)), content @ ..] => (rule.rule(sub_rule)?, content),
            [Var(var), content @ ..] => {
                debug_assert!(
                    var.get_specified_type().is_none(),
                    "got a specified type for `{var}`, expected just an identifier"
                );
                (rule.rule(var.name())?, content)
            }
            _ => return None,
        };

        // Be sure that we capture everything in that rule. We don't have the `..` like in rust...
        if rule.len() != content.len() {
            return None;
        }

        Some(
            // Do the matching for the sub-rule.
            rule.iter().zip(content).map(|(rule, pat)| match pat {
                // If a constant is provided we must ensure that the types are compatible!
                Const(pat) => rule.coercible_to(&pat.get_const_type()).then_some(None).ok_or(()),
                // If a variable is declared then we will return it. If the variable has a
                // specified type, we must ensure that the element in the variant can be coerced
                // into this specified type.
                Var(var) => match var.get_specified_type() {
                    Some(ty) => rule.coercible_to(ty).then(|| Some((var.clone(), rule.clone()))).ok_or(()),
                    None => Ok(Some((var.clone(), rule.clone()))),
                },
            })
            // Get ride of any errors because of impossible mappings.
            .collect::<Result<Vec<Option<(ASTVar, ASTType)>>, ()>>().ok()?
            // We collect the successful bindings into a vector, we don't care when a constant
            // mapped with an element of the variant.
            .into_iter().flatten().collect(),
        )
    }

    /// Get a set of rule for a specified variant if it exists.
    pub fn variant_complies_with_rule<T: Typed>(&self, rule: impl AsRef<str>, variant: &ASTVariant<T>) -> bool {
        let rules = self.content().1.get(rule.as_ref());
        rules.map_or(false, |rules| rules.complies(self, variant))
    }

    /// Create a new context, don't clone as we want to do other things...
    #[inline]
    pub fn for_scope(&self) -> Self {
        self.clone()
    }

    /// Get the type of a variable by its name.
    pub fn get(&self, name: &ASTVar) -> ASTType {
        let (scope, ..) = self.content();
        scope.get(name.name().as_ref()).unwrap().clone()
    }

    /// Declare a new variable. Returns an error on redefinition.
    pub fn declare(&mut self, var: ASTVar, ty: ASTType) {
        if var.is_joker() {
            log::debug!("can't declare the joker as a variable at {}", var.span());
        } else if self.content_mut().0.insert(var.name(), ty).is_some() {
            panic!()
        }
    }

    pub fn has_import(&self, import: impl AsRef<str>) -> bool {
        self.content().3.iter().any(|scoped| scoped.as_ref() == import.as_ref())
    }

    /// Get the string cache out of the module pointer stored in the context.
    #[inline]
    pub fn strings(&self) -> &ASTStringCacheHandle {
        &self.content().2
    }
}
