use crate::ast::*;

/// Types. Every object that can be manipulated by the user has a type. We have base types and
/// collection types and special ASS types.
///
/// Some coercion is possible between collection types...
#[derive(Debug, PartialEq, Eq, Clone, Default)]
pub enum ASTType {
    /// Null, unit, nullptr, etc. This is the default type
    #[default]
    Nil,

    /// An integer
    Integer,

    /// A floating point number
    Floating,

    /// A string.
    String,

    /// A boolean, true/false.
    Boolean,

    /// A color
    Color,

    /// A movement of a syllabe or line.
    Movement,

    /// A variant, can be a color or a movement.
    Variant,

    /// A tuple
    Tuple(Vec<ASTType>),

    /// Special ASS type, a syllabe. A simple syllabe.
    Syllabe,

    /// Special ASS type, a line. It's a collection of [ASTType::Syllabe] with extra steps, can
    /// iterate over it.
    Line,

    /// A table where the content is not really defined. Because we don't know the keys we can't
    /// iterate over it.
    AnyTable,

    /// A table where everything inside is defined, a struct. We can't iterate over it but we
    /// statically know the keys of the table.
    Table(ASTTable<ASTType>),

    /// A uniform table, where all the values has the same types, like a sequence but which string
    /// keys.
    UniformTable(Box<ASTType>),

    /// A sequence where the content is not really defined, like the any table but with sequences.
    AnySequence,

    /// A sequence where every value has the same type. We can iterate over it, like a uniform
    /// table, but with integer and continuous keys.
    Sequence(Box<ASTType>),

    /// A function type, needs the arguments and the return value.
    Function(Vec<ASTType>, Box<ASTType>),

    /// A type that is not known at compile time, the user must check its type with a switch case.
    Any,
}

impl ASTType {
    /// Try to see if a type can be pattern matched and if it's the case, returns [Some] bindings
    /// to add them to the type context, otherwise return [None].
    pub fn try_match(&self, mctx: &ASTTypeContext, to: &[ASTPatternElement]) -> Option<Vec<(ASTVar, ASTType)>> {
        self.get_variant_rule()
            .and_then(|rule| mctx.variant_try_map_rule(rule, to))
    }

    /// Tells wether the type is buildable in a const context or not.
    #[inline]
    pub fn is_const_constructible(&self) -> bool {
        matches!(self, ASTType::Nil | ASTType::Tuple(_) | ASTType::Any)
            || self.is_numeric()
            || self.is_table()
            || self.is_sequence()
    }

    /// Tells wether the type is numeric or not.
    #[inline]
    pub fn is_numeric(&self) -> bool {
        matches!(self, ASTType::Boolean | ASTType::Integer | ASTType::Floating)
    }

    /// Tells whether the type is any sort of sequence or not.
    #[inline]
    pub fn is_sequence(&self) -> bool {
        matches!(self, ASTType::Sequence(_) | ASTType::AnySequence)
    }

    /// Tells whether the type is any sort of table or not.
    #[inline]
    pub fn is_table(&self) -> bool {
        matches!(self, ASTType::UniformTable(_) | ASTType::Table(_) | ASTType::AnyTable)
    }

    /// Tells whether the type is any sort of variant or not.
    #[inline]
    pub fn is_variant(&self) -> bool {
        matches!(self, ASTType::Variant | ASTType::Color | ASTType::Movement)
    }

    /// Get the name of the rules to follow for this variant if it's really a variant.
    #[inline]
    pub fn get_variant_rule(&self) -> Option<&'static str> {
        match self {
            ASTType::Color => Some("color"),
            ASTType::Movement => Some("movement"),
            _ => None,
        }
    }

    /// Tells whether a type is trivially copiable, i.e. this type won't be managed.
    #[inline]
    pub fn is_trivialy_copiable(&self) -> bool {
        use ASTType::*;
        matches!(self, Nil | Integer | Floating | Boolean | Tuple(_))
    }

    pub fn coercible_to(&self, to: &ASTType) -> bool {
        use ASTType::*;
        match (self, to) {
            // Equals...
            (x, y)
                if (x == y)
                    || (x.is_table() && matches!(y, AnyTable))
                    || (x.is_sequence() && matches!(y, AnySequence))
                    || (x.is_variant() && y.is_variant()) =>
            {
                true
            }

            // Nil things
            (_, Nil) | (Nil, Integer | Floating | Boolean) => true,
            (Nil, to) => to.is_table() || to.is_sequence(),

            // Integer things
            (Integer, Floating | Boolean) | (Boolean, Integer | Floating) | (Floating, Integer) => true,

            // Sequence / ASS things
            (String, Sequence(ty)) | (Sequence(ty), String) => matches!(&**ty, String),
            (Sequence(x), Line) | (Line, Sequence(x)) => matches!(&**x, Syllabe),
            (Syllabe, Sequence(x)) => matches!(&**x, String),

            // Table / Sequence things
            (from, AnyTable) => from.is_sequence() || from.is_table() || from.is_variant(),
            (Sequence(to), UniformTable(from)) | (UniformTable(from), Sequence(to)) => from.coercible_to(to),
            (Table(table), Sequence(ty)) => {
                let integer_keys = table.keys().all(|key| key.parse::<i64>().is_ok());
                let mut uniform_values = table.values().collect::<Vec<_>>();
                let uniform_values = uniform_values
                    .pop()
                    .map(|ty| uniform_values.into_iter().all(|a| ty.eq(a)).then_some(ty));
                integer_keys && matches!(uniform_values, Some(Some(table_ty)) if table_ty.eq(ty))
            }

            _ => false,
        }
    }
}

impl Typed for ASTType {
    #[inline]
    fn get_type(&self, _: &ASTTypeContext) -> ASTType {
        self.clone()
    }
}

impl std::fmt::Display for ASTType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ASTType::Nil => f.write_str("nil"),
            ASTType::Any => f.write_str("any"),

            ASTType::Integer => f.write_str("integer"),
            ASTType::Floating => f.write_str("float"),
            ASTType::String => f.write_str("string"),
            ASTType::Boolean => f.write_str("boolean"),
            ASTType::Color => f.write_str("color"),
            ASTType::Variant => f.write_str("variant"),
            ASTType::Movement => f.write_str("movement"),

            ASTType::Function(args, ret) => {
                f.write_str("function (")?;
                if let Some(ty) = args.iter().next() {
                    write!(f, " {ty}")?;
                }
                for ty in args.iter().skip(1) {
                    write!(f, ", {ty}")?;
                }
                write!(f, " ) -> {ret}")
            }

            ASTType::Tuple(tuple) => {
                f.write_str("(")?;
                if let Some(ty) = tuple.iter().next() {
                    write!(f, " {ty}")?;
                }
                for ty in tuple.iter().skip(1) {
                    write!(f, ", {ty}")?;
                }
                f.write_str(" )")
            }

            ASTType::Line => f.write_str("line"),
            ASTType::Syllabe => f.write_str("syllabe"),
            ASTType::AnySequence => f.write_str("sequence { any }"),
            ASTType::Sequence(ty) => write!(f, "sequence {{ {ty} }}"),

            ASTType::AnyTable => f.write_str("table { any }"),
            ASTType::UniformTable(inner) => write!(f, "table {{ {inner} }}"),
            ASTType::Table(table) => {
                f.write_str("table ")?;
                write_ast_table(f, table)
            }
        }
    }
}
