use crate::ast::*;
use std::sync::atomic::{AtomicU64, Ordering};

/// Variable thing. Have a name, a where-it-is-defined span and optionally a type. Having to no
/// type means that the type was not already found.
///
/// To get the type of a variable, it is either present in the context type hashmap, then we check
/// the specified type, if all failed we return the [ASTType::Nil] type.
#[derive(Debug, Clone)]
pub struct ASTVar(ASTSpan, ASTString, Option<ASTType>);

#[derive(Debug, Clone)]
pub struct ASTVarBuilder<'a> {
    span: ASTSpan,
    cache: &'a ASTStringCacheHandle,
    name: Option<ASTString>,
    specified_ty: Option<ASTType>,
}

impl<'a> ASTVarBuilder<'a> {
    pub fn build(self) -> ASTVar {
        let name = self.name.unwrap_or_else(|| {
            static ANON: AtomicU64 = AtomicU64::new(0);
            self.cache.get(format!("_{}", ANON.fetch_add(1, Ordering::SeqCst)))
        });
        ASTVar(self.span, name, self.specified_ty)
    }

    pub fn span(mut self, span: impl Into<ASTSpan>) -> Self {
        self.span = span.into();
        self
    }

    pub fn with_type(mut self, ty: ASTType) -> Self {
        self.specified_ty = Some(ty);
        self
    }

    pub fn name_from(mut self, name: impl AsRef<str>) -> Self {
        self.name = Some(self.cache.get(name.as_ref()));
        self
    }

    pub fn name(mut self, name: ASTString) -> Self {
        self.name = Some(name);
        self
    }
}

impl PartialEq for ASTVar {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        self.1 == other.1 && self.2 == other.2
    }
}

impl ASTVar {
    #[inline]
    #[allow(clippy::new_ret_no_self)]
    pub fn new(cache: &ASTStringCacheHandle) -> ASTVarBuilder {
        ASTVarBuilder { span: Default::default(), cache, name: None, specified_ty: None }
    }

    #[inline]
    pub fn span(&self) -> ASTSpan {
        self.0
    }

    /// Get the specified type if any.
    #[inline]
    pub fn get_specified_type(&self) -> Option<&ASTType> {
        self.2.as_ref()
    }

    /// Get the specified type. If no type is specified then we say that we return [ASTType::Nil].
    #[inline]
    pub fn get_specified_type_or_nil(&self) -> ASTType {
        self.get_specified_type().cloned().unwrap_or_default()
    }

    /// Set the specified type.
    #[inline]
    pub fn set_specified_type(&mut self, ty: ASTType) {
        self.2 = Some(ty);
    }

    /// Get the name.
    #[inline]
    pub fn name(&self) -> ASTString {
        self.1.clone()
    }

    /// Tells wether the variable is the joker or not. We can assign into the joker, in this case
    /// we yeet the result, but we can never read it.
    pub fn is_joker(&self) -> bool {
        self.1.as_ref().eq("_")
    }
}

impl From<&ASTVar> for ASTSpan {
    #[inline]
    fn from(value: &ASTVar) -> Self {
        value.span()
    }
}

impl std::fmt::Display for ASTVar {
    #[inline]
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.1.as_ref())
    }
}

impl Typed for ASTVar {
    #[inline]
    fn get_type(&self, ctx: &ASTTypeContext) -> ASTType {
        ctx.get(self)
    }
}
