use crate::ast::*;

/// A trait to describes types that are variants. Can be used to handle colors and movements in a
/// generic way.
///
/// Represents a color in the AST. A color can be:
/// - `#(rgb 0, 0, 0)` for black
/// - `#(green)` for green
/// - etc...
///
/// Represents a movement in the AST, a movement can be:
/// - `#[pos 500, 500]`
/// - `#[cmove 0, 0, 100, 50, 30]`
/// - etc...
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct ASTVariant<T: Typed> {
    pub(crate) variant: ASTString,
    pub(crate) args: Vec<T>,
}

pub(crate) struct ASTFmtColor<'a, T: Typed>(&'a ASTVariant<T>, &'a ASTTypeContext);
pub(crate) struct ASTFmtMovement<'a, T: Typed>(&'a ASTVariant<T>, &'a ASTTypeContext);

impl<T: Typed> ASTVariant<T> {
    #[inline]
    pub fn variant(&self) -> ASTString {
        self.variant.clone()
    }

    #[inline]
    pub fn args(&self) -> &[T] {
        &self.args
    }

    #[inline]
    pub fn args_mut(&mut self) -> &mut [T] {
        &mut self.args
    }

    #[inline]
    pub fn args_len(&self) -> usize {
        self.args.len()
    }
}

impl ASTVariant<ASTExpr> {
    #[inline]
    pub fn is_const_expr(&self) -> bool {
        self.args.iter().all(|expr| expr.is_const_expr())
    }
}

impl ASTVariant<ASTConst> {
    #[inline]
    pub fn is_const_expr(&self) -> bool {
        true
    }
}

impl<'a, T: Typed> ASTFmtColor<'a, T> {
    pub(crate) fn new(variant: &'a ASTVariant<T>, ctx: &'a ASTTypeContext) -> Self {
        Self(variant, ctx)
    }
}

impl<'a, T: Typed> ASTFmtMovement<'a, T> {
    pub(crate) fn new(variant: &'a ASTVariant<T>, ctx: &'a ASTTypeContext) -> Self {
        Self(variant, ctx)
    }
}

impl<'a, T: Typed> std::fmt::Display for ASTFmtColor<'a, T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self(ASTVariant { variant, args }, ctx) = self;
        write!(f, "#({variant}")?;
        if let Some(arg) = args.iter().next() {
            write!(f, " {}", arg.get_type(ctx))?;
        }
        for arg in args.iter().skip(1) {
            write!(f, ", {}", arg.get_type(ctx))?;
        }
        f.write_str(" )")
    }
}

impl<'a, T: Typed> std::fmt::Display for ASTFmtMovement<'a, T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self(ASTVariant { variant, args }, ctx) = self;
        write!(f, "#[{variant}")?;
        if let Some(arg) = args.iter().next() {
            write!(f, " {}", arg.get_type(ctx))?;
        }
        for arg in args.iter().skip(1) {
            write!(f, ", {}", arg.get_type(ctx))?;
        }
        f.write_str(" ]")
    }
}

/// Describes a set of rules for a variant.
#[derive(Debug, Clone)]
pub struct ASTVariantRules {
    family: ASTString,
    rules: HashMap<ASTString, Vec<ASTType>>,
}

macro_rules! rule {
    ($cache: expr, $name: literal) => { ($cache.get($name), vec![]) };
    ($cache: expr, $name: literal -> [$($arg: ident),*]) => { ($cache.get($name), vec![$(ASTType::$arg,)*]) };
}

impl ASTVariantRules {
    /// Default empty ruleset.
    #[inline]
    pub fn new(name: ASTString) -> Self {
        Self { family: name, rules: Default::default() }
    }

    /// Get the name of the family of variants.
    #[inline]
    pub fn name(&self) -> ASTString {
        self.family.clone()
    }

    /// Get all the rules of this family.
    #[inline]
    pub fn rules(&self) -> impl Iterator<Item = (&ASTString, &Vec<ASTType>)> {
        self.rules.iter()
    }

    /// Get a specific rule if present.
    #[inline]
    pub fn rule(&self, variant: impl AsRef<str>) -> Option<&[ASTType]> {
        self.rules.get(variant.as_ref()).map(|vec| vec.as_ref())
    }

    /// Create a new rule set for colors, the default one.
    #[inline]
    pub fn new_colors(cache: ASTStringCacheHandle) -> Self {
        Self {
            family: cache.get("color"),
            rules: HashMap::from_iter([
                rule!(cache, "rgb"  -> [Integer, Integer, Integer]),
                rule!(cache, "rgba" -> [Integer, Integer, Integer, Integer]),
                rule!(cache, "black"),
                rule!(cache, "white"),
                rule!(cache, "grey"),
                rule!(cache, "red"),
                rule!(cache, "green"),
                rule!(cache, "blue"),
            ]),
        }
    }

    /// Create a new rule set for movements, the default one.
    #[inline]
    pub fn new_movements(cache: ASTStringCacheHandle) -> Self {
        Self {
            family: cache.get("movement"),
            rules: HashMap::from_iter([
                rule![cache, "implicit"],
                rule!(cache, "fixed"       -> [Integer, Integer]),
                rule!(cache, "linear"      -> [Integer, Integer, Integer, Integer]),
                rule!(cache, "accelerated" -> [Integer, Integer, Integer, Integer, Floating]),
            ]),
        }
    }

    /// Insert a new rule in the ruleset, returns [true] if the rule was inserted successfully,
    /// [false] otherwise.
    #[must_use]
    #[inline]
    pub fn insert_rule(&mut self, name: ASTString, types: Vec<ASTType>) -> bool {
        self.rules.insert(name, types).is_none()
    }

    /// Returns wether a variant complies with the set of allowed one.
    #[must_use]
    #[inline]
    pub fn complies<T: Typed>(&self, ctx: &ASTTypeContext, variant: &ASTVariant<T>) -> bool {
        match self.rules.get(&variant.variant) {
            None => false,
            Some(rule) => {
                (rule.len() == variant.args.len())
                    && rule
                        .iter()
                        .zip(variant.args.iter().map(|arg| arg.get_type(ctx)))
                        .all(|(rule, arg)| arg.coercible_to(rule))
            }
        }
    }
}
