#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum ASTVisibility {
    #[default]
    Private,
    Public,
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum ASTVisibilityRule {
    Only(ASTVisibility),
    AtLeast(ASTVisibility),
}

impl ASTVisibilityRule {
    /// Checks if the rule allows to show an item with the specified visibility [ASTVisibility].
    /// Used to factorize the functions in [crate::ast::ASTModule] and [crate::ast::ASTProgram].
    #[inline]
    pub fn allows(&self, vis: ASTVisibility) -> bool {
        match self {
            ASTVisibilityRule::Only(this) => vis.eq(this),
            ASTVisibilityRule::AtLeast(this) => PartialOrd::le(this, &vis),
        }
    }

    /// Allows all the visibilities.
    #[inline]
    pub fn any() -> Self {
        Self::AtLeast(ASTVisibility::Private)
    }
}

impl Default for ASTVisibilityRule {
    fn default() -> Self {
        Self::Only(ASTVisibility::Private)
    }
}

impl ASTVisibility {
    #[inline]
    pub fn as_str(&self) -> &str {
        match self {
            ASTVisibility::Private => "private",
            ASTVisibility::Public => "public",
        }
    }
}

impl AsRef<str> for ASTVisibility {
    #[inline]
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

#[test]
fn test_visibility() {
    assert!(ASTVisibility::Private < ASTVisibility::Public);
}

#[test]
fn test_visibility_rules() {
    use ASTVisibility::*;
    use ASTVisibilityRule::*;

    assert!(Only(Private).allows(Private));
    assert!(!Only(Private).allows(Public));
    assert!(Only(Public).allows(Public));
    assert!(!Only(Public).allows(Private));

    assert!(AtLeast(Private).allows(Public));
    assert!(AtLeast(Private).allows(Private));
    assert!(AtLeast(Public).allows(Public));
    assert!(!AtLeast(Public).allows(Private));

    assert!(ASTVisibilityRule::any().allows(Public));
    assert!(ASTVisibilityRule::any().allows(Private));
}
