//! Contains wrappers for the container, the line, syllabes, etc...
#![allow(clippy::missing_safety_doc)]

use crate::{AuxTable, StringSlice};
use alloc::{boxed::Box, vec::Vec};
use core::{
    ffi::{c_char, CStr},
    mem,
    ptr::{self, NonNull},
};
use vvs_ass::{ass_container_from_file, ASSStyle};

/// Wraps the container that holds all the content of the ASS file / Vivy file.
pub struct ASSContainer(vvs_ass::ASSContainer<AuxTable>);

/// Wraps a line in an ASS file.
pub struct ASSLine(vvs_ass::ASSLine<AuxTable>);

/// Wraps syllabes, contained in lines in ASS files.
pub struct ASSSyllabe(vvs_ass::ASSSyllabe<AuxTable>);

/// Get the number of lines in the container.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSContainerGetLinesCount(this: *const ASSContainer) -> i32 { (*this).0.lines.len() as i32 }

/// Get the number of syllabes in the line.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSLineGetSyllabesCount(this: *const ASSLine) -> i32 { (*this).0.content.len() as i32 }

/// Get a line by its number in the container.
#[no_mangle]
pub unsafe extern "C" fn ASSContainerGetLineAt(this: *mut ASSContainer, idx: i32) -> *mut ASSLine {
    let line = (*this).0.lines.get_mut(idx as usize).map(|l| l as *mut _ as *mut _);
    line.unwrap_or_else(|| {
        log::error!("failed to get line n°{idx}");
        core::ptr::null_mut()
    })
}

/// Get a syllabe in a line by its index.
#[no_mangle]
pub unsafe extern "C" fn ASSLineGetSyllabeAt(this: *mut ASSLine, idx: i32) -> *mut ASSSyllabe {
    let syl = (*this).0.content.get_mut(idx as usize).map(|s| s as *mut _ as *mut _);
    syl.unwrap_or_else(|| {
        log::error!("failed to get syllabe n°{idx}");
        core::ptr::null_mut()
    })
}

/// Tells whever a line is commented or not.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSLineIsCommented(this: *const ASSLine) -> bool { (*this).0.is_comment }

/// Gets the style of the line, or 'Default' if it was not found/specified.
#[no_mangle]
pub unsafe extern "C" fn ASSLineGetStyle<'a>(this: *const ASSLine) -> StringSlice<'a> {
    match (*this).0.aux.0.get("style") {
        Some(str) => StringSlice::from(str),
        _ => StringSlice::from("Default"),
    }
}

/// Get the content of a syllabe.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSSyllabeGetContent<'a>(this: *const ASSSyllabe) -> StringSlice<'a> { StringSlice::from(&(*this).0.content) }

/// Get the duration of a syllabe.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSSyllabeGetDuration(this: *const ASSSyllabe) -> i32 { (*this).0.fini.saturating_sub((*this).0.start) }

/// Get the start of a syllabe.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSSyllabeGetStartTime(this: *const ASSSyllabe) -> i32 { (*this).0.start }

/// Get the end time of a syllabe.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSSyllabeGetFiniTime(this: *const ASSSyllabe) -> i32 { (*this).0.fini }

/// Set the duration of a syllabe.
#[no_mangle]
pub unsafe extern "C" fn ASSSyllabeSetDuration(this: *mut ASSSyllabe, duration: i32) {
    (*this).0.fini = (*this).0.start.saturating_add(duration);
}

/// Set the start of a syllabe.
#[no_mangle]
pub unsafe extern "C" fn ASSSyllabeSetStartTime(this: *mut ASSSyllabe, time: i32) {
    (*this).0.start = time;
    (*this).0.fini = (*this).0.fini.max(time);
}

/// Set the end of a syllabe.
#[no_mangle]
pub unsafe extern "C" fn ASSSyllabeSetFiniTime(this: *mut ASSSyllabe, time: i32) {
    (*this).0.fini = time;
    (*this).0.start = (*this).0.fini.min(time);
}

/// Load the ASS from a file, returns nullptr if an error was encountred.
#[no_mangle]
pub unsafe extern "C" fn ASSContainerFromFile(path: NonNull<c_char>) -> *mut ASSContainer {
    let container = CStr::from_ptr(path.as_ptr() as *const _).to_str().map(|path| {
        ass_container_from_file(path)
            .map(|ass: vvs_ass::ASSContainer<AuxTable>| Box::into_raw(Box::new(ass)) as *mut ASSContainer)
            .map_err(|err| log::error!("failed to load file: {err}"))
    });
    match container.inspect_err(|err| log::error!("invalid utf8 filename: {err}")) {
        Ok(Ok(container)) => container,
        _ => ptr::null_mut(),
    }
}

/// Get a style out of the container by its name.
#[no_mangle]
pub unsafe extern "C" fn ASSContainerGetStyleByName(this: *mut ASSContainer, name: StringSlice) -> *mut ASSStyle {
    let style = (*this).0.styles.get_mut(name.as_str()).map(|s| s as *mut _);
    style.unwrap_or_else(|| {
        log::error!("no sytle `{}` in ass container", name.as_str());
        core::ptr::null_mut()
    })
}

/// Get the name of a style by its position.
#[no_mangle]
pub unsafe extern "C" fn ASSContainerGetStyleNameAt<'a>(this: *const ASSContainer, idx: i32) -> StringSlice<'a> {
    let mut keys: Vec<_> = (*this).0.styles.keys().collect();
    keys.sort();
    keys.get(idx as usize)
        .map(StringSlice::from)
        .unwrap_or_else(|| StringSlice::from("Default"))
}

/// Get the number of styles in the container.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSContainerGetStylesCount(this: *const ASSContainer) -> i32 { (*this).0.styles.len() as i32 }

/// Set the start time of the line.
#[no_mangle]
pub unsafe extern "C" fn ASSLineSetStartTime(this: *mut ASSLine, time: i32) {
    (*this).0.start = time;
    (*this).0.fini = (*this).0.fini.max(time);
    for syl in &mut (*this).0.content {
        syl.start = syl.start.max(time);
        syl.fini = syl.fini.max(time);
    }
}

/// Set the end time of a line.
#[no_mangle]
pub unsafe extern "C" fn ASSLineSetFiniTime(this: *mut ASSLine, time: i32) {
    (*this).0.fini = time;
    (*this).0.start = (*this).0.start.min(time);
    for syl in &mut (*this).0.content {
        syl.start = syl.start.min(time);
        syl.fini = syl.fini.min(time);
    }
}

/// Get the duration of a line.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSLineGetDuration(this: *const ASSLine) -> i32 { (*this).0.fini.saturating_sub((*this).0.start) }

/// Set the duration of a line.
#[no_mangle]
pub unsafe extern "C" fn ASSLineSetDuration(this: *mut ASSLine, duration: i32) {
    (*this).0.fini = (*this).0.start.saturating_add(duration);
}

/// Get the start time of a line.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSLineGetStartTime(this: *const ASSLine) -> i32 { (*this).0.start }

/// Get the end time of a line.
#[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ASSLineGetFiniTime(this: *const ASSLine) -> i32 { (*this).0.fini }

/// Drop the container. It is file to pass a null pointer to this function.
#[no_mangle]
pub unsafe extern "C" fn ASSContainerDrop(this: *mut ASSContainer) {
    #[allow(dead_code)]
    const ASSERT_STRUCT_SIZES: () = {
        macro_rules! assert_sizes {
            ($ty: ident) => {
                assert!(
                    mem::size_of::<$ty>() == mem::size_of::<vvs_ass::$ty<AuxTable>>(),
                    concat!("invalid sizes for ", stringify!($ty), " wrapper")
                );
                assert!(
                    mem::align_of::<$ty>() == mem::align_of::<vvs_ass::$ty<AuxTable>>(),
                    concat!("invalid alignements for ", stringify!($ty), " wrapper")
                );
            };
        }
        assert_sizes!(ASSContainer);
        assert_sizes!(ASSLine);
        assert_sizes!(ASSSyllabe);
    };

    if !this.is_null() {
        drop(unsafe { Box::from_raw(this) });
    }
}
