use crate::prelude::*;
use llvm_sys::{core::*, prelude::*};
use std::{
    ffi::{CString, NulError},
    marker,
};

crate::bindings::declare! { mut LLVMContextRef as Context<'a> }

macro_rules! get_type {
    ($llvm:ident as $name:ident) => {
        paste::paste! {
            #[doc = "Get the type for " $llvm "."]
            pub fn [< type_ $name >](&self) -> Type {
                unsafe { Type::from_ptr([< LLVM $llvm TypeInContext >](self.inner)) }
            }
        }
    };
}

impl Default for Context<'_> {
    fn default() -> Self {
        Self::new()
    }
}

impl Context<'_> {
    /// Create a new context from scratch. This will automatically initialize LLVM if needed.
    pub fn new() -> Self {
        if let Err(err) = crate::init::initialize_llvm() {
            // Should fail only on re-init. If a real problem occured, we would have already
            // aborted the program.
            log::debug!("{err}");
        }
        Self { inner: unsafe { LLVMContextCreate() }, marker: marker::PhantomData }
    }

    /// Create a new builder for this context.
    pub fn create_builder(&self) -> Builder {
        unsafe { Builder::from_ptr(LLVMCreateBuilderInContext(self.inner)) }
    }

    /// Create a new module in this context.
    pub fn create_module(&self, name: impl AsRef<str>) -> Result<Module, NulError> {
        let name = CString::new(name.as_ref())?;
        Ok(unsafe { Module::from_ptr(LLVMModuleCreateWithName(name.as_ptr())) })
    }

    get_type! {  Int8 as   i8 }
    get_type! { Int32 as  i32 }
    get_type! { Float as  f32 }
    get_type! {  Void as void }
}

impl Drop for Context<'_> {
    fn drop(&mut self) {
        unsafe { LLVMContextDispose(self.inner) }
    }
}
