use crate::prelude::*;
use llvm_sys::{core::*, prelude::*, LLVMLinkage, LLVMVisibility};

crate::bindings::declare! { mut LLVMValueRef as FunctionDeclaration<'a> }
crate::bindings::declare! { mut LLVMValueRef as Function<'a> {
    entry_point: BB<'a>,
} }

impl<'a> FunctionDeclaration<'a> {
    /// Set the [LLVMLinkage] and [LLVMVisibility] accordingly.
    pub fn into_public(self, public: bool) -> Self {
        match public {
            true => unsafe {
                LLVMSetLinkage(self.as_ptr(), LLVMLinkage::LLVMCommonLinkage);
                LLVMSetVisibility(self.as_ptr(), LLVMVisibility::LLVMProtectedVisibility);
            },
            false => unsafe {
                LLVMSetLinkage(self.as_ptr(), LLVMLinkage::LLVMLinkerPrivateLinkage);
                LLVMSetVisibility(self.as_ptr(), LLVMVisibility::LLVMHiddenVisibility);
            },
        }
        self
    }

    /// Get the type of this function declaration.
    pub fn ty(&self) -> FunctionType {
        unsafe { FunctionType::from_ptr(LLVMTypeOf(self.as_ptr())) }
    }

    /// Transform the function declaration into a proper function, with a function body, or at
    /// lease an entry point.
    pub fn into_function(self) -> Function<'a> {
        unsafe {
            let entry_point = BB::from_ptr(LLVMAppendBasicBlockInContext(
                LLVMGetTypeContext(self.ty().as_ptr()),
                self.as_ptr(),
                crate::bindings::cstr(b"entrypoint\0").as_ptr(),
            ));
            Function::from_ptr(self.as_ptr(), entry_point)
        }
    }
}

impl<'a> Function<'a> {
    /// Get the type of this function.
    pub fn ty(&self) -> FunctionType {
        unsafe { FunctionType::from_ptr(LLVMTypeOf(self.as_ptr())) }
    }

    /// Get the entry point of the function.
    pub fn entry_point(&self) -> &BB {
        &self.entry_point
    }

    /// Get the entry point of the function in a mutable way.
    pub fn entry_point_mut(&'a mut self) -> &'a mut BB {
        &mut self.entry_point
    }

    /// Iterate over all the basic blocks of the function.
    pub fn iter(&'a self) -> FunctionIter<'a, &'a Function> {
        unsafe {
            match LLVMCountBasicBlocks(self.as_ptr()) {
                0 => FunctionIter { curr: None, last: None, func: self },
                _ => FunctionIter {
                    curr: Some(BB::from_ptr(LLVMGetFirstBasicBlock(self.inner))),
                    last: Some(BB::from_ptr(LLVMGetLastBasicBlock(self.inner))),
                    func: self,
                },
            }
        }
    }

    /// Iterate over all the basic blocks of the function in a mutable way.
    pub fn iter_mut(&'a mut self) -> FunctionIter<'a, &'a mut Function> {
        unsafe {
            match LLVMCountBasicBlocks(self.as_ptr()) {
                0 => FunctionIter { curr: None, last: None, func: self },
                _ => FunctionIter {
                    curr: Some(BB::from_ptr(LLVMGetFirstBasicBlock(self.inner))),
                    last: Some(BB::from_ptr(LLVMGetLastBasicBlock(self.inner))),
                    func: self,
                },
            }
        }
    }
}

mod sealed {
    pub trait FuncRef<'a>: AsRef<super::Function<'a>> {}
}

impl<'a> sealed::FuncRef<'a> for &'a Function<'a> {}
impl<'a> sealed::FuncRef<'a> for &'a mut Function<'a> {}

impl<'a> AsRef<Function<'a>> for &'a Function<'a> {
    fn as_ref(&self) -> &Function<'a> {
        self
    }
}

impl<'a> AsRef<Function<'a>> for &'a mut Function<'a> {
    fn as_ref(&self) -> &Function<'a> {
        self
    }
}

/// Iterate over the basic blocks in a function.
pub struct FunctionIter<'a, F: sealed::FuncRef<'a>> {
    curr: Option<BB<'a>>,
    last: Option<BB<'a>>,
    func: F,
}

impl<'a, F: sealed::FuncRef<'a>> Iterator for FunctionIter<'a, F> {
    type Item = BB<'a>;

    fn next(&mut self) -> Option<Self::Item> {
        match (self.curr.take(), self.last.as_ref()) {
            (Some(_), None) | (None, Some(_)) => unreachable!(),
            (None, None) => None,
            (Some(curr), Some(last)) if curr == *last => {
                self.last = None;
                Some(curr)
            }
            (Some(curr), Some(_)) => unsafe {
                self.curr = Some(BB::from_ptr(LLVMGetNextBasicBlock(curr.as_ptr())));
                Some(curr)
            },
        }
    }

    fn last(self) -> Option<Self::Item> {
        self.last
    }

    fn count(self) -> usize {
        (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let count = (unsafe { LLVMCountBasicBlocks(self.func.as_ref().as_ptr()) }) as usize;
        (count, Some(count))
    }
}
