#![forbid(missing_docs)]

//! Bindings for the C-API of LLVM. Documentation can be found here: https://llvm.org/doxygen/group__LLVMCCore.html

pub mod basic_block;
pub mod builder;
pub mod context;
pub mod function;
pub mod module;
pub mod types;
pub mod value;

macro_rules! declare {
    ($owned: tt $llvm:ident as $name:ident <$lifetime:lifetime> $({
        $($field:ident: $ty:ty,)*
    })?) => {
        #[doc = concat!("A ", stringify!($name), ", to safely use functions for ", stringify!($llvm))]
        pub struct $name<$lifetime> {
            inner: $llvm,
            marker: $crate::bindings::declare!(@marker $owned $llvm $lifetime),
            $($($field: $ty)*)?
        }

        #[allow(dead_code)]
        impl<$lifetime> $name<$lifetime> {
            /// Create a new instance from a pointer.
            ///
            /// # Safety
            /// Must be used from functions of the [crate::bindings::context::Context] to be sure that the
            /// usage is correct...
            pub(super) unsafe fn from_ptr(inner: $llvm $($(, $field: $ty)*)?) -> Self {
                Self {
                    inner $($(, $field)*)?,
                    marker: core::marker::PhantomData
                }
            }

            /// Get the inner pointer.
            pub(super) fn as_ptr(&self) -> $llvm {
                self.inner
            }

        }
    };

    (@marker mut   $llvm:ident $lifetime:lifetime) => { core::marker::PhantomData<&$lifetime mut $llvm> };
    (@marker const $llvm:ident $lifetime:lifetime) => { core::marker::PhantomData<&$lifetime     $llvm> };
}
use declare;

const fn cstr(content: &[u8]) -> &std::ffi::CStr {
    match std::ffi::CStr::from_bytes_with_nul(content) {
        Ok(string) => string,
        Err(_) => panic!("you forgot the null byte at the end"),
    }
}
