use llvm_sys::{core::*, prelude::*, LLVMTypeKind};

crate::bindings::declare! { const LLVMTypeRef as Type<'a> }
crate::bindings::declare! { const LLVMTypeRef as FunctionType<'a> }

impl FunctionType<'_> {
    /// Create a new function type.
    pub fn new<const N: usize>(args: [Type; N], returns: Type) -> Self {
        unsafe {
            Self::from_ptr(LLVMFunctionType(returns.as_ptr(), args.map(|ty| ty.as_ptr()).as_mut_ptr(), N as u32, 0))
        }
    }

    /// Get the returned type of the function.
    pub fn returns(&self) -> Type {
        unsafe { Type::from_ptr(LLVMGetReturnType(self.as_ptr())) }
    }

    /// Get the number of arguments.
    pub fn args_len(&self) -> usize {
        (unsafe { LLVMCountParamTypes(self.as_ptr()) }) as usize
    }

    /// Get the arguments of the function type.
    pub fn args(&self) -> Vec<Type> {
        let params_count = self.args_len();
        let mut params = Vec::with_capacity(params_count);
        params.resize(params_count, std::ptr::null_mut());
        unsafe { LLVMGetParamTypes(self.as_ptr(), params.as_mut_ptr()) };
        params.into_iter().map(|ptr| unsafe { Type::from_ptr(ptr) }).collect()
    }
}

macro_rules! type_is {
    ($llvm:ident as $name:ident) => {
        paste::paste! {
            #[doc = "Tells wether the type is " $name " or not."]
            pub fn [< is_$name >](&self) -> bool {
                (unsafe { LLVMGetTypeKind(self.as_ptr()) }) == LLVMTypeKind::$llvm
            }
        }
    };
}

impl Type<'_> {
    type_is! { LLVMIntegerTypeKind  as int   }
    type_is! { LLVMVoidTypeKind     as void  }
    type_is! { LLVMArrayTypeKind    as array }
    type_is! { LLVMFloatTypeKind    as f32   }
    type_is! { LLVMLabelTypeKind    as label }
    type_is! { LLVMPointerTypeKind  as ptr   }
    type_is! { LLVMVectorTypeKind   as vec   }
}
