//! Initialize LLVM targets and such. This module only exports one function:
//! [self::initialize_llvm]. Depending on the target architecture we select
//! the correct function...

use anyhow::anyhow;
use llvm_sys::target_machine::*;
use std::ffi::CStr;

/// Be sure to init LLVM only once. We use an atomic usize because it is supported on all the
/// platforms!
fn call_once() -> bool {
    use std::sync::atomic::{AtomicUsize, Ordering::*};
    static INIT_ONCE: AtomicUsize = AtomicUsize::new(0);
    INIT_ONCE.compare_exchange(0, 1, SeqCst, SeqCst).is_ok()
}

fn print_llvm_infos() {
    let triple = unsafe { CStr::from_ptr(LLVMGetDefaultTargetTriple()) }.to_string_lossy();
    let cpu = unsafe { CStr::from_ptr(LLVMGetHostCPUName()) }.to_string_lossy();
    let cpu_features = unsafe { CStr::from_ptr(LLVMGetHostCPUFeatures()) }.to_string_lossy();

    log::debug!("got default triple: `{triple}`");
    log::debug!("got cpu `{cpu}` with features: {cpu_features}");
}

#[cfg(target_arch = "x86_64")]
pub fn initialize_llvm() -> anyhow::Result<()> {
    call_once()
        .then(|| unsafe {
            llvm_sys::target::LLVMInitializeX86Target();
            print_llvm_infos();
        })
        .ok_or_else(|| anyhow!("llvm was already initialized for this executable!"))
}

#[cfg(not(target_arch = "x86_64"))]
compile_error!("this target is not supported for now");
