//! Utility functions and structures to iterate over an LLVM function and in the basic blocks.

use llvm_sys::{core::*, prelude::*};
use std::ptr;

/// Iterate over the basic blocks in a function.
#[derive(Clone, Copy)]
pub struct LLVMFunctionIter {
    curr_bb: LLVMBasicBlockRef,
    last_bb: LLVMBasicBlockRef,
    function: LLVMValueRef,
    index: u32,
}

#[derive(Clone, Copy)]
pub struct LLVMBasicBlockIter {
    curr_instr: LLVMValueRef,
    last_instr: LLVMValueRef,
}

impl Iterator for LLVMFunctionIter {
    type Item = LLVMBasicBlockRef;

    fn next(&mut self) -> Option<Self::Item> {
        (!self.curr_bb.is_null() && !self.last_bb.is_null())
            .then(|| match unsafe { LLVMGetNextBasicBlock(self.curr_bb) } {
                ret if ret.is_null() => None,
                ret => {
                    self.curr_bb = if ret == self.last_bb { ptr::null_mut() } else { ret };
                    self.index += 1;
                    Some(ret)
                }
            })
            .flatten()
    }

    fn last(self) -> Option<Self::Item> {
        (!self.last_bb.is_null()).then_some(self.last_bb)
    }

    fn count(self) -> usize {
        unsafe { LLVMFunctionIter::count_bbs(&self) }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let count = unsafe { LLVMFunctionIter::count_bbs(self) };
        (count, Some(count))
    }
}

impl Iterator for LLVMBasicBlockIter {
    type Item = LLVMValueRef;

    fn next(&mut self) -> Option<Self::Item> {
        (!self.curr_instr.is_null() && !self.last_instr.is_null())
            .then(|| match unsafe { LLVMGetNextInstruction(self.curr_instr) } {
                ret if ret.is_null() => None,
                ret => {
                    self.curr_instr = if ret == self.last_instr { ptr::null_mut() } else { ret };
                    Some(ret)
                }
            })
            .flatten()
    }

    fn last(self) -> Option<Self::Item> {
        (!self.last_instr.is_null()).then_some(self.last_instr)
    }
}

impl LLVMFunctionIter {
    /// Create a new iterator over a function. If the passed value is not a function this
    /// constructor will panic.
    ///
    /// # Safety
    /// The passed value must be a valid pointer to [LLVMValue]. You may not create new basic
    /// blocks while using this iterator.
    pub unsafe fn new(function: LLVMValueRef) -> Self {
        assert!(!LLVMIsAFunction(function).is_null(), "should pass functions...");
        Self {
            curr_bb: LLVMGetFirstBasicBlock(function),
            last_bb: LLVMGetLastBasicBlock(function),
            index: 0,
            function,
        }
    }

    unsafe fn count_bbs(&self) -> usize {
        unsafe { LLVMCountBasicBlocks(self.function) }.saturating_sub(self.index) as usize
    }
}

impl LLVMBasicBlockIter {
    /// Create a new iterator over a basic block.
    ///
    /// # Safety
    /// The passed value must be a valid pointer to [LLVMBasicBlock]. You may not insert new values
    /// into the block while using this iterator.
    pub unsafe fn new(bb: LLVMBasicBlockRef) -> Self {
        Self { curr_instr: LLVMGetFirstInstruction(bb), last_instr: LLVMGetBasicBlockTerminator(bb) }
    }
}
