//! Re-exports LLVM stuff

#[cfg(feature = "link")]
pub mod build {
    //! Utility functions to link against LLVM without relying to [llvm-sys]. The function
    //! [llvm_link] should be run from the build.rs script of any executable or library depending
    //! on [vvs_llvm].

    use anyhow::{anyhow, bail, Context};
    use std::process::{Command, Output};

    /// Run a command with said arguments and returns the stdout of said command.
    fn run<const N: usize>(cmd: &'static str, args: [&'static str; N]) -> anyhow::Result<Vec<u8>> {
        match Command::new(cmd).args(args).output()? {
            Output { status, stdout, .. } if status.success() => Ok(stdout),
            Output { stderr, .. } => Err(anyhow!(
                "failed to run `{cmd}`: {}",
                std::str::from_utf8(&stderr).with_context(|| format!("`{cmd}` failed with non-utf8 stderr"))?
            )),
        }
    }

    /// Prints info for cargo to link against LLVM libraries
    pub fn llvm_link() -> anyhow::Result<()> {
        let libdir = std::str::from_utf8(&run("llvm-config", ["--libdir"])?)
            .with_context(|| "`llvm-config --libdir` failed with non-utf8 stdout")?
            .trim()
            .to_string();
        println!("cargo:rustc-link-search={libdir}");

        let ty = match std::str::from_utf8(&run("llvm-config", ["--shared-mode"])?)?.trim() {
            // In dynamic mode, we may need to do something about the rpath...
            "shared" => {
                println!("cargo:rustc-link-arg=-Wl,-rpath={libdir}"); // rpath for executable
                println!("cargo:rustc-env=LD_LIBRARY_PATH={libdir}"); // for tests
                "dylib"
            }

            // Static mode, just include all the .a files. We also need some system libs...
            "static" => {
                println!("cargo:rustc-link-arg=-lstdc++");
                std::str::from_utf8(&run("llvm-config", ["--system-libs"])?)?
                    .trim()
                    .split_ascii_whitespace()
                    .for_each(|lib| println!("cargo:rustc-link-arg={lib}"));
                "static"
            }

            // Err...
            out => bail!("invalid returned value `{out}`"),
        };

        std::str::from_utf8(&run("llvm-config", ["--libs"])?)?
            .trim()
            .split_ascii_whitespace()
            .flat_map(|name| name.strip_prefix("-l"))
            .for_each(|name| println!("cargo:rustc-link-lib={ty}={name}"));

        println!("cargo:rerun-if-changed=build.rs");
        Ok(())
    }
}

#[cfg(feature = "sys")]
pub mod sys {
    pub use llvm_sys::{
        analysis::*,
        core::*,
        error::*,
        orc2::{ee::*, lljit::*, *},
        prelude::*,
        target::*,
        target_machine::*,
        transforms::pass_builder::*,
        *,
    };
}

#[cfg(feature = "bindings")]
pub mod prelude {
    pub use crate::bindings::{basic_block::*, builder::*, context::*, function::*, module::*, types::*, value::*};
}

#[cfg(feature = "bindings")]
mod bindings;

#[cfg(feature = "init")]
pub mod init;

#[cfg(feature = "bindings")]
mod iter;
#[cfg(feature = "bindings")]
pub use iter::*;
