//! Option table passed to the Vivy program at runtime.

use crate::{ast::Expression, ShortString};
use serde::{Deserialize, Serialize};

/// Options for a program, represent the thing that is red from a file.
///
/// The syntax was a bit modified:
/// ```ini
/// [module]
/// option = some Lua/VivyScript expression
/// ```
#[derive(Default, Serialize, Deserialize)]
pub struct OptionTable {
    modules: Vec<(ShortString, Vec<(ShortString, Expression)>)>,
}

pub(crate) struct OptionTableSectionMut<'a> {
    options: &'a mut Vec<(ShortString, Expression)>,
}

impl OptionTable {
    /// Create a new empty default option table.
    pub fn new() -> Self {
        Default::default()
    }

    /// Get the number of options in the table
    pub fn len(&self) -> usize {
        self.modules.iter().map(|(_, section)| section.len()).sum()
    }

    /// Tells whever we have options in the table or not.
    pub fn is_empty(&self) -> bool {
        self.modules.iter().all(|(_, section)| section.is_empty())
    }

    /// Get the asked option if it exists in this table.
    pub fn get(&self, module: impl AsRef<str>, option: impl AsRef<str>) -> Option<&Expression> {
        self.modules
            .iter()
            .find_map(|(name, section)| (name.as_str() == module.as_ref()).then_some(section))?
            .iter()
            .find_map(|(name, value)| (name.as_str() == option.as_ref()).then_some(value))
    }

    /// Tells whether we have a said section or not in the option table.
    pub(crate) fn has_section(&self, section: impl AsRef<str>) -> bool {
        self.modules.iter().any(|(m, _)| m.as_str() == section.as_ref())
    }

    /// Get a section of the table in mutable access. If the section doesn't exist it will be
    /// created and will be empty.
    pub(crate) fn section_mut(&mut self, section: impl Into<ShortString>) -> OptionTableSectionMut<'_> {
        let section = section.into();
        let mut sections = self.modules.iter().enumerate();
        let idx = sections
            .find_map(|(idx, (module, _))| (*module == section).then_some(idx))
            .unwrap_or_else(|| {
                self.modules.push((section, vec![]));
                self.modules.len() - 1
            });
        OptionTableSectionMut::new(&mut self.modules[idx].1)
    }
}

impl<'a> OptionTableSectionMut<'a> {
    fn new(options: &'a mut Vec<(ShortString, Expression)>) -> Self {
        Self { options }
    }
}

impl<'a> Iterator for OptionTableSectionMut<'a> {
    type Item = (ShortString, Expression);

    fn next(&mut self) -> Option<Self::Item> {
        self.options.pop()
    }
}

impl<'a> Extend<(ShortString, Expression)> for OptionTableSectionMut<'a> {
    fn extend<T: IntoIterator<Item = (ShortString, Expression)>>(&mut self, iter: T) {
        self.options.extend(iter)
    }
}

impl Iterator for OptionTable {
    type Item = (ShortString, ShortString, Expression);

    fn next(&mut self) -> Option<Self::Item> {
        let (module, mut section) = self
            .modules
            .iter_mut()
            .find(|(_, opts)| !opts.is_empty())
            .map(|(module, section)| (module, OptionTableSectionMut::new(section)))?;
        let (option, expression) = section.next().unwrap();
        Some((module.clone(), option, expression))
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        let count = self.len();
        (count, Some(count))
    }

    fn count(self) -> usize {
        self.len()
    }
}
