use crate::{
    ast::{
        parsers::{parse_option_table, ParserState},
        OptionTable,
    },
    tokenizer::Lexer,
};

/// A produced [`OptionTable`](crate::ast::OptionTable), along with any errors found during
/// parsing. This option table may not be exactly the same as the input code, as reconstruction may
/// have occurred. For more information.
pub struct OptionTableResult {
    option_table: OptionTable,
    errors: Vec<crate::Error>,
}

impl OptionTableResult {
    /// Returns a reference to the [OptionTable] that was parsed. If there were any errors, this
    /// will not be exactly the same, as reconstruction will have occurred.
    pub fn option_table(&self) -> &OptionTable {
        &self.option_table
    }

    /// Consumes the [OptionTable] that was parsed. If there were any errors, this will not be
    /// exactly the same, as reconstruction will have occurred.
    pub fn into_option_table(self) -> OptionTable {
        self.option_table
    }

    /// Returns all errors that occurred during parsing.
    pub fn errors(&self) -> &[crate::Error] {
        &self.errors
    }

    pub(crate) fn parse_fallible(code: &str) -> Self {
        let lexer = Lexer::new(code);
        let mut parser_state = ParserState::new(lexer);
        let option_table = parse_option_table(&mut parser_state).unwrap_or_defaut();
        Self { option_table, errors: parser_state.errors }
    }

    /// Consumes this OptionTableResult, returning the [OptionTable] that was parsed.
    pub fn into_result(self) -> Result<OptionTable, Vec<crate::Error>> {
        self.into()
    }
}

impl From<OptionTableResult> for Result<OptionTable, Vec<crate::Error>> {
    fn from(result: OptionTableResult) -> Self {
        if result.errors.is_empty() {
            Ok(result.option_table)
        } else {
            Err(result.errors)
        }
    }
}
