#[derive(Debug)]
pub enum ParserResult<T> {
    // This doesn't necessarily mean that there were no errors,
    // because this can sometimes be a recovered value.
    Value(T),

    // Couldn't get any sort of value, but the lexer has moved.
    // This should always come with an error.
    LexerMoved,

    NotFound,
}

impl<T> ParserResult<T> {
    pub fn ok(self) -> Option<T> {
        match self {
            ParserResult::Value(value) => Some(value),
            ParserResult::LexerMoved | ParserResult::NotFound => None,
        }
    }

    pub fn ok_or<E>(self, error: E) -> Result<T, E> {
        self.ok().ok_or(error)
    }

    pub fn expect(self, msg: &str) -> T {
        self.ok().expect(msg)
    }

    pub fn unwrap(self) -> T {
        match self {
            ParserResult::Value(value) => value,
            ParserResult::LexerMoved => panic!("unwrap() called when value was LexerMoved"),
            ParserResult::NotFound => panic!("unwrap() called when value was NotFound"),
        }
    }

    pub fn unwrap_or(self, default: T) -> T {
        match self {
            ParserResult::Value(value) => value,
            _ => default,
        }
    }

    pub fn unwrap_or_else(self, cb: impl FnOnce() -> T) -> T {
        match self {
            ParserResult::Value(value) => value,
            _ => cb(),
        }
    }
}

impl<T: Default> ParserResult<T> {
    pub fn unwrap_or_defaut(self) -> T {
        self.unwrap_or_else(Default::default)
    }
}
