use crate::{ast::AstResult, tests::common::run_test_folder, tokenizer};
use codespan_reporting::{
    diagnostic::{Diagnostic, Label},
    files::SimpleFiles,
};
use insta::{assert_snapshot, assert_yaml_snapshot};
use std::fs;

fn process_codespan_display(source_name: &str, source: &str, errors: &[crate::Error]) {
    let mut files = SimpleFiles::new();
    let file_id = files.add(source_name, source);
    let config = codespan_reporting::term::Config::default();
    let mut output = termcolor::NoColor::new(Vec::new());

    for error in errors {
        let range = error.range();
        let diagnostic = Diagnostic::error()
            .with_message(error.error_message())
            .with_code(match error {
                crate::Error::AstError(_) => "ast",
                crate::Error::TokenizerError(_) => "tokenizer",
            })
            .with_labels(vec![Label::primary(file_id, (range.0.bytes())..(range.1.bytes()))]);
        codespan_reporting::term::emit(&mut output, &config, &files, &diagnostic).unwrap();
    }

    assert_snapshot!("error_display", String::from_utf8(output.into_inner()).unwrap());
}

fn run_parser_fail_cases(folder: &str) {
    run_test_folder(folder, |path| {
        let source = fs::read_to_string(path.join("source.lua")).expect("couldn't read source.lua");
        let tokens = tokenizer::Lexer::new(&source).collect().unwrap();
        assert_yaml_snapshot!("tokens", tokens);

        let result = AstResult::parse_fallible(&source);

        if result.errors().is_empty() {
            panic!("fail case passed for {path:?}");
        }

        assert_yaml_snapshot!("errors", result.errors());
        assert_yaml_snapshot!("ast", result.ast());

        process_codespan_display("source.lua", &source, result.errors());

        let ast = result.into_ast().update_positions();
        assert_yaml_snapshot!("ast_to_string", format!("{ast}"));
    });
}

#[test]
fn test_roblox_parser_fail_cases() {
    run_parser_fail_cases("./src/tests/roblox_cases/fail/parser");
}

#[test]
fn test_lua52_parser_fail_cases() {
    run_parser_fail_cases("./src/tests/lua52_cases/fail/parser");
}

#[test]
fn test_lua53_parser_fail_cases() {
    run_parser_fail_cases("./src/tests/lua53_cases/fail/parser");
}

#[test]
fn test_lua54_parser_fail_cases() {
    run_parser_fail_cases("./src/tests/lua54_cases/fail/parser");
}

#[test]
fn test_vivy_parser_fail_cases() {
    run_parser_fail_cases("./src/tests/vivy_cases/fail/parser");
}

#[test]
fn test_option_parser_fail_cases() {
    run_test_folder("./src/tests/option_cases/fail/parser", |path| {
        let source = fs::read_to_string(path.join("source.ini")).expect("couldn't read source.ini");
        let tokens = tokenizer::Lexer::new(&source).collect().unwrap();
        assert_yaml_snapshot!("tokens", tokens);

        let result = crate::ast::OptionTableResult::parse_fallible(&source);
        if result.errors().is_empty() {
            panic!("fail case passed for {path:?}");
        }

        assert_yaml_snapshot!("errors", result.errors());
        assert_yaml_snapshot!("option_table", result.option_table());

        process_codespan_display("source.ini", &source, result.errors());
    });
}
