#![allow(dead_code)]

//! Usefull traits for the vivy parser crate.

use std::borrow::Cow;

/// Extensions for the [Result] enum.
pub trait ResultExtension<T, E> {
    /// Inspect the value part of the result and transform it with a callback.
    fn inspect_mut(self, cb: impl FnOnce(&mut T)) -> Self;

    /// Inspect the error part of the result and transform it with a callback.
    fn inspect_err_mut(self, cb: impl FnOnce(&mut E)) -> Self;
}

impl<T, E> ResultExtension<T, E> for Result<T, E> {
    fn inspect_err_mut(mut self, cb: impl FnOnce(&mut E)) -> Self {
        if let Err(error) = &mut self {
            cb(error);
        }
        self
    }

    fn inspect_mut(mut self, cb: impl FnOnce(&mut T)) -> Self {
        if let Ok(value) = &mut self {
            cb(value);
        }
        self
    }
}

/// Extensions for the [Vec] struct.
pub trait VecExtension<T> {
    /// Push an item at the end of the vector and returns the result.
    fn join(self, value: T) -> Self;

    /// Returns the without the elements as selected by the callback.
    fn without(self, cb: impl FnMut(&T) -> bool) -> Self;
}

impl<T> VecExtension<T> for Vec<T> {
    fn join(mut self, value: T) -> Self {
        self.push(value);
        self
    }

    fn without(mut self, cb: impl FnMut(&T) -> bool) -> Self {
        self.retain(cb);
        self
    }
}

impl<T> VecExtension<T> for std::collections::VecDeque<T> {
    fn join(mut self, value: T) -> Self {
        self.push_back(value);
        self
    }

    fn without(mut self, cb: impl FnMut(&T) -> bool) -> Self {
        self.retain(cb);
        self
    }
}

/// Like [Default], but we get a reference to a static instead.
pub trait DefaultRef {
    fn default_ref() -> &'static Self;
}

/// A string that can be lazy.
pub trait MaybeLazyString {
    fn to_str(self) -> Cow<'static, str>;
}

impl MaybeLazyString for &'static str {
    fn to_str(self) -> Cow<'static, str> {
        Cow::Borrowed(self)
    }
}

impl<F: FnOnce() -> String> MaybeLazyString for F {
    fn to_str(self) -> Cow<'static, str> {
        Cow::Owned(self())
    }
}
