//! Used to create visitors that recurse through [`Ast`](ast::Ast) nodes.

use crate::{
    ast::*,
    private::Sealed,
    tokenizer::{Token, TokenReference},
};

macro_rules! create_visitor {
    (ast: {
        $($visit_name:ident => $ast_type:ident,)+
    }, token: {
        $($visit_token:ident,)+
    }) => {
        /// A trait that implements functions to listen for specific nodes/tokens.
        /// Unlike [`VisitorMut`], nodes/tokens passed are immutable.
        pub trait Visitor {
            /// Visit the nodes of an [`Ast`](crate::ast::Ast)
            fn visit_ast(&mut self, ast: &Ast) where Self: Sized {
                ast.nodes().visit(self);
                ast.eof().visit(self);
            }

            paste::item! {
                $(
                    #[allow(missing_docs)]
                    fn $visit_name(&mut self, _node: &$ast_type) { }
                    #[allow(missing_docs)]
                    fn [<$visit_name _end>](&mut self, _node: &$ast_type) { }
                )+
            }

            $(
                #[allow(missing_docs)]
                fn $visit_token(&mut self, _token: &Token) { }
            )+
        }

        /// A trait that implements functions to listen for specific nodes/tokens.
        /// Unlike [`Visitor`], nodes/tokens passed are mutable.
        pub trait VisitorMut {
            /// Visit the nodes of an [`Ast`](crate::ast::Ast)
            fn visit_ast(&mut self, ast: Ast) -> Ast where Self: Sized {
                // TODO: Visit tokens?
                let eof = ast.eof().to_owned();
                let nodes = ast.nodes.visit_mut(self);

                Ast {
                    nodes,
                    // Everything gets cloned with this visitor, so there's no original tokens
                    eof: self.visit_eof(eof),
                }
            }

            paste::item! { $(
                #[allow(missing_docs)]
                fn $visit_name(&mut self, node: $ast_type) -> $ast_type {
                    node
                }

                #[allow(missing_docs)]
                fn [<$visit_name _end>](&mut self, node: $ast_type) -> $ast_type {
                    node
                }
            )+ }

            $(
                #[allow(missing_docs)]
                fn $visit_token(&mut self, token: Token) -> Token {
                    token
                }
            )+
        }
    };
}

#[doc(hidden)]
pub trait Visit: Sealed {
    fn visit<V: Visitor>(&self, visitor: &mut V);
}

#[doc(hidden)]
pub trait VisitMut: Sealed
where
    Self: Sized,
{
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self;
}

impl<T: Visit> Visit for &T {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        (**self).visit(visitor);
    }
}

impl<T: Visit> Visit for &mut T {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        (**self).visit(visitor);
    }
}

impl<T: Visit> Visit for Vec<T> {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        for item in self {
            item.visit(visitor);
        }
    }
}

impl<T: VisitMut> VisitMut for Vec<T> {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        self.into_iter().map(|item| item.visit_mut(visitor)).collect()
    }
}

impl<T: Visit> Visit for Option<T> {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        if let Some(item) = self {
            item.visit(visitor);
        }
    }
}

impl<T: VisitMut> VisitMut for Option<T> {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        self.map(|item| item.visit_mut(visitor))
    }
}

impl<A: Visit, B: Visit> Visit for (A, B) {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        self.0.visit(visitor);
        self.1.visit(visitor);
    }
}

impl<A: VisitMut, B: VisitMut> VisitMut for (A, B) {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        (self.0.visit_mut(visitor), self.1.visit_mut(visitor))
    }
}

impl<T: Visit> Visit for Box<T> {
    fn visit<V: Visitor>(&self, visitor: &mut V) {
        (**self).visit(visitor);
    }
}

impl<T: VisitMut> VisitMut for Box<T> {
    fn visit_mut<V: VisitorMut>(self, visitor: &mut V) -> Self {
        Box::new((*self).visit_mut(visitor))
    }
}

create_visitor!(ast: {
    visit_anonymous_call => FunctionArgs,
    visit_assignment => Assignment,
    visit_block => Block,
    visit_call => Call,
    visit_contained_span => ContainedSpan,
    visit_do => Do,
    visit_else_if => ElseIf,
    visit_eof => TokenReference,
    visit_expression => Expression,
    visit_field => Field,
    visit_function_args => FunctionArgs,
    visit_function_body => FunctionBody,
    visit_function_call => FunctionCall,
    visit_function_declaration => FunctionDeclaration,
    visit_function_name => FunctionName,
    visit_generic_for => GenericFor,
    visit_if => If,
    visit_index => Index,
    visit_local_assignment => LocalAssignment,
    visit_local_function => LocalFunction,
    visit_last_stmt => LastStmt,
    visit_method_call => MethodCall,
    visit_numeric_for => NumericFor,
    visit_parameter => Parameter,
    visit_prefix => Prefix,
    visit_return => Return,
    visit_repeat => Repeat,
    visit_stmt => Stmt,
    visit_suffix => Suffix,
    visit_table_constructor => TableConstructor,
    visit_token_reference => TokenReference,
    visit_un_op => UnOp,
    visit_var => Var,
    visit_var_expression => VarExpression,
    visit_while => While,
    visit_compound_assignment => CompoundAssignment,
    visit_compound_op => CompoundOp,
    visit_else_if_expression => ElseIfExpression,
    visit_if_expression => IfExpression,

    // Types
    visit_type_argument => TypeArgument,
    visit_type_assertion => TypeAssertion,
    visit_type_declaration => TypeDeclaration,
    visit_type_field => TypeField,
    visit_type_field_key => TypeFieldKey,
    visit_type_info => TypeInfo,
    visit_type_specifier => TypeSpecifier,

    // Lua >=5.2
    visit_attribute => Attribute,

    // Vivy
    visit_main => Main,
    visit_write => Write,
    visit_yield => Yield,
    visit_import => Import,
    visit_call_list => CallList,
    visit_option_decl => OptionDecl,
    visit_call_list_item => CallListItem,
    visit_main_assignement => MainAssignement,
}, token: {
    visit_identifier,
    visit_multi_line_comment,
    visit_number,
    visit_single_line_comment,
    visit_string_literal,
    visit_symbol,
    visit_token,
    visit_whitespace,
    visit_interpolated_string_segment,
});
