//! Contains the main code for the compiler. The compiler parses and do checks on the code, do some
//! optimisations and then returns the [ast::Ast] if no errors where found.

use crate::{
    ast::*,
    tokenizer::TokenReference,
    traits::VecExtension as _,
    vivy::{
        error_report::ErrorReport,
        library::Library,
        passes::*,
        search_path::{SearchPath, DEFAULT_SEARCH_PATH},
        symbol_table::SymbolTable,
        *,
    },
    ShortString,
};
use derive_more::Display;
use hashbrown::{HashMap, HashSet};
use std::{borrow::Cow, cell::RefCell, fmt, fs, mem, path::Path, rc::Rc};
use symbol_table::SymbolTableLocked;

/// Process a program.
pub struct FrontendPipeline<'a> {
    /// Options from a possibly passed file to the compiler.
    options: Option<Rc<OptionTable>>,

    /// The content of the program / main module.
    program: Cow<'a, str>,

    /// List of imported modules.
    imports: HashMap<ShortString, Rc<(Ast, Cow<'static, str>)>>,

    /// List of imported symbol tables.
    symbols: Vec<Rc<SymbolTable<'a, TypeInfo, SymbolTableLocked>>>,

    /// The search path.
    search: &'a SearchPath,

    /// Is the current module being compiled the main module?
    is_main: bool,

    /// The include path, upper in the vector is the modules importing the ones down in the vector.
    process_path: Vec<ShortString>,

    /// The library of available types.
    library: Rc<RefCell<Library>>,

    /// List of user defined passes. Will be run in order after we are done with our own passes.
    passes: &'a [&'a dyn UserFrontendPass],
}

/// Record of all the options defined, unused, setted in the program.
pub struct OptionsRecord {
    /// List of all unused declared options.
    unused_options: Vec<(ShortString, ShortString)>,

    /// All setted options.
    setted_options: Vec<(ShortString, ShortString, Expression)>,

    /// All declared options.
    declared_options: Vec<(ShortString, ShortString, Expression)>,
}

impl fmt::Display for OptionsRecord {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self { setted_options, .. } = self;
        match setted_options.iter().map(|(a, b, _)| a.len() + b.len()).max() {
            None | Some(0) => Ok(()),
            Some(len) => setted_options.iter().try_for_each(|(module, name, expression)| {
                writeln!(f, "{module}.{name: >len$} {expression}", len = len - module.len())
            }),
        }
    }
}

/// Result of the front-end. Don't show that directly to the user, will show [FrontendOutput].
struct CompilerResult {
    /// The [Ast] of the main program's module.
    ast: Ast,

    /// List of all imported modules' [Ast]
    imports: HashMap<ShortString, Rc<(Ast, Cow<'static, str>)>>,

    /// The main program.
    main: MainProgram,

    /// Record of all options in the program.
    options: OptionsRecord,

    /// All the defined types.
    library: Rc<RefCell<Library>>,
}

/// The result of the frontend.
pub struct FrontendOutput {
    /// The [Ast] of the main program's module.
    pub ast: Ast,

    /// List of all imported modules' [Ast]
    pub imports: HashMap<ShortString, Ast>,

    /// The main program.
    pub main: MainProgram,

    /// Record of all options in the program.
    pub options: OptionsRecord,

    /// All the defined types.
    pub library: Library,
}

impl From<CompilerResult> for FrontendOutput {
    fn from(value: CompilerResult) -> Self {
        let CompilerResult { ast, imports, main, options, library } = value;
        let library = RefCell::into_inner(Rc::into_inner(library).expect("unique owner"));
        let imports = HashMap::from_iter(
            imports
                .into_iter()
                .map(|(module, ptr)| (module, Rc::into_inner(ptr).expect("unique owner").0)),
        );
        Self { ast, main, options, imports, library }
    }
}

impl fmt::Display for FrontendOutput {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self { imports, main, options, library, .. } = self;
        write!(f, "imported modules: ")?;
        imports.keys().try_for_each(|module| write!(f, "{module} "))?;
        writeln!(f)?;

        writeln!(f, "{options}")?;
        writeln!(f, "{library}")?;
        writeln!(f, "{main:#?}")
    }
}

/// Tells if the frontend failed or not.
#[derive(Debug, Display, Clone, Copy)]
#[display("frontend error")]
pub struct FrontendError;

impl std::error::Error for FrontendError {}

impl From<()> for FrontendError {
    fn from(_: ()) -> Self {
        Self {}
    }
}

#[derive(Default)]
struct ImportAllResult {
    declared_options: Vec<(ShortString, ShortString, Expression)>,
    unused: Vec<(ShortString, ShortString)>,
}

impl FromIterator<ImportAllResult> for ImportAllResult {
    fn from_iter<T: IntoIterator<Item = ImportAllResult>>(iter: T) -> Self {
        let mut iter = iter.into_iter();
        let Some(ret) = iter.next() else {
            return Default::default();
        };
        iter.fold(ret, |mut ret, item| {
            let Self { mut declared_options, unused } = item;
            ret.declared_options.append(&mut declared_options);
            ret.unused.retain(|option| unused.contains(option));
            ret
        })
    }
}

impl<'a> FrontendPipeline<'a> {
    /// Create a new pipeline to parse, to check and do black magic.
    pub fn new(SourceCode { name, code }: &'a SourceCode) -> Self {
        Self {
            program: Cow::Borrowed(code.as_ref()),
            options: None,
            imports: Default::default(),
            symbols: vec![],
            is_main: true,
            search: &DEFAULT_SEARCH_PATH,
            process_path: vec![ShortString::new(name.as_ref())],
            library: Rc::new(RefCell::new(Library::default())),
            passes: &[],
        }
    }

    /// Add passes to do for each module at the end of mendatory passes.
    pub fn passes(self, passes: &'a [&'a dyn UserFrontendPass]) -> Self {
        Self { passes, ..self }
    }

    /// Get the name of the module we are compiling here. We should always have a name, if we parse
    /// the stdin we still have a name for the module: stdin...
    pub fn name(&self) -> &ShortString {
        self.process_path.last().expect("module name")
    }

    /// Set the option file to use for this instance of the program.
    pub fn options(self, options: Option<impl AsRef<Path>>) -> Result<Self, FrontendError> {
        let Some(options) = options else { return Ok(self) };
        let options = options.as_ref();

        let name = options
            .file_name()
            .ok_or_else(|| log::error!("expected a file with an actual name"))?
            .to_str()
            .ok_or_else(|| log::error!("expected files with utf8 names"))?;

        let code = fs::read_to_string(options)
            .map_err(|err| log::error!("failed to read option file '{}': {err}", options.display()))?;

        Ok(crate::ast::OptionTableResult::parse_fallible(&code)
            .into_result()
            .map(|options| Self { options: Some(Rc::new(options)), ..self })
            .map_err(|errors| ErrorReport::new().add_errors(name, &code, errors).report_and_ignore())?)
    }

    /// Set the [SearchPath] to use for this instance of the program.
    pub fn search(self, search: &'a SearchPath) -> Self {
        Self { search, ..self }
    }

    /// Import all modules from the given list and add them to the current pipeline.
    fn import_all(&mut self, import_list: Vec<TokenReference>) -> Result<ImportAllResult, FrontendError> {
        Iterator::collect(import_list.into_iter().map(|import| {
            let name = import.token_type().as_str();
            if self.imports.contains_key(name) {
                return Ok(ImportAllResult::default());
            } else if self.process_path.iter().any(|path| path.as_str() == name) {
                ErrorReport::error(
                    self.name(),
                    &self.program,
                    AstError::from_parts(import.clone(), "recursive indlude of module"),
                );
                return Err(FrontendError);
            }

            let program = self.search.resolve(name).ok_or_else(|| {
                ErrorReport::error(
                    self.name(),
                    &self.program,
                    AstError::from_parts(import.clone(), "failed to find the module to import"),
                )
            })?;

            let CompilerResult {
                ast, imports, options: OptionsRecord { unused_options, declared_options, .. }, ..
            } = FrontendPipeline {
                process_path: self.process_path.clone().join(ShortString::new(name)),
                imports: mem::take(&mut self.imports), // Will be replaced.
                program: match program {
                    Cow::Borrowed(program) => Cow::Borrowed(program),
                    Cow::Owned(ref program) => Cow::Borrowed(program.as_str()),
                },
                options: self.options.clone(),
                symbols: self.symbols.clone(),
                library: self.library.clone(),
                is_main: false,
                ..*self
            }
            .process()?;

            debug_assert!({
                let prev = HashSet::<&ShortString>::from_iter(self.imports.keys());
                let news = HashSet::<&ShortString>::from_iter(imports.keys());
                (prev.len() <= news.len()) && (prev.intersection(&news).count() == prev.len())
            });
            let _ = mem::replace(&mut self.imports, imports); // Replace with a more complete import list.
            self.imports.insert(ShortString::new(name), Rc::new((ast, program)));

            Ok(ImportAllResult { declared_options, unused: unused_options })
        }))
    }

    /// Process the pipeline. Some returned values may be hidden from the user, or things
    /// transformed a but more for the main result, but here we keep them.
    fn process(mut self) -> Result<CompilerResult, FrontendError> {
        let ast = AstResult::parse_fallible(&self.program).into_result().map_err(|err| {
            let _ = ErrorReport::new().add_errors(self.name(), &self.program, err).report();
        })?;

        let StmtCheckerResult { decl_opts, setd_opts, main, imported } =
            StmtChecker::new(&ast, (self.name().as_str(), self.program.as_ref()))
                .is_main(self.is_main)
                .process()?;
        let ImportAllResult { mut declared_options, unused } = self.import_all(imported)?;

        declared_options.extend(
            decl_opts
                .into_iter()
                .map(|(name, value)| (self.name().clone(), name, value.into())),
        );

        let symbols = SymbolTable::try_from(&ast)
            .map_err(|err| {
                let _ = ErrorReport::new()
                    .add_errors(self.name(), &self.program, [err])
                    .report();
            })?
            .import(self.symbols.iter().map(Rc::as_ref));

        let loc = (self.name().as_str(), self.program.as_ref());
        TypeCollector::new(&ast, loc, &mut self.library.borrow_mut()).process()?;
        TypeChecker::new(&ast, loc, &self.library.borrow(), symbols).process()?;

        let (ast, OptsTransformResult { unused }) = match self.options.clone() {
            None => (ast, Default::default()),
            Some(options) => {
                let (ast, results) = OptsTransform::new(ast, loc, &options).process()?;
                (ast, OptsTransformResult { unused: results.unused.without(|opt| unused.contains(opt)) })
            }
        };

        Ok(CompilerResult {
            ast: self.passes.iter().try_fold(ast, |prog, pass| {
                log::debug!("apply pass {} on module {}", pass.name(), self.name());
                pass.process(prog, self.name(), self.program.as_ref())
            })?,
            imports: self.imports,
            options: OptionsRecord { declared_options, setted_options: setd_opts, unused_options: unused },
            library: self.library.clone(),
            main: main.unwrap_or_default(),
        })
    }

    /// Process the pipeline. Returns the [Ast] or an error. Note that if an error occured, they
    /// will be printed out to the terminal.
    pub fn run(self) -> Result<FrontendOutput, FrontendError> {
        self.process().map(Into::into)
    }
}
