//! Store informations about methods and fields for types.

use crate::{ast::TypeInfo, ShortString};
use derive_more::{IntoIterator, IsVariant};
use hashbrown::HashMap;
use std::{fmt, sync::LazyLock};

/// Tells whether a field is mutable or constant.
#[derive(PartialEq, Eq, Clone, Copy, Debug, IsVariant)]
pub enum FieldMutability {
    /// The field can be mutated by a user.
    Mutable,

    /// The field can't be mutated by a user and can represent an internal variant of the variable.
    Constant,
}

/// Informations about a field of a [TypeRecord].
#[derive(PartialEq, Clone, Copy)]
pub struct TypeRecordField<'a>(pub FieldMutability, pub &'a TypeInfo);

/// Informations about a method of a [TypeRecord].
#[derive(PartialEq, Clone, Copy)]
pub struct TypeRecordMethod<'a>(pub &'a [TypeInfo], pub &'a TypeInfo, pub bool);

impl TypeRecordMethod<'_> {
    /// Tells whether the method is really a method or just a free function.
    pub fn is_method(&self) -> bool {
        let TypeRecordMethod(_, _, is_method) = self;
        *is_method
    }
}

/// A record stores informations about a type, the fields and the methods.
#[derive(Clone)]
pub struct TypeRecord {
    type_info: TypeInfo,
    is_std: bool,
    fields: HashMap<ShortString, (FieldMutability, TypeInfo)>,
    methods: HashMap<ShortString, (Vec<TypeInfo>, TypeInfo)>,
}

static STANDARD_RECORDS: LazyLock<Vec<TypeRecord>> = LazyLock::new(|| {
    macro_rules! records {
        ($( struct $ty: ident $struct: tt impl $methods: tt )*) => { Vec::from_iter([$(records!(@ $ty $struct $methods)),*]) };
        (@ $type: ident $fields: tt { $($(fn $method: ident ($($arg: ident),*) -> $ret: ident;)+)? }) => {
            std::mem::take(records!(@ TypeRecord::new_std(TypeInfo::$type().clone()), $fields)
                $(.with_methods([
                    $((ShortString::new(stringify!($method)), vec![$(TypeInfo::$arg().clone()),*], TypeInfo::$ret().clone())),*
                ]))?
            )
        };
        (@ $a: expr, const $f: ident : $fty: ident ) => { $a.with_field(stringify!($f).into(), FieldMutability::Constant, TypeInfo::$fty().clone()) };
        (@ $a: expr, mut   $f: ident : $fty: ident ) => { $a.with_field(stringify!($f).into(), FieldMutability::Mutable,  TypeInfo::$fty().clone()) };
        (@ $acc: expr, {}) => { $acc };
        (@ $acc: expr, { $v: tt $f: ident : $fty: ident $(, $vs: tt $fs: ident : $ftys: ident)* $(,)? }) => {
            records!(@ records!(@ $acc, $v $f: $fty), { $($vs $fs: $ftys),* })
        };
    }

    records! {
        struct number {} impl {
            fn is_nan()   -> number;
            fn is_zero()  -> number;
            fn is_true()  -> number;
            fn is_false() -> number;
            fn floor()    -> number;
            fn ceil()     -> number;
        }

        struct string {} impl {
            fn new() -> string;
            fn put(char) -> string;
            fn append(string) -> string;

            fn len() -> number;
            fn get(number) -> char;

            fn is_lowercase() -> number;
            fn is_uppercase() -> number;
            fn to_lowercase() -> string;
            fn to_uppercase() -> string;
        }

        struct char {} impl {
            fn is_lowercase() -> number;
            fn is_uppercase() -> number;
            fn to_lowercase() -> char;
            fn to_uppercase() -> char;
        }

        struct table {} impl {
            fn new() -> table;
            fn len() -> number;
            fn get(any) -> any;
            fn set(any, any) -> nil;
        }

        struct any {} impl {
            fn new() -> any;

            fn to_string() -> string;
            fn to_number() -> number;
            fn to_table()  -> table;
        }

        struct line {
            const index: number, // Index of the line in the original document.
            mut   start: number, // Start time of the line.
            mut   end:   number, // End time of the line.
        } impl {
            fn len() -> number;
            fn syllabe(number) -> syllabe;
            fn syllabes()      -> syllabes;
        }

        struct lines {} impl {
            fn len() -> number;
            fn line(number) -> line;
            fn syllabes()   -> syllabes;
        }

        struct syllabes {} impl {
            fn len() -> number;
            fn syllabe(number) -> syllabe;
        }

        struct syllabe {
            const index: number, // Index in the parent line.
            mut   start: number, // Start time of the syllabe.
            mut   end:   number, // End time of the syllabe.
        } impl {
            fn len() -> number;
            fn text() -> string;
            fn get(number) -> char;
        }
    }
});

impl TypeRecord {
    /// Create a new type record.
    pub fn new(type_info: TypeInfo) -> Self {
        Self { type_info, is_std: false, ..Default::default() }
    }

    /// Create a new standard type record.
    fn new_std(type_info: TypeInfo) -> Self {
        Self { type_info, is_std: true, ..Default::default() }
    }

    /// Add or override a field in the record.
    pub fn with_field(&mut self, field: ShortString, mutability: FieldMutability, type_info: TypeInfo) -> &mut Self {
        self.fields.insert(field, (mutability, type_info));
        self
    }

    /// Add or override a method in the record.
    pub fn with_method(&mut self, method: ShortString, arguments: Vec<TypeInfo>, returns: TypeInfo) -> &mut Self {
        self.methods.insert(method, (arguments, returns));
        self
    }

    /// Add or override fields in the record.
    pub fn with_fields(
        &mut self,
        iter: impl IntoIterator<Item = (ShortString, FieldMutability, TypeInfo)>,
    ) -> &mut Self {
        iter.into_iter().fold(self, |s, (k, m, t)| s.with_field(k, m, t))
    }

    /// Add or override methods in the record.
    pub fn with_methods(
        &mut self,
        iter: impl IntoIterator<Item = (ShortString, Vec<TypeInfo>, TypeInfo)>,
    ) -> &mut Self {
        iter.into_iter().fold(self, |s, (k, m, t)| s.with_method(k, m, t))
    }

    /// Get the fields of instances of this type.
    pub fn fields(&self) -> impl Iterator<Item = (&ShortString, TypeRecordField)> {
        self.fields
            .iter()
            .map(|(field, (mutability, type_info))| (field, TypeRecordField(*mutability, type_info)))
    }

    /// Get the methods of instances of this type.
    pub fn methods(&self) -> impl Iterator<Item = (&ShortString, TypeRecordMethod)> {
        self.methods.iter().map(|(method, (args, returns))| {
            (method, TypeRecordMethod(args, returns, args.first().map(|_| todo!()).unwrap_or_default()))
        })
    }

    /// Get a field's informations if it exists, otherwise get [None].
    pub fn field(&self, field: impl AsRef<str>) -> Option<TypeRecordField> {
        self.fields
            .get(field.as_ref())
            .map(|(mutability, type_info)| TypeRecordField(*mutability, type_info))
    }

    /// Get a method's informations if it exists, otherwise get [None].
    pub fn method(&self, method: impl AsRef<str>) -> Option<TypeRecordMethod> {
        self.methods
            .get(method.as_ref())
            .map(|(args, returns)| TypeRecordMethod(args, returns, args.first().map(|_| todo!()).unwrap_or_default()))
    }

    /// Get the [TypeInfo] of this record.
    pub fn type_info(&self) -> &TypeInfo {
        &self.type_info
    }
}

impl Default for TypeRecord {
    fn default() -> Self {
        Self { is_std: true, type_info: Default::default(), fields: Default::default(), methods: Default::default() }
    }
}

impl fmt::Display for TypeRecord {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let this = self.type_info();
        let std_comment = match self.is_std {
            true => "--[[ std! ]]",
            false => "",
        };
        writeln!(f, "struct {std_comment} {this} {{")?;
        self.fields()
            .try_for_each(|(name, TypeRecordField(mutability, type_info))| {
                let mutability = (mutability == FieldMutability::Mutable)
                    .then_some("mut ")
                    .unwrap_or_default();
                writeln!(f, "\t{mutability} {name}: {type_info},")
            })?;
        writeln!(f, "}}")?;

        self.methods()
            .try_for_each(|(name, func @ TypeRecordMethod(args, ret, ..))| {
                write!(f, "function {this}:{name}(")?;
                if func.is_method() && !args.is_empty() {
                    write!(f, "{this},")?;
                } else if func.is_method() {
                    write!(f, "{this}")?;
                }
                args.iter().enumerate().try_for_each(|(idx, arg)| match idx {
                    0 => write!(f, "{arg}"),
                    _ => write!(f, ", {arg}"),
                })?;
                writeln!(f, ") -> {ret};")
            })
    }
}

/// Stores all the knowned [TypeRecord] for a program. Can differentiate between standard type
/// records and users' type records.
#[derive(IntoIterator)]
pub struct Library(Vec<TypeRecord>);

impl Library {
    /// Create a new library.
    pub fn new() -> Self {
        Self::from_iter(STANDARD_RECORDS.iter())
    }

    /// Iterate throu the records.
    pub fn iter(&self) -> impl Iterator<Item = &TypeRecord> {
        self.0.iter()
    }

    /// Iterate throu the records in a mutable way. Not that you can only mutate non-standard
    /// types' record.
    pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut TypeRecord> {
        self.0.iter_mut().filter(|TypeRecord { is_std, .. }| !*is_std)
    }

    /// Tells wether a specific record for a type name is present or not.
    pub fn has(&self, name: impl AsRef<str>) -> bool {
        self.get(name).is_some()
    }

    /// Get the record for a type name, if not present returns a newly created record.
    pub fn get(&self, name: impl AsRef<str>) -> Option<&TypeRecord> {
        self.iter().find(|TypeRecord { type_info, .. }| match type_info {
            TypeInfo::Basic(type_info) => type_info.token_type().as_str() == name.as_ref(),
            _ => false,
        })
    }

    /// Get the record for a type name in a mutable way, if not present returns [None]. If the
    /// record is from standard type, also returns [None].
    pub fn get_mut(&mut self, name: impl AsRef<str>) -> Option<&mut TypeRecord> {
        self.iter_mut().find(|TypeRecord { type_info, .. }| match type_info {
            TypeInfo::Basic(type_info) => type_info.token_type().as_str() == name.as_ref(),
            _ => false,
        })
    }

    /// Set a type record for a type. If the record is already present, replace it. If the already
    /// present record is of a standard type, it won't be replaced.
    pub fn put(&mut self, record: TypeRecord) {
        let find = |(idx, TypeRecord { type_info, is_std, .. }): (_, &_)| {
            (*type_info == record.type_info).then_some((idx, *is_std))
        };
        match self.0.iter().enumerate().find_map(find) {
            Some((_, true)) => {}
            Some((idx, false)) => self.0[idx] = record,
            None => self.0.push(record),
        }
    }
}

impl FromIterator<TypeRecord> for Library {
    fn from_iter<T: IntoIterator<Item = TypeRecord>>(iter: T) -> Self {
        Self(iter.into_iter().collect())
    }
}

impl<'a> FromIterator<&'a TypeRecord> for Library {
    fn from_iter<T: IntoIterator<Item = &'a TypeRecord>>(iter: T) -> Self {
        Self(iter.into_iter().cloned().collect())
    }
}

impl Default for Library {
    fn default() -> Self {
        Self::new()
    }
}

impl fmt::Display for Library {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.iter().try_for_each(|record| write!(f, "{record}"))
    }
}
