//! Main program and utilities to be able to complete the work.

use crate::{ast::Expression, ShortString};
use derive_more::{IntoIterator, IsVariant};

/// The main program to execute in the parsed script.
#[derive(Default, Debug)]
pub struct MainProgram {
    /// The initial variable name of the read ASS file.
    pub initial: ShortString,

    /// Steps to do, in order, to make the program progress.
    pub steps: Vec<MainProgramStep>,

    /// The variables to write at the end of the program.
    pub returns: MainProgramReturns,
}

impl MainProgram {
    /// Set the initial variable name.
    pub fn with_initial(self, initial: ShortString) -> Self {
        Self { initial, ..self }
    }

    /// Set the steps for the main program.
    pub fn with_steps(self, steps: Vec<MainProgramStep>) -> Self {
        Self { steps, ..self }
    }

    /// Set the returned variables
    pub fn with_returns(self, returns: impl IntoIterator<Item = ShortString>) -> Self {
        Self { returns: returns.into_iter().collect(), ..self }
    }
}

/// The returned variables from the main program.
#[derive(Debug, Default, IntoIterator)]
pub struct MainProgramReturns {
    #[into_iterator(owned)]
    variables: Vec<ShortString>,
}

impl FromIterator<ShortString> for MainProgramReturns {
    fn from_iter<T: IntoIterator<Item = ShortString>>(iter: T) -> Self {
        Self { variables: iter.into_iter().collect() }
    }
}

/// A step to do for the main program.
#[derive(Debug)]
pub struct MainProgramStep {
    /// The name of the assignated variable.
    destination: ShortString,

    /// The name of the job with its source module, in this order: `(module, job)`.
    job: (ShortString, ShortString),

    /// The arguments to pass to the job, the variables to read (and possibly rearange), and the
    /// parameters to set.
    call_list: Vec<MainProgramCallListItem>,
}

impl MainProgramStep {
    /// Create a new program step.
    pub fn new(
        destination: ShortString,
        job: (ShortString, ShortString),
        call_list: impl IntoIterator<Item = MainProgramCallListItem>,
    ) -> Self {
        Self { destination, job, call_list: call_list.into_iter().collect() }
    }

    /// Get the name of the variable to write the result of the job into.
    pub fn destination(&self) -> &ShortString {
        &self.destination
    }

    /// Get the name of the source module where to find the job we are trying to call.
    pub fn module(&self) -> &ShortString {
        &self.job.0
    }

    /// Get the name of the job we are trying to call.
    pub fn job(&self) -> &ShortString {
        &self.job.1
    }

    /// Get the arguments of the job.
    pub fn call_list_items(&self) -> impl Iterator<Item = &MainProgramCallListItem> {
        self.call_list.iter()
    }
}

/// An item to pass to a job when we call it.
#[derive(Debug, IsVariant)]
pub enum MainProgramCallListItem {
    /// We add a variable to process.
    Variable(ShortString),

    /// We set a parameter for the job.
    Parameter(ShortString, Expression),
}

impl MainProgramCallListItem {
    /// Treat the call list item as a variable if possible.
    pub fn as_variable(&self) -> Option<&ShortString> {
        match self {
            MainProgramCallListItem::Variable(variable) => Some(variable),
            _ => None,
        }
    }

    /// Treat the call list item as a parameter setting if possible.
    pub fn as_parameter(&self) -> Option<(&ShortString, &Expression)> {
        match self {
            MainProgramCallListItem::Parameter(argument, value) => Some((argument, value)),
            _ => None,
        }
    }
}
