use crate::{
    ast::{Punctuated, TypeInfo, *},
    node::Node,
    tokenizer::{Token, TokenKind, TokenReference, TokenType},
    visitors::Visit,
    vivy::{
        main_program::{MainProgram, MainProgramCallListItem, MainProgramStep},
        passes::declare_pass,
    },
    ShortString,
};

declare_pass! {
    /// The statement checker is here to verify that statements are not used in incorrect positions.
    visit::StmtChecker: 'a {
        in_job: bool,
    }

    opt    { is_main: bool }
    result {
        imported:  Vec<TokenReference>,
        setd_opts: Vec<(ShortString, ShortString, Expression)>,
        decl_opts: Vec<(ShortString, OptionDefaultValue)>,
        main:      Option<MainProgram>,
    }
}

impl crate::visitors::Visitor for StmtChecker<'_> {
    fn visit_ast(&mut self, ast: &Ast) {
        ast.nodes().stmts().for_each(|stmt| match stmt {
            Stmt::Import(import) => self.check_import(import),
            Stmt::Main(main) => self.check_main(main),
            Stmt::Assignment(assignment) => self.check_option_set(assignment),
            Stmt::OptionDecl(declaration) => self.check_option_decl(declaration),

            stmt @ Stmt::FunctionDeclaration(_) | stmt @ Stmt::LocalFunction(_) => stmt.visit(self),
            stmt @ Stmt::JobDeclaration(_) => {
                self.in_job = true;
                stmt.visit(self);
                self.in_job = false;
            }

            Stmt::TypeDeclaration(_) => unimplemented!("custom types are tbd"),

            stmt => self.error(stmt.tokens().next().unwrap(), "unexpected statement"),
        });
        if !ast.eof().is_kind(TokenKind::Eof) {
            self.error(ast.eof(), "expected eof token here");
        }
        self.warn_on_reimport();
    }

    fn visit_yield(&mut self, yields: &Yield) {
        if !self.in_job {
            self.error(yields.token(), "the yield statement can only be used from inside a job")
        }
    }

    fn visit_import(&mut self, node: &Import) {
        self.error(node.import_token(), "the 'import' statement is only valid in top-level position")
    }

    fn visit_option_decl(&mut self, node: &OptionDecl) {
        self.error(node.option_token(), "the 'option' statement is only valid in top-level position")
    }

    fn visit_write(&mut self, node: &Write) {
        self.error(node.write_token(), "the 'write' statement is only valid in the main statement")
    }

    fn visit_main(&mut self, node: &Main) {
        self.error(node.main_token(), "the 'main' statement is only valid in top-level position")
    }
}

impl StmtChecker<'_> {
    fn warn_on_reimport(&mut self) {
        let mut vec: Vec<_> = self
            .imported
            .iter()
            .map(|token| ShortString::new(token.token_type().as_str()))
            .collect();
        vec.sort();

        enum State {
            Empty,
            Import(ShortString),
            AlreadyWarned(ShortString),
        }

        let mut state = State::Empty;
        for current in vec {
            match state {
                // Double import.
                State::Import(last) if last == current => {
                    state = State::AlreadyWarned(current);
                    let token = self
                        .imported
                        .iter()
                        .find(|import| import.token_type().as_str() == last.as_str())
                        .unwrap()
                        .clone();
                    self.warning(token, "already imported");
                }

                // Multi-import, but already warned.
                State::AlreadyWarned(ref last) if *last == current => {}

                // Current changed.
                State::AlreadyWarned(_) | State::Import(_) | State::Empty => state = State::Import(current),
            }
        }
    }

    fn check_option_decl(&mut self, declaration: &OptionDecl) {
        declaration.declarations().for_each(|(n, ty, v)| match n.token_type() {
            TokenType::Identifier { identifier } if self.decl_opts.iter().any(|(opt, _)| opt == identifier) => {
                self.error(n, "redefinition of option");
            }
            TokenType::Identifier { identifier } => self.decl_opts.push((
                identifier.clone(),
                OptionDefaultValue { expression: v.clone(), type_info: ty.map(|ty| ty.type_info().clone()) },
            )),
            token => unreachable!("expected an identifier! {token:?}"),
        })
    }

    fn check_import(&mut self, import: &Import) {
        assert!(import.name().is_kind(TokenKind::StringLiteral), "import name token should be a string literal");
        self.imported.push(import.name().clone());
    }

    fn check_option_set(&mut self, assignation: &Assignment) {
        if !self.is_main.unwrap_or_default() {
            return self.error(
                assignation.tokens().next().expect("assignation"),
                "options can be set only from the main script file",
            );
        }

        let handle = |v: &Var| match v {
            Var::Expression(expression) => {
                let module = match expression.prefix() {
                    Prefix::Name(TokenReference {
                        token: Token { token_type: TokenType::Identifier { identifier }, .. },
                        ..
                    }) => identifier.clone(),
                    prefix => {
                        let token = prefix.tokens().next().unwrap();
                        self.error(token, "invalid module name, expected an identifier");
                        return None;
                    }
                };

                let error_token = match expression.suffixes().next() {
                    Some(Suffix::Index(Index::Dot {
                        name:
                            TokenReference { token: Token { token_type: TokenType::Identifier { identifier }, .. }, .. },
                        ..
                    })) if expression.suffixes().count() == 1 => return Some((module, identifier.clone())),
                    Some(suffix) => suffix.tokens().next().unwrap().clone(),
                    None => expression.tokens().next().unwrap().clone(),
                };
                self.error(error_token, "expected a reference to an option by its name, e.g. `module.option`");
                None
            }

            Var::Name(name) => match name.token_type() {
                TokenType::Identifier { identifier } => Some((ShortString::new(self.name), identifier.clone())),
                _ => unreachable!(),
            },
        };

        let options = Vec::from_iter(assignation.variables().iter().flat_map(handle))
            .into_iter()
            .zip(assignation.expressions().into_iter().cloned())
            .map(|((module, name), val)| (module, name, val));
        self.setd_opts.extend(options);
    }

    fn check_main(&mut self, main: &Main) {
        if !self.is_main.unwrap_or_default() {
            return self.error(main.main_token(), "the main statement must always be in the main script file");
        } else if self.main.is_some() {
            return self.error(main.main_token(), "multiple main found");
        }

        let mut declared_variables = vec![main.initial_variable().clone()];
        let mut program_steps = Vec::with_capacity(main.assignements().len());
        for step in main.assignements() {
            let call_list = step.call_list_items().flat_map(|item| match item {
                CallListItem::Variable(var) => declared_variables
                    .contains(var)
                    .then(|| MainProgramCallListItem::Variable(var.token_type().as_str().into()))
                    .ok_or_else(|| self.error(var, "use of undeclared variable")),

                CallListItem::SetParameter { parameter, expression, .. } => {
                    let parameter = parameter.token_type().as_str();
                    log::error!("check if the job has the parameter {parameter}");
                    Ok(MainProgramCallListItem::Parameter(ShortString::new(parameter), expression.clone()))
                }
            });

            if !declared_variables.contains(step.destination()) {
                self.error(step.destination(), "redefined variable");
                continue;
            }

            let (module, job) = step.called_job_identifiers();
            log::error!("check if the module and the job are correct…");

            program_steps.push(MainProgramStep::new(
                step.destination().token_type().as_str().into(),
                (module.clone(), job.clone()),
                call_list,
            ));
            declared_variables.push(step.destination().clone());
        }

        let written_variables = main
            .returns()
            .map(|write| write.variables())
            .into_iter()
            .flat_map(Punctuated::iter);
        let written_variables = written_variables.flat_map(|written| {
            declared_variables
                .contains(written)
                .then(|| ShortString::new(written.token_type().as_str()))
                .ok_or_else(|| self.error(written, "try to write a variable that was not declared"))
        });

        self.main = Some(
            MainProgram::default()
                .with_initial(ShortString::new(main.initial_variable().token_type().as_str()))
                .with_steps(program_steps)
                .with_returns(written_variables),
        );
    }
}

pub struct OptionDefaultValue {
    expression: Expression,
    type_info: Option<TypeInfo>,
}

impl From<OptionDefaultValue> for Expression {
    fn from(value: OptionDefaultValue) -> Self {
        value.expression
    }
}

impl OptionDefaultValue {
    /// Get the default value for the option.
    pub fn expression(&self) -> &Expression {
        &self.expression
    }

    /// Get the type info if specified
    pub fn type_info(&self) -> Option<&TypeInfo> {
        self.type_info.as_ref()
    }
}
