use crate::{
    ast::*,
    node::Node,
    tokenizer::*,
    traits::DefaultRef,
    vivy::{library::*, passes::declare_pass, symbol_table::*},
};

declare_pass! {
    /// The type checker is here to verify type corectness of the program and infer types, etc. To
    /// check the types we will need to do custom things...
    ///
    /// TODO: Verify unique functions at top level, with correct names
    visit::TypeChecker: 'a {}

    aux {
        library: &'a Library,
        symbols: SymbolTable<'a, TypeInfo, SymbolTableLocked>,
    }
}

impl crate::visitors::Visitor for TypeChecker<'_> {
    fn visit_ast(&mut self, ast: &Ast) {
        log::error!("todo: verify unique functions at top level, with correct names");
        for error in Vec::from_iter(ast.nodes().stmts().flat_map(|stmt| match stmt {
            Stmt::JobDeclaration(decl) => self.type_compute().callable(decl.body(), true).err(),
            Stmt::FunctionDeclaration(decl) => self.type_compute().callable(decl.body(), false).err(),
            Stmt::LocalFunction(decl) => self.type_compute().callable(decl.body(), false).err(),
            Stmt::Assignment(assignments) => self.type_compute().assignments(assignments).map(|_| ()).err(),
            Stmt::OptionDecl(_) | Stmt::Import(_) => None, // NOOP / Types already registered...
            Stmt::TypeDeclaration(_) => unimplemented!("custom types are tbd"),
            _ => unreachable!("already checked that they are not here"),
        })) {
            self.report.add_errors(self.name, self.source, [*error]);
        }
        if !ast.eof().is_kind(TokenKind::Eof) {
            self.error(ast.eof(), "expected eof token here");
        }
    }
}

impl TypeChecker<'_> {
    fn type_compute(&self) -> TypeCompute {
        TypeCompute::new(self.library, self.symbols.sub_scope())
    }
}

/// The thing that will actually compute types and returns errors in case of problems.
struct TypeCompute<'a> {
    library: &'a Library,
    symbols: SymbolTable<'a, TypeInfo, SymbolTableMutable>,
    in_loop: bool,
    in_job: bool,
    returns: &'a TypeInfo,
}

type TypeComputeError = Box<AstError>;
type TypeComputeResult<'a> = Result<(TypeInfo, SymbolTable<'a, TypeInfo, SymbolTableMutable>), TypeComputeError>;

macro_rules! box_error {
    ($token: expr, $format: literal                   $(,)?) => { Box::new(AstError::from_parts($token.into(), format!($format            ))) };
    ($token: expr, $format: literal, $($args: expr),+ $(,)?) => { Box::new(AstError::from_parts($token.into(), format!($format, $($args),+))) };
}

fn unsupported<T>(token: &TokenReference, whats: &str) -> Result<T, TypeComputeError> {
    Err(box_error!(token, "{whats} are not supported for now"))
}

impl<'a> TypeCompute<'a> {
    fn new(library: &'a Library, symbols: SymbolTable<'a, TypeInfo, SymbolTableMutable>) -> Self {
        Self { library, symbols, returns: TypeInfo::default_ref(), in_loop: false, in_job: false }
    }

    fn sub(&'a self) -> Self {
        Self { symbols: self.symbols.clone(), ..*self }
    }

    fn set(mut self, symbol: &TokenReference, ty: TypeInfo) -> Result<Self, TypeComputeError> {
        self.symbols.set(symbol.token_type().as_str(), ty).map_err(|var| {
            box_error!(symbol, "{var} was already defined at the root level of the program and is thus a constant")
        })?;
        Ok(self)
    }

    fn in_loop(self) -> Self {
        Self { in_loop: true, ..self }
    }

    fn assignments(self, assignments: &'a Assignment) -> TypeComputeResult<'a> {
        let mut assignments = assignments.variables().iter().zip(assignments.expressions());
        assignments.try_for_each(|(dest, expr)| self.sub().assignation(dest, expr))?;
        Ok((TypeInfo::default(), self.symbols))
    }

    fn check_method_call(
        self,
        record: &TypeRecord,
        method: &MethodCall,
        caller: &TokenReference,
        token: &TokenReference,
    ) -> Result<TypeInfo, TypeComputeError> {
        let method_name = method.name().token_type().as_str();
        let (arguments, returns) = match record
            .method(method_name)
            .ok_or_else(|| box_error!(token, "undefined method '{caller}:{method_name}'"))?
        {
            TypeRecordMethod([this, args @ ..], ret, true) if this == record.type_info() => (args, ret),
            TypeRecordMethod(args, ret, false) => (args, ret),
            TypeRecordMethod([this, ..], _, true) => {
                return Err(box_error!(token, "try to call '{this}:{method_name}' on expression of type '{caller}'"))
            }
            TypeRecordMethod(.., true) => unreachable!(),
        };
        self.check_func_args(arguments, returns, method.args(), token)
    }

    fn check_func_args(
        self,
        arguments: &[TypeInfo],
        returns: &TypeInfo,
        func_args: &FunctionArgs,
        token: &TokenReference,
    ) -> Result<TypeInfo, TypeComputeError> {
        match func_args {
            FunctionArgs::String(_) => match arguments.first() {
                Some(type_info) if arguments.len() == 1 && type_info == TypeInfo::string() => Ok(returns.clone()),
                Some(type_info) if arguments.len() == 1 => Err(box_error!(
                    token,
                    "try to pass a string argument to a function that a variable of type {type_info}"
                )),
                Some(_) => {
                    Err(box_error!(token, "try to pass an argument to a function that takes {}", arguments.len()))
                }
                None => Err(box_error!(token, "try to pass an argument to a function that takes none",)),
            },

            FunctionArgs::TableConstructor(_) if arguments.len() != 1 => Err(box_error!(
                token,
                "try to pass a table as argument to a function that takes {} arguments",
                arguments.len()
            )),
            FunctionArgs::TableConstructor(args) => {
                debug_assert_eq!(arguments.len(), 1);
                todo!("find a way to correctly parse expected types...");
                // We can pass a table because we expect a table, or because we pass named
                // arguments...
            }

            FunctionArgs::Parentheses { arguments: args, .. } => {
                if args.len() != arguments.len() {
                    let (arg_len, func_len) = (args.len(), arguments.len());
                    return Err(box_error!(
                        token,
                        "try to pass {arg_len} arguments to a function that takes {func_len}"
                    ));
                }
                let mut arguments = args.iter().enumerate().zip(arguments.iter());
                match arguments.find_map(|((idx, arg), type_info)| {
                    self.sub()
                        .expr_as_type(type_info, arg)
                        .is_err()
                        .then(|| (idx, arg.tokens().next().unwrap(), type_info))
                }) {
                    Some((n, tok, ty)) => Err(box_error!(tok, "expected argument n°{n} to be of type {ty}")),
                    None => Ok(returns.clone()),
                }
            }
        }
    }

    fn expr(self, expression: &'a Expression) -> TypeComputeResult<'a> {
        match expression {
            Expression::Parentheses { expression, .. } => self.expr(expression),
            Expression::FunctionCall(call) => self.suffixes(call.prefix(), call.suffixes()),

            Expression::Number(_) => Ok((TypeInfo::number().clone(), self.symbols)),
            Expression::String(_) => Ok((TypeInfo::string().clone(), self.symbols)),
            Expression::TypeAssertion { .. } => Ok((TypeInfo::default(), self.symbols)),

            Expression::Var(Var::Expression(var)) => self.suffixes(var.prefix(), var.suffixes()),
            Expression::Var(Var::Name(token)) => {
                let ty = self
                    .symbols
                    .get(token.token_type().as_str())
                    .ok_or_else(|| box_error!(token, "undefined variable"))?;
                Ok((ty.clone(), self.symbols))
            }

            Expression::Symbol(token) => match token.token_type() {
                TokenType::Symbol { symbol: Symbol::Nil } => Ok((TypeInfo::default(), self.symbols)),
                TokenType::Symbol { symbol: Symbol::False | Symbol::True } => {
                    Ok((TypeInfo::number().clone(), self.symbols))
                }
                _ => unreachable!(),
            },

            Expression::UnaryOperator { unop, expression } => {
                use {TypeInfo::*, UnOp::*};
                let (ty, symbols) = self.expr(expression)?;
                let is_container =
                    matches!(ty, Array { .. } | Basic(_) | Optional { .. } | Table { .. } | Tuple { .. });
                match unop {
                    Minus(_) | Not(_) | Tilde(_) if ty == *TypeInfo::number() => Ok((ty, symbols)),
                    Hash(_) if is_container => Ok((TypeInfo::number().clone(), symbols)),
                    _ => {
                        Err(box_error!(unop.token(), "can't apply this unary operator to an expression of type '{ty}'"))
                    }
                }
            }

            Expression::BinaryOperator { lhs, binop, rhs } => {
                use BinOp::*;
                let ty = self.sub().expr(lhs)?.0;
                self.sub().expr_as_type(&ty, rhs)?;
                match binop {
                    TwoEqual(_) | TildeEqual(_) => Ok((TypeInfo::number().clone(), self.symbols)),
                    TwoDots(_) if ty == *TypeInfo::string() => Ok((TypeInfo::number().clone(), self.symbols)),

                    GreaterThan(_) | GreaterThanEqual(_) | LessThan(_) | LessThanEqual(_) | DoubleGreaterThan(_)
                    | DoubleLesserThan(_) | Caret(_) | Pipe(_) | Ampersand(_) | Tilde(_) | Percent(_) | Star(_)
                    | Slash(_) | DoubleSlash(_) | Minus(_) | Plus(_) | And(_) | Or(_)
                        if ty == *TypeInfo::number() =>
                    {
                        Ok((ty, self.symbols))
                    }

                    _ => Err(box_error!(
                        binop.token(),
                        "can't apply this binary operator to an expression of type '{ty}'",
                    )),
                }
            }

            Expression::IfExpression(_) => todo!(),
            Expression::TableConstructor(_) => todo!(),

            Expression::Function(_) => unsupported(expression.tokens().next().unwrap(), "lambdas"),
        }
    }

    fn suffixes(self, prefix: &'a Prefix, suffixes: impl Iterator<Item = &'a Suffix> + 'a) -> TypeComputeResult<'a> {
        let prefix = match prefix {
            Prefix::Expression(expression) => self.sub().expr(expression)?.0,
            Prefix::Name(token) => self
                .symbols
                .get(token.token_type().as_str())
                .cloned()
                .ok_or_else(|| box_error!(token, "try to reference an undefined variable or function"))?,
        };

        let ty = suffixes.into_iter().try_fold(prefix, |prefix, suffix| match prefix {
            TypeInfo::Callback { arguments, return_type, .. } => match suffix {
                Suffix::Call(Call::AnonymousCall(call)) => {
                    let arguments =
                        Vec::from_iter(arguments.into_iter().map(|TypeArgument { type_info, .. }| type_info));
                    self.sub()
                        .check_func_args(&arguments, &return_type, call, suffix.tokens().next().unwrap())
                }
                suffix => unreachable!("{suffix:?}"),
            },

            TypeInfo::Array { type_info, .. } => match suffix {
                Suffix::Index(Index::Dot { name, .. }) => match name.token_type().as_str() {
                    "first" | "last" => Ok(*type_info),
                    field => Err(box_error!(name, "unknown field '{field}' for array")),
                },
                Suffix::Index(Index::Brackets { expression, .. }) => self
                    .sub()
                    .expr_as_type(TypeInfo::number(), expression)
                    .map(|_| *type_info),
                _ => unreachable!(),
            },

            TypeInfo::Basic(type_name) => {
                let suffix_token = suffix.tokens().next().unwrap();
                let record = self
                    .library
                    .get(type_name.token_type().as_str())
                    .ok_or_else(|| box_error!(suffix_token, "try to use undefined type '{type_name}'"))?;
                match suffix {
                    Suffix::Call(Call::MethodCall(method)) => {
                        self.sub().check_method_call(record, method, &type_name, suffix_token)
                    }
                    Suffix::Index(Index::Dot { name, .. }) => {
                        let TypeRecordField(_, field) = record
                            .field(name.token_type().as_str())
                            .ok_or_else(|| box_error!(suffix_token, "undefined field for type '{type_name}'"))?;
                        Ok(field.clone())
                    }
                    Suffix::Index(Index::Brackets { expression, .. }) => todo!("{expression}"),
                    suffix => unreachable!("{suffix}"),
                }
            }

            TypeInfo::Tuple { types, .. } => {
                let index = match suffix {
                    Suffix::Index(Index::Dot { name: token, .. })
                    | Suffix::Index(Index::Brackets { expression: Expression::Number(token), .. }) => token,
                    _ => unreachable!(),
                };
                let handle = |str: &str| {
                    str.parse::<usize>().map_err(|err| {
                        box_error!(index, "index into a tuple must be a known integer at compilation time: {err}",)
                    })
                };
                let ty = types.iter().nth(match index.token_type() {
                    TokenType::Identifier { identifier } => match identifier.as_str() {
                        "zero" | "first" => 0,
                        "one" | "second" => 1,
                        "two" => 2,
                        idx => return Err(box_error!(index, "unknown tuple index: '{idx}'")),
                    },
                    TokenType::Number { text } => handle(text.as_str())?,
                    idx @ TokenType::StringLiteral { .. } => handle(idx.as_str())?,
                    _ => todo!(),
                });
                ty.cloned()
                    .ok_or_else(|| box_error!(index, "out of range in tuple: ({types})"))
            }

            TypeInfo::Table { fields, .. } => {
                let field = match suffix {
                    Suffix::Index(Index::Brackets { expression: Expression::String(name), .. })
                    | Suffix::Index(Index::Dot { name, .. }) => name.token_type().as_str(),
                    _ => unreachable!(),
                };
                let field = fields.iter().find_map(|TypeField { key, value, .. }| {
                    let TypeFieldKey::Name(key) = key else { unreachable!() };
                    let key = key.token_type().as_str();
                    (field == key).then_some(value)
                });
                field.cloned().ok_or_else(|| {
                    box_error!(suffix.tokens().next().unwrap(), "failed to find field in {{ {fields} }}")
                })
            }

            TypeInfo::Optional { base, .. } => match suffix {
                Suffix::Call(Call::MethodCall(call)) => match call.args() {
                    FunctionArgs::Parentheses { arguments, .. } if arguments.is_empty() => {
                        match call.name().token_type().as_str() {
                            "is_nil" | "is_value" => Ok(TypeInfo::number().clone()),
                            "value" => Ok(*base),
                            m => Err(box_error!(suffix.tokens().next().unwrap(), "unknown method '{base}::{m}'")),
                        }
                    }
                    _ => Err(box_error!(
                        suffix.tokens().next().unwrap(),
                        "can't pass parguments to '{base}::{}'",
                        call.name(),
                    )),
                },
                _ => Err(box_error!(suffix.tokens().next().unwrap(), "incorrect usage of optional value")),
            },

            TypeInfo::Typeof { .. } => Ok(TypeInfo::string().clone()),
        })?;

        Ok((ty, self.symbols))
    }

    fn expr_as_type(self, as_ty: &TypeInfo, expression: &'a Expression) -> Result<(), TypeComputeError> {
        let (ty, _) = self.sub().expr(expression)?;
        (ty == *as_ty).then_some(()).ok_or_else(|| {
            box_error!(expression.tokens().next().unwrap(), "the expression should be of type '{as_ty}', got: '{ty}'",)
        })
    }

    fn local_assignment(mut self, assignment: &'a LocalAssignment) -> TypeComputeResult<'a> {
        let variables = assignment.names().iter();
        let variables = variables
            .zip(assignment.type_specifiers().map(|ty| ty.map(TypeSpecifier::type_info)))
            .zip(assignment.expressions());
        let variables = variables.map(|((name, ty), val)| {
            let ty = match ty {
                None => self.sub().expr(val)?.0,
                Some(ty) => {
                    self.sub().expr_as_type(ty, val)?;
                    ty.clone()
                }
            };
            Ok::<_, TypeComputeError>((name.token_type().as_str().into(), ty))
        });
        self.symbols
            .extend(variables.collect::<Result<Vec<_>, TypeComputeError>>()?)
            .map_err(|variable| {
                box_error!(assignment.local_token(), "can't declare {variable} because it shadows a global variable",)
            })?;
        Ok((TypeInfo::default(), self.symbols))
    }

    fn compound_assignment(self, assignment: &'a CompoundAssignment) -> TypeComputeResult<'a> {
        todo!()
    }

    fn yields(self, yields: &'a Yield) -> TypeComputeResult<'a> {
        if !self.in_job {
            Err(box_error!(yields.token(), "yielding values is only possible from jobs"))
        } else {
            todo!()
        }
    }

    fn assignation(self, variable: &'a Var, expression: &'a Expression) -> Result<(), TypeComputeError> {
        match variable {
            Var::Expression(_) => todo!(),
            Var::Name(variable) => match variable.token_type() {
                TokenType::Identifier { identifier } if self.symbols.is_global(identifier) => {
                    Err(box_error!(variable, "try to assign a global variable"))
                }
                TokenType::Identifier { identifier } => match self.symbols.get(identifier).cloned() {
                    Some(ty) => self.expr_as_type(&ty, expression),
                    None => Err(box_error!(variable, "try to assign an undefined variable")),
                },
                _ => unreachable!(),
            },
        }
    }

    fn stmt(self, stmt: &'a Stmt) -> TypeComputeResult<'a> {
        match stmt {
            Stmt::CompoundAssignment(assignments) => self.compound_assignment(assignments),
            Stmt::LocalAssignment(assignment) => self.local_assignment(assignment),
            Stmt::Assignment(assignments) => self.assignments(assignments),
            Stmt::FunctionCall(call) => self.suffixes(call.prefix(), call.suffixes()),
            Stmt::Yield(yields) => self.yields(yields),
            Stmt::Do(block) => self.block(block.block()),

            Stmt::If(if_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), if_stmt.condition())?;
                self.sub().block(if_stmt.block())?;
                for elseif in if_stmt.else_if().into_iter().flatten() {
                    self.sub().expr_as_type(TypeInfo::number(), elseif.condition())?;
                    self.sub().block(elseif.block())?;
                }
                if let Some(block) = if_stmt.else_block() {
                    self.sub().block(block)?;
                }
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::Repeat(repeat_stmt) => {
                let symbols = self.sub().block(repeat_stmt.block())?.1;
                TypeCompute { symbols, ..self }
                    .in_loop()
                    .expr_as_type(TypeInfo::number(), repeat_stmt.until())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::While(while_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), while_stmt.condition())?;
                self.sub().in_loop().block(while_stmt.block())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::NumericFor(for_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), for_stmt.start())?;
                self.sub().expr_as_type(TypeInfo::number(), for_stmt.end())?;
                if let Some(step) = for_stmt.step() {
                    self.sub().expr_as_type(TypeInfo::number(), step)?;
                }
                self.sub()
                    .set(for_stmt.index_variable(), TypeInfo::number().clone())?
                    .block(for_stmt.block())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::GenericFor(_) => todo!(),

            Stmt::JobDeclaration(decl) => {
                Err(box_error!(decl.function_token(), "nested declaration of job is forbidden"))
            }

            Stmt::FunctionDeclaration(decl) => unsupported(decl.function_token(), "nested declaration of functions"),
            Stmt::LocalFunction(decl) => unsupported(decl.function_token(), "nested declaration of functions"),
            Stmt::TypeDeclaration(_) => unsupported(stmt.tokens().next().unwrap(), "custom types"),

            Stmt::Import(_) | Stmt::Write(_) | Stmt::Main(_) | Stmt::OptionDecl(_) => {
                unreachable!("should have been already checked / can't be inside function and job declarations")
            }
        }
    }

    fn last_stmt(self, stmt: &'a LastStmt) -> TypeComputeResult<'a> {
        match stmt {
            LastStmt::Break(token) | LastStmt::Continue(token) => match self.in_loop {
                true => Ok((TypeInfo::default(), self.symbols)),
                false => Err(box_error!(token, "invalid statement, not inside a loop")),
            },

            LastStmt::Return(returns) if returns.returns().len() >= 2 => {
                unsupported(returns.token(), "tuples to return multiple things")
            }
            LastStmt::Return(returns) if returns.returns().len() == 1 => (self.returns != TypeInfo::default_ref())
                .then(|| self.expr(returns.returns().iter().next().unwrap()))
                .ok_or_else(|| {
                    box_error!(returns.token(), "function or job returns something, but this return statement is empty")
                })?,
            LastStmt::Return(returns) if returns.returns().len() == 1 => (self.returns != TypeInfo::default_ref())
                .then(|| (TypeInfo::default(), self.symbols))
                .ok_or_else(|| {
                    box_error!(
                        returns.token(),
                        "expected to return an expression of type {}, but returns nothing here",
                        self.returns,
                    )
                }),
            LastStmt::Return(_) => unreachable!(),
        }
    }

    fn block(self, block: &'a Block) -> TypeComputeResult<'a> {
        let symbols = block.stmts().try_fold(self.symbols.clone(), |symbols, stmt| {
            Ok::<_, TypeComputeError>(TypeCompute { symbols, ..self }.stmt(stmt)?.1)
        })?;
        Ok(match block.last_stmt() {
            Some(stmt) => (TypeCompute { symbols, ..self }.last_stmt(stmt)?.0, self.symbols),
            None => (TypeInfo::default(), self.symbols),
        })
    }

    fn callable(mut self, callable: &'a FunctionBody, is_job: bool) -> TypeComputeResult<'a> {
        self.in_job = is_job;
        self.returns = callable
            .return_type()
            .map(|specifier| specifier.type_info())
            .unwrap_or(TypeInfo::default_ref());

        if let Some((idx, parameter)) = callable
            .parameters()
            .iter()
            .zip(callable.type_specifiers().map(|ty| ty.map(|ty| ty.type_info())))
            .enumerate()
            .find_map(|(idx, (parameter, specifier))| match specifier {
                Some(ty) => self
                    .symbols
                    .set(parameter.token_type().as_str(), ty.clone())
                    .is_err()
                    .then_some((idx + 1, parameter)),
                None => Some((idx + 1, parameter)),
            })
        {
            return Err(box_error!(
                parameter.tokens().next().unwrap(),
                "parameter n°{idx} named '{parameter}' has no type specifier or is an ellipse '...' or shadows a global variable",
            ));
        }

        self.block(callable.block())
    }
}
