//! Used to search for other files when using the `import` statement in Vivy Script.

use std::{borrow::Cow, fs, path::PathBuf};

/// Search path, can pass the name of a module and will try to return the associated code.
#[derive(Default)]
pub struct SearchPath {
    folders: Vec<PathBuf>,
}

/// The extension for vivy script files.
pub const VIVY_SCRIPT_EXTENSION: &str = "vvs";

impl SearchPath {
    /// Create a new search path.
    pub const fn new() -> Self {
        Self { folders: Vec::new() }
    }

    /// Resolve the search path to get the desired code.
    pub fn resolve(&self, import: &str) -> Option<Cow<'static, str>> {
        self.folders.iter().find_map(|folder| {
            let mut path = folder.join(import);
            path.set_extension(VIVY_SCRIPT_EXTENSION);
            let code = fs::read_to_string(path).ok()?;
            Some(Cow::Owned(code))
        })
    }

    /// Add a folder to the search list.
    pub fn push(&mut self, folder: PathBuf) {
        self.folders.push(folder)
    }
}

impl Extend<PathBuf> for SearchPath {
    /// Add multiple folders to the search list.
    fn extend<T: IntoIterator<Item = PathBuf>>(&mut self, iter: T) {
        self.folders.extend(iter)
    }
}

impl FromIterator<PathBuf> for SearchPath {
    /// Create a search list from a list of folders.
    fn from_iter<T: IntoIterator<Item = PathBuf>>(iter: T) -> Self {
        Self { folders: iter.into_iter().collect() }
    }
}

/// The default search path, by default we only search the pre-packaged files.
pub(crate) static DEFAULT_SEARCH_PATH: SearchPath = SearchPath::new();
