use crate::{
    ast::*,
    tokenizer::{Symbol, Token, TokenReference, TokenType},
    traits::DefaultRef,
    ShortString,
};
use hashbrown::HashMap;
use std::{
    marker,
    ops::{self, Deref},
};

#[derive(Debug, Clone)]
pub struct SymbolTableLocked;

#[derive(Debug, Clone)]
pub struct SymbolTableMutable;

/// The symbol table stores the available symbols with theyr type or whatever.
#[derive(Clone)]
pub struct SymbolTable<'a, T: Clone, State> {
    /// A possible parent for this scope.
    parent: Option<&'a SymbolTable<'a, T, SymbolTableLocked>>,

    /// Imported scopes.
    imported: Vec<&'a SymbolTable<'a, T, SymbolTableLocked>>,

    /// List of symbols from this scope.
    symbols: HashMap<ShortString, T>,

    /// The state of the table, can be locked or mutable.
    _state: marker::PhantomData<State>,
}

impl<T: Clone, State> Default for SymbolTable<'_, T, State> {
    fn default() -> Self {
        Self { parent: None, symbols: HashMap::new(), imported: vec![], _state: marker::PhantomData }
    }
}

impl<'a, T: Clone> SymbolTable<'a, T, SymbolTableLocked> {
    /// Create a new sub-scoped symbol table.
    pub fn sub_scope(&'a self) -> SymbolTable<'a, T, SymbolTableMutable> {
        SymbolTable { parent: Some(self), symbols: Default::default(), imported: vec![], _state: marker::PhantomData }
    }

    /// Get the root scope.
    pub fn root(&self) -> &SymbolTable<'a, T, SymbolTableLocked> {
        let mut root = self;
        while let Some(parent) = root.parent {
            root = parent
        }
        root
    }
}

impl<'a, T: Clone> SymbolTable<'a, T, SymbolTableMutable> {
    /// Lock the table.
    pub fn lock(self) -> SymbolTable<'a, T, SymbolTableLocked> {
        let Self { parent, imported, symbols, .. } = self;
        SymbolTable::<'a, T, SymbolTableLocked> { parent, imported, symbols, _state: marker::PhantomData }
    }

    /// Add a new variable to the current scope. We override symbols (non-UB shadowing of
    /// variables). Note that we can't shadow or set variables from the root scope as it denotes
    /// the global state. In case of error we return the name of the problematic variable.
    pub fn set(&mut self, symbol: impl Into<ShortString> + AsRef<str>, value: T) -> Result<&mut Self, ShortString> {
        match self.parent {
            Some(parent) if parent.root().get(symbol.as_ref()).is_none() => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            None => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            Some(_) => Err(symbol.into()),
        }
    }

    /// Add variables from an iterator. Same rules as [Self::set].
    pub fn extend(&mut self, iter: impl IntoIterator<Item = (ShortString, T)>) -> Result<&mut Self, ShortString> {
        iter.into_iter()
            .try_fold(self, |this, (symbol, value)| this.set(symbol, value))
    }

    /// Imports other scopes into this one.
    pub fn import(
        mut self,
        iter: impl IntoIterator<Item = &'a SymbolTable<'a, T, SymbolTableLocked>>,
    ) -> SymbolTable<'a, T, SymbolTableLocked> {
        self.imported.extend(iter);
        self.lock()
    }
}

impl<'a, T: Clone, State> SymbolTable<'a, T, State> {
    /// Get a variable/callback/function from this scope or any parent scope.
    pub fn get(&self, symbol: impl AsRef<str>) -> Option<&T> {
        self.symbols
            .get(symbol.as_ref())
            .or_else(|| self.imported.iter().find_map(|scope| scope.get(symbol.as_ref())))
            .or_else(|| self.parent?.get(symbol.as_ref()))
    }

    /// Tells if a variable/callback/function is present in this scope or any parent scope.
    pub fn has(&self, symbol: impl AsRef<str>) -> bool {
        self.get(symbol).is_some()
    }

    /// Tells if a symbol is global or not.
    pub fn is_global(&self, symbol: impl AsRef<str>) -> bool {
        (self.parent.is_none() && self.get(symbol.as_ref()).is_some())
            || self
                .parent
                .map(SymbolTable::root)
                .map_or(false, |root| root.get(symbol).is_some())
    }
}

impl<T: Clone> FromIterator<(ShortString, T)> for SymbolTable<'_, T, SymbolTableMutable> {
    fn from_iter<I: IntoIterator<Item = (ShortString, T)>>(iter: I) -> Self {
        let mut table = Self::default();
        let _ = table.extend(iter);
        table
    }
}

impl<Idx: AsRef<str>, T: Clone + DefaultRef + 'static, State> ops::Index<Idx> for SymbolTable<'_, T, State> {
    type Output = T;

    /// Same as [SymbolTable::get] but we returns a reference to a default value if the variable was not found.
    fn index(&self, index: Idx) -> &Self::Output {
        self.get(index).unwrap_or_else(|| T::default_ref())
    }
}

fn fold_table_on_stmt<'a>(
    symbol: SymbolTable<'a, TypeInfo, SymbolTableMutable>,
    stmt: &'a Stmt,
) -> Result<SymbolTable<'a, TypeInfo, SymbolTableMutable>, Box<AstError>> {
    let handle_decl = |mut symbol: SymbolTable<'a, TypeInfo, SymbolTableMutable>, name: &str, body: &FunctionBody| {
        let (arrow, return_type) = body
            .return_type()
            .map(|type_specifier| (type_specifier.punctuation().clone(), Box::new(type_specifier.type_info().clone())))
            .unwrap_or_else(|| {
                let arrow = Token::new(TokenType::Symbol { symbol: Symbol::ThinArrow });
                (TokenReference::new(vec![], arrow, vec![]), Box::new(TypeInfo::default()))
            });

        let arguments = Punctuated::from_iter(body.parameters().pairs().zip(body.type_specifiers()).map(
            |(param, ty)| {
                let punctuation = ty.map(|ty| ty.punctuation()).cloned().unwrap_or_else(|| {
                    let token = Token::new(crate::tokenizer::TokenType::Symbol { symbol: Symbol::TwoDots });
                    TokenReference::new(vec![], token, vec![])
                });
                let name = Some((param.value().clone().deref().clone(), punctuation));
                let type_info = ty.map(TypeSpecifier::type_info).cloned().unwrap_or_default();
                let type_argument = TypeArgument::new(type_info).with_name(name);
                match param.punctuation() {
                    Some(punctuation) => Pair::Punctuated(type_argument, punctuation.clone()),
                    None => Pair::End(type_argument),
                }
            },
        ));

        let parentheses = body.parameters_parentheses().clone();
        let _ = symbol.set(name, TypeInfo::Callback { parentheses, arguments, arrow, return_type });
        Ok(symbol)
    };

    match stmt {
        // Handled by the type collector
        Stmt::FunctionDeclaration(d) if d.name().method_name().is_some() => Ok(symbol),

        // Function declarations errors
        Stmt::FunctionDeclaration(d) if d.name().names().len() != 1 => todo!("error: invalid name"),
        Stmt::FunctionDeclaration(d) if matches!(d.name().names().first_value(), Some(n) if symbol.has(n.token_type().as_str())) =>
        {
            todo!("error: multiple definitions")
        }
        Stmt::LocalFunction(local) if symbol.has(local.name().token_type().as_str()) => {
            todo!("error: multiple definitions")
        }

        // Function declarations
        Stmt::LocalFunction(local) => handle_decl(symbol, local.name().token_type().as_str(), local.body()),
        Stmt::FunctionDeclaration(decl) => {
            let name = decl.name().names().into_iter().next().unwrap().token_type().as_str();
            handle_decl(symbol, name, decl.body())
        }

        // Options & constants
        Stmt::OptionDecl(_) => todo!(),
        Stmt::LocalAssignment(_) => todo!("handle constants"),

        // Ignore
        _ => Ok(symbol),
    }
}

impl<'a> TryFrom<&'a Ast> for SymbolTable<'a, TypeInfo, SymbolTableMutable> {
    type Error = AstError;

    fn try_from(ast: &'a Ast) -> Result<Self, Self::Error> {
        ast.nodes()
            .stmts()
            .try_fold(Self::default(), fold_table_on_stmt)
            .map_err(|boxed| *boxed)
    }
}
