//! Implement what is needed to make life easier with C-like enums in Rust.

use proc_macro2::*;
use quote::*;

pub(crate) fn impl_derive_enum_variant_count(syn_item: syn::DeriveInput) -> TokenStream {
    let (len, name) = match syn_item.data {
        syn::Data::Enum(enum_item) => (
            enum_item.variants.len(),
            Ident::new(&format!("{}_LENGTH", syn_item.ident.to_string().to_uppercase()), syn_item.ident.span()),
        ),
        _ => panic!("EnumVariantCount only works on Enums"),
    };
    quote! { pub const #name : usize = #len; }
}

pub(crate) fn impl_derive_enum_variant_iter(syn_item: syn::DeriveInput) -> TokenStream {
    let (enum_name, name, content) = match syn_item.data {
        syn::Data::Enum(enum_item) => {
            let variants = enum_item
                .variants
                .into_iter()
                .map(|variant| format!("{}::{}", syn_item.ident, variant.ident));
            let variants = variants.collect::<Vec<_>>().join(", ");
            (
                syn_item.ident.clone(),
                Ident::new(&format!("{}_VALUES", syn_item.ident.to_string().to_uppercase()), syn_item.ident.span()),
                syn::parse_str::<syn::Expr>(&format!("&[ {variants} ]"))
                    .expect("failed generation of enum's variant list"),
            )
        }
        _ => panic!("EnumVariantIter only works on Enums"),
    };
    quote! { pub const #name : &[#enum_name] = #content; }
}

pub(crate) fn impl_derive_enum_variant_from_str(syn_item: syn::DeriveInput) -> TokenStream {
    let enum_item = match syn_item.data {
        syn::Data::Enum(enum_item) => enum_item,
        _ => panic!("EnumVariantFromStr only works on Enums"),
    };
    let name = syn_item.ident;
    let match_branches = enum_item
        .variants
        .into_iter()
        .map(|variant| format!("{:?} => Ok({name}::{}),", variant.ident.to_string().to_lowercase(), variant.ident))
        .chain(Some("_ => Err(format!(\"unknown field '{{s}}'\")),".to_string()))
        .collect::<Vec<_>>()
        .join("");
    let match_branches =
        syn::parse_str::<syn::Expr>(&format!("match s.trim().to_lowercase().as_str() {{ {match_branches} }}"))
            .expect("failed generation of enum's FromStr implementation");
    quote! {
        impl std::str::FromStr for #name {
            type Err = std::string::String;
            fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
                #match_branches
            }

        }
    }
}
