mod enums;
mod vvrt;

use proc_macro::TokenStream;
use quote::quote;
use syn::parse_macro_input;

#[proc_macro_derive(EnumVariantCount)]
pub fn derive_enum_variant_count(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as syn::DeriveInput);
    enums::impl_derive_enum_variant_count(input).into()
}

#[proc_macro_derive(EnumVariantIter)]
pub fn derive_enum_variant_iter(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as syn::DeriveInput);
    enums::impl_derive_enum_variant_iter(input).into()
}

#[proc_macro_derive(EnumVariantFromStr)]
pub fn derive_enum_variant_from_str(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as syn::DeriveInput);
    enums::impl_derive_enum_variant_from_str(input).into()
}

/// Processes everything to create the export rule for LLVM. We will rename the functions
/// accordingly.
///
/// You need to have the `ManageFunctions` and `Properties` in scope to use this macro.
///
/// Re-export what is inside the module for `super`. With the name passed as argument we export the
/// function `{name}_llvm_properties` that will returns the properties for LLVM.
#[proc_macro_attribute]
pub fn vvrt_bridge(arg: TokenStream, input: TokenStream) -> TokenStream {
    let name = parse_macro_input!(arg as syn::Ident);
    let module = parse_macro_input!(input as syn::ItemMod);

    let (module, funcs) = vvrt::impl_vvrt_bridge(name.to_string(), module).unwrap();
    let module_name = &module.ident;

    let has_clone = funcs.iter().any(|(_, role, ..)| {
        role.as_ref()
            .map(|role| matches!(role, vvrt::Role::Clone))
            .unwrap_or_default()
    });
    let has_drop = funcs.iter().any(|(_, role, ..)| {
        role.as_ref()
            .map(|role| matches!(role, vvrt::Role::Drop))
            .unwrap_or_default()
    });
    if has_clone ^ has_drop {
        panic!("you can't have a drop implementation without the clone one, or the other way around")
    }

    let funcs = vvrt::gen_method_infos(name, funcs).unwrap();

    quote! {
        #module
        pub(super) use self::#module_name::*;
        #funcs
    }
    .into()
}
