use crate::vvrt::{MethodInfo, Role, Type};
use anyhow::bail;
use proc_macro2::TokenStream;
use quote::quote;

fn gen_type_call(ty: &Type, arg: &syn::Ident) -> TokenStream {
    match ty {
        Type::None => quote! { LLVMVoidTypeInContext(#arg) },
        Type::Integer => quote! { LLVMInt32TypeInContext(#arg) },
        Type::Floating => quote! { LLVMFloatTypeInContext(#arg) },
        Type::Boolean => quote! { LLVMInt1TypeInContext(#arg) },
        Type::StrPtr => quote! { LLVMPointerTypeInContext(#arg, 0) },
        Type::VVRT(vvrt) => {
            let vvrt = syn::Ident::new(vvrt, arg.span());
            quote! { #vvrt::llvm_type(#arg) }
        }
    }
}

fn get_asttype(ty: &Type) -> TokenStream {
    match ty {
        Type::None => quote! { ASTType::Nil },
        Type::Integer => quote! { ASTType::Integer },
        Type::Floating => quote! { ASTType::Floating },
        Type::Boolean => quote! { ASTType::Boolean },
        Type::StrPtr => quote! { ASTType::String },
        Type::VVRT("VVRTString") => quote! { ASTType::String },
        Type::VVRT("VVRTTable") => quote! { ASTType::AnyTable },
        Type::VVRT("VVRTSeq") => quote! { ASTType::AnySequence },
        Type::VVRT("VVRTAny") => quote! { ASTType::Any },
        Type::VVRT("VVRTVariant") => quote! { ASTType::Variant },
        Type::VVRT("VVRTType") => quote! { ASTType::Integer },
        Type::VVRT("VVRTLine") => quote! { ASTType::Line },
        Type::VVRT("VVRTSyllabe") => quote! { ASTType::Syllabe },
        ty => unreachable!("unknown type {ty:?}"),
    }
}

pub(crate) fn gen_method_infos(name: syn::Ident, infos: Vec<MethodInfo>) -> anyhow::Result<TokenStream> {
    macro_rules! get_manage {
        ($role: ident) => {
            infos
                .iter()
                .find_map(|(name, role, ..)| matches!(role, Some(Role::$role)).then_some(name.as_str()))
        };
    }

    macro_rules! opt_string {
        ($expr: expr) => {
            match $expr {
                None => quote! { None },
                Some(str) => quote! { Some(#str) },
            }
        };
    }

    macro_rules! quote_vec {
        ($inner: expr) => {
            match $inner {
                Some(inner) => quote! { vec![#inner] },
                None => quote! { vec![] },
            }
        };
    }

    let arg = syn::Ident::new("c", name.span());

    let manage_len = opt_string!(get_manage!(Len));
    let manage_eq = opt_string!(get_manage!(Eq));

    let manage = match (get_manage!(Drop), get_manage!(Clone)) {
        (None, None) => quote! { None },
        (Some(drop), Some(clone)) => quote! {
            Some(ManageFunctions {
                clone: #clone,
                drop: #drop,
            })
        },
        _ => bail!("clone and drop functions must be present at the same time, or none of them"),
    };

    let methods = infos.into_iter().map(|(name, _, args, returns)| {
        let llvm_returns = gen_type_call(&returns, &arg);
        let llvm_args_len: u32 = args.len().try_into().expect("too many arguments");
        let llvm_args = quote_vec!(args
            .iter()
            .map(|ty| gen_type_call(ty, &arg))
            .reduce(|acc, item| quote! { #acc, #item }));

        let ast_returns = get_asttype(&returns);
        let ast_args = quote_vec!(args
            .into_iter()
            .map(|ty| get_asttype(&ty))
            .reduce(|acc, item| quote! { #acc, #item }));

        let func = syn::Ident::new(&name, arg.span());

        quote! {{
            let func: unsafe extern "C" fn() = unsafe { std::mem::transmute(#func as *const ()) };
            let mut llvm_args = #llvm_args; // We compute the vector here for safety.
            let ast_type = ASTType::Function(#ast_args, Box::new(#ast_returns)); // Also needs the ast type.
            let functy = vvs_llvm::sys::LLVMFunctionType(
                #llvm_returns,                          // Should be Ok there.
                llvm_args.as_mut_ptr(), #llvm_args_len, // We pre-computed the length here.
                0                                       // No variadic functions there.
            );
            (#name, (func, functy, ast_type))
        }}
    });
    let methods = quote_vec!(methods.reduce(|acc, item| quote! { #acc, #item }));

    Ok(quote! {
        impl LLVMExported for #name {
            unsafe fn llvm_type(c: LLVMContextRef) -> LLVMTypeRef {
                #[allow(dead_code)]
                static SIZE_IS_VALID: () = {
                    assert!(std::mem::size_of::<#name>() == 8);
                    assert!(std::mem::align_of::<#name>() == 8);
                };
                LLVMPointerTypeInContext(c, 0)
            }

            unsafe fn llvm_properties(c: LLVMContextRef) -> Properties {
                use vvs_lang::ast::ASTType;
                Properties {
                    manage: #manage,
                    equality: #manage_eq,
                    len: #manage_len,
                    methods: hashbrown::HashMap::from_iter(#methods),
                }
            }
        }
    })
}
