//! Utility macros used in the Vivy Script Runtime.

mod gen;
mod structs;

pub(crate) use self::{gen::*, structs::*};

use anyhow::anyhow;
use proc_macro2::{Ident, TokenStream, TokenTree};
use quote::quote;
use quote::ToTokens;
use syn::{
    token::{Pub, Unsafe},
    Generics, Visibility,
};

fn handle_function(name: &str, func: &mut syn::ItemFn) -> anyhow::Result<MethodInfo> {
    func.vis = Visibility::Public(Pub::default());
    func.sig.asyncness = None;
    func.sig.unsafety = Some(Unsafe::default());
    func.sig.constness = None;
    func.sig.variadic = None;
    func.sig.generics = Generics::default();

    let func_name = func.sig.ident.to_string().to_lowercase();
    let role = Role::try_from(func_name.as_str()).ok();
    let func_name = format!("{name}_{}", func_name);
    func.sig.ident = Ident::new(&func_name, func.sig.ident.span());

    let args = func.sig.inputs.iter().map(|inputs| {
        Type::try_from(
            syn::parse2::<syn::Type>(
                inputs
                    .to_token_stream()
                    .into_iter()
                    .skip_while(|tok| !matches!(tok, TokenTree::Punct(p) if p.as_char() == ':'))
                    .skip(1)
                    .collect::<TokenStream>(),
            )
            .unwrap(),
        )
    });
    let args = args.collect::<Result<Vec<_>, _>>()?;

    // Skip the arrow
    let returns = if func.sig.output.to_token_stream().into_iter().count().eq(&0) {
        Type::None
    } else {
        let returns = func.sig.output.to_token_stream().into_iter().skip(2);
        Type::try_from(syn::parse2::<syn::Type>(returns.collect())?)?
    };

    Ok((func_name, role, args, returns))
}

pub(crate) fn impl_vvrt_bridge(
    name: String,
    mut module: syn::ItemMod,
) -> anyhow::Result<(syn::ItemMod, Vec<MethodInfo>)> {
    if let Some((_, items)) = module.content.as_mut() {
        let funcs = items.iter_mut().map(|item| match item {
            syn::Item::Fn(item) => handle_function(&name, item).map(Some),
            _ => Ok(None),
        });
        let funcs = funcs
            .collect::<Result<Vec<_>, _>>()?
            .into_iter()
            .flatten()
            .collect::<Vec<_>>();
        items.push(syn::parse2(quote! { use super::*; })?);
        Ok((module, funcs))
    } else {
        Err(anyhow!("expected a module"))
    }
}
