use anyhow::anyhow;
use quote::ToTokens;

pub(crate) type MethodInfo = (String, Option<Role>, Vec<Type>, Type);

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub(crate) enum Role {
    Clone,
    Eq,
    Drop,
    Len,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[allow(clippy::upper_case_acronyms)]
pub(crate) enum Type {
    /// Void, None, Null, etc.
    None,

    /// Special type for the `*const u8` type. We only have const pointers here. This is used to
    /// create strings from static slices...
    StrPtr,

    /// Integers, we use 32 bit integers.
    Integer,

    /// Floating numbers, we use 32 bit floats.
    Floating,

    /// Boolean.
    Boolean,

    /// A builtin VVRT type (Vivy Script Runtime)
    VVRT(&'static str),
}

impl TryFrom<&str> for Role {
    type Error = anyhow::Error;
    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value.to_lowercase().as_ref() {
            "clone" => Ok(Self::Clone),
            "eq" => Ok(Self::Eq),
            "drop" => Ok(Self::Drop),
            "len" => Ok(Self::Len),
            str => Err(anyhow!("unknown role `{str}`")),
        }
    }
}

impl TryFrom<syn::Type> for Type {
    type Error = anyhow::Error;
    fn try_from(value: syn::Type) -> Result<Self, Self::Error> {
        match value {
            syn::Type::Path(path) => match &path.path.require_ident()?.to_string()[..] {
                str if str.starts_with("VVRT") => Ok(Type::VVRT(String::leak(str.to_string()))),
                "i32" => Ok(Type::Integer),
                "f32" => Ok(Type::Floating),
                "bool" => Ok(Type::Boolean),
                "()" => Ok(Type::None),
                str => Err(anyhow!("unknown type `{str}`")),
            },
            syn::Type::Ptr(inner) => match *inner.elem {
                syn::Type::Path(ty) if ty.path.require_ident()? == "u8" => Ok(Type::StrPtr),
                ty => Err(anyhow!("expected u8 to form string pointer, got: {}", ty.into_token_stream().to_string())),
            },
            ty => Err(anyhow!("type can't be casted into base_type: {}", ty.into_token_stream().to_string())),
        }
    }
}
