use crate::workers::Workers;
use anyhow::{bail, Context};
use std::{
    ffi::{c_int, c_void, CStr, CString},
    ptr,
    sync::{
        atomic::{AtomicU64, Ordering},
        Arc,
    },
};
use vvs_llvm::sys::*;
use vvs_runtime_types::VVRTSymbol;

#[allow(dead_code)]
pub struct JIT {
    jit: *mut LLVMOrcOpaqueLLJIT,
    es: *mut LLVMOrcOpaqueExecutionSession,

    tsctx: *mut LLVMOrcOpaqueThreadSafeContext,

    irtl: *mut LLVMOrcOpaqueIRTransformLayer,
    objl: *mut LLVMOrcOpaqueObjectLayer,
    dylib: *mut LLVMOrcOpaqueJITDylib,

    workers: Workers,
}

macro_rules! llvm_expect {
    (unsafe $err: expr, $msg: literal $(, $arg: expr)* $(,)?) => {{
        unsafe { llvm_expect!($err, $msg $(, $arg)*) }
    }};

    ($err: expr, $msg: literal $(, $arg: expr)* $(,)?) => {{
        let err = $err;
        if !err.is_null() {
            let msg = std::ffi::CStr::from_ptr(LLVMGetErrorMessage(err))
                .to_string_lossy()
                .to_string();
            LLVMConsumeError(err);
            bail!("{}: {}", format_args!($msg, $(, $arg)*), msg)
        }
    }};
}

/// This function should be visible because it starts with `VVRT`
#[allow(non_snake_case)]
pub extern "C" fn VVRTHelloWorld() {
    println!("Hello World from LLVM JIT");
}

/// This function should not be visible because it doesn't start with `VVRT`
pub extern "C" fn hello_world() {
    println!("wtf !?");
}

extern "C" fn orc_sym_filter(_: *mut c_void, sym: LLVMOrcSymbolStringPoolEntryRef) -> c_int {
    eprintln!("toto");
    let Ok(sym) = unsafe { CStr::from_ptr(LLVMOrcSymbolStringPoolEntryStr(sym)) }.to_str() else {
        log::error!("got a symbol which name was not valid utf8 from orc");
        return 0;
    };
    sym.starts_with("VVRT") as c_int
}

impl JIT {
    /// Create a new JIT thingy to execute Vivy Script code.
    pub fn new() -> anyhow::Result<JIT> {
        if let Err(err) = vvs_llvm::init::initialize_llvm() {
            log::error!("{err}");
        }

        unsafe {
            // Create the jit.
            let mut jit = ptr::null_mut();
            let mut tmb = ptr::null_mut();
            llvm_expect!(LLVMOrcJITTargetMachineBuilderDetectHost(&mut tmb), "failed to create orc jit builder");

            let jit_builder = LLVMOrcCreateLLJITBuilder();
            LLVMOrcLLJITBuilderSetJITTargetMachineBuilder(jit_builder, tmb);
            llvm_expect!(LLVMOrcCreateLLJIT(&mut jit, jit_builder), "failed to create the orc jit");

            log::debug!("got data layout: {}", CStr::from_ptr(LLVMOrcLLJITGetDataLayoutStr(jit)).to_string_lossy());

            // Create things, in this order.
            let es = LLVMOrcLLJITGetExecutionSession(jit);
            let objl = LLVMOrcCreateRTDyldObjectLinkingLayerWithSectionMemoryManager(es);
            let irtl = LLVMOrcLLJITGetIRTransformLayer(jit);
            let dylib = LLVMOrcExecutionSessionCreateBareJITDylib(es, b"<main>\0".as_ptr() as *const _);

            // Allow to search for symbols in the current process, only if they start with `VVRT`.
            let mut dg = std::ptr::null_mut();
            llvm_expect!(
                LLVMOrcCreateDynamicLibrarySearchGeneratorForProcess(
                    &mut dg,
                    LLVMOrcLLJITGetGlobalPrefix(jit),
                    Some(orc_sym_filter),
                    std::ptr::null_mut()
                ),
                "failed to create an orc dylib search generator for the current process"
            );
            LLVMOrcJITDylibAddGenerator(dylib, dg);

            // Ok. Create workers and return the struct.
            let tsctx = LLVMOrcCreateNewThreadSafeContext();
            let workers = Workers::new(|flag: Arc<AtomicU64>| loop {
                match flag.load(Ordering::SeqCst) {
                    0 => std::thread::yield_now(),
                    u64::MAX => return Ok(()),
                    x => bail!("won't handle work package n°{x}"),
                }
            })?;
            Ok(JIT { jit, es, dylib, tsctx, irtl, objl, workers })
        }
    }

    /// Get the context out of the JIT thingy. Used to create a LLVM module to be used latter by
    /// the JIT.
    pub fn ctx(&self) -> LLVMContextRef {
        unsafe { LLVMOrcThreadSafeContextGetContext(self.tsctx) }
    }

    /// Add a new module to the JIT engine.
    ///
    /// # Safety
    /// The passed module ([LLVMModuleRef]) must not be a null pointer and must be a valid LLVM
    /// module with valid code for the runtime. This module should be created by the [vvs_codegen]
    /// crate or you can have problems executing it...
    pub unsafe fn add_module(&self, module: LLVMModuleRef) -> anyhow::Result<()> {
        let tsm = LLVMOrcCreateNewThreadSafeModule(module, self.tsctx);
        let rc = LLVMOrcLLJITAddLLVMIRModule(self.jit, self.dylib, tsm);
        llvm_expect!(rc, "failed to add the module into the jit");
        Ok(())
    }

    /// LookUp a symbol defined in the JIT.
    pub fn lookup<'a>(&self, name: &'a str) -> anyhow::Result<VVRTSymbol<'a>> {
        let mut addr = Default::default();
        let symbol = VVRTSymbol::try_from(name)?;
        let str = CString::new(name).with_context(|| "CString::new failed")?;
        llvm_expect!(unsafe
            LLVMOrcLLJITLookup(self.jit, &mut addr, str.as_ptr()),
            "failed to lookup symbol `{name}`"
        );
        Ok(symbol.with_addr(addr))
    }
}

impl Drop for JIT {
    fn drop(&mut self) {
        self.workers.pre_drop();
        unsafe {
            LLVMOrcDisposeThreadSafeContext(self.tsctx);
            LLVMOrcDisposeLLJIT(self.jit);
            LLVMOrcDisposeObjectLayer(self.objl);
        }
    }
}
