//! A simple implementation of worker threads, unix only for now. HWloc seems to have difficulties
//! on MacOS, so this should be Ok for Linux and BSDs...

use anyhow::{bail, Context, Result};
use std::{
    cell::Cell,
    sync::{
        atomic::{AtomicU64, Ordering},
        Arc,
    },
    thread::JoinHandle,
};

thread_local! {
    pub static WORKER_THREAD_IDX: Cell<u16> = const { Cell::new(u16::MAX) };
}

/// Contains worker threads, we bind one per physical core. We can distribute work to those threads
/// and wait for the distributed work to be done.
pub(crate) struct Workers {
    handles: Vec<(JoinHandle<()>, Arc<AtomicU64>)>,
}

fn join_handles<T>(handles: impl Iterator<Item = JoinHandle<T>>) {
    for handle in handles {
        if let Err(err) = handle.join() {
            let idx = WORKER_THREAD_IDX.get();
            log::error!("failed to join thread n°{idx} with error: {err:?}")
        }
    }
}

fn cancel_threads(handles: impl Iterator<Item = Arc<AtomicU64>>) {
    for atomic in handles {
        atomic.store(u64::MAX, Ordering::SeqCst);
    }
}

impl Workers {
    /// Create new worker threads. We use a closure passed by the user. When work is asked for the
    /// workers the passed integer is set to a non-zero constant (you can't assume it's value),
    /// when the thread has finish it's work the atomic must be set to 0.
    ///
    /// If we want to cancel the thread, the passed atomic will be set to `U64_MAX`.
    pub fn new(user_func: impl Fn(Arc<AtomicU64>) -> Result<()> + Send + Clone + 'static) -> Result<Self> {
        let thread_count = std::thread::available_parallelism()
            .map(std::num::NonZeroUsize::get)
            .unwrap_or(1);
        (0..thread_count)
            .map(|idx| Self::launch_thread(user_func.clone(), idx))
            .try_fold(vec![], |mut acc, item| match item {
                Err(err) => Err((acc, err)),
                Ok(item) => {
                    acc.push(item);
                    Ok(acc)
                }
            })
            .map(|handles| Self { handles })
            .map_err(|(handles, err)| {
                cancel_threads(handles.iter().map(|(_, rc)| rc.clone()));
                join_handles(handles.into_iter().map(|(hdl, _)| hdl));
                err
            })
    }

    /// Launch a thread with a user function on a core of the topology.
    fn launch_thread(
        user_func: impl Fn(Arc<AtomicU64>) -> Result<()> + Send + 'static,
        idx: usize,
    ) -> Result<(JoinHandle<()>, Arc<AtomicU64>)> {
        let flag = Arc::new(AtomicU64::new(0));
        let ret_flag = flag.clone();

        let idx_i16: u16 = idx.try_into().with_context(|| "try to launch to many threads")?;
        if idx_i16 == u16::MAX {
            bail!("try to launch to many threads");
        }

        let handle = std::thread::Builder::new()
            .name(format!("VVRT-Worker-{idx}"))
            .spawn(move || {
                WORKER_THREAD_IDX.set(idx_i16);
                log::debug!("start thread n°{idx}");
                if let Err(err) = user_func(flag.clone()) {
                    log::error!("thread n°{idx} exits with error: {err}");
                }
                flag.store(0, Ordering::SeqCst);
                log::debug!("exit thread n°{idx}");
            })?;

        Ok((handle, ret_flag))
    }

    /// Pre-drop all the threads and cancel the work to do.
    pub fn pre_drop(&mut self) {
        let handles = std::mem::take(&mut self.handles);
        cancel_threads(handles.iter().map(|(_, rc)| rc.clone()));
        join_handles(handles.into_iter().map(|(hdl, _)| hdl));
    }
}

impl Drop for Workers {
    fn drop(&mut self) {
        self.pre_drop();
    }
}
