use anyhow::{anyhow, bail};
use std::{ffi::CString, str::FromStr};
use vvs_lang::ast::ASTType;
use vvs_llvm::sys::*;

/// Materialize which type of symbol we have.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum VVRTSymbolType {
    Constant,
    Function,
    Option,
    Method,
}

/// Describes a symbol. Can specify an ORCv2 execution address.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct VVRTSymbol<'a> {
    pub sym: VVRTSymbolType,
    pub module: &'a str,
    pub name: &'a str,
    addr: Option<LLVMOrcExecutorAddress>,
}

impl VVRTSymbolType {
    /// Get the string representation of the symbol type.
    pub fn as_str(&self) -> &'static str {
        match self {
            VVRTSymbolType::Constant => "Constant",
            VVRTSymbolType::Function => "Function",
            VVRTSymbolType::Option => "Option",
            VVRTSymbolType::Method => "Method",
        }
    }
}

impl FromStr for VVRTSymbolType {
    type Err = anyhow::Error;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.trim() {
            "Constant" => Ok(VVRTSymbolType::Constant),
            "Function" => Ok(VVRTSymbolType::Function),
            "Option" => Ok(VVRTSymbolType::Option),
            "Method" => Ok(VVRTSymbolType::Method),
            s => Err(anyhow!("unknown symbol type `{s}`")),
        }
    }
}

impl std::fmt::Display for VVRTSymbolType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_str())
    }
}

impl AsRef<str> for VVRTSymbolType {
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl<'a> VVRTSymbol<'a> {
    /// Create a new symbol.
    pub fn new(ty: VVRTSymbolType, module: &'a str, name: &'a str) -> anyhow::Result<Self> {
        if let Some(idx) = module.find(|c: char| !c.is_alphanumeric()) {
            bail!("invalid character found in module name at position `{idx}`: {module}")
        }
        if let Some(idx) = name.find(|c: char| c.is_whitespace() || c == '\0') {
            bail!("invalid character found in name at position `{idx}`: {name}")
        }
        Ok(Self { sym: ty, module, name, addr: None })
    }

    pub fn method(ty: &ASTType, name: &'a str) -> anyhow::Result<Self> {
        let module = match ty {
            ASTType::Any => "Any",
            ASTType::Line => "Line",
            ASTType::String => "String",
            ASTType::Syllabe => "Syllabe",
            ty if ty.is_variant() => "Variant",
            ty if ty.is_table() => "Table",
            ty if ty.is_sequence() => "Sequence",
            ty => bail!("can't have methods for type `{ty}`"),
        };
        Ok(VVRTSymbol { sym: VVRTSymbolType::Method, module, name, addr: None })
    }

    /// Builder-like syntax to take a symbol and associate it to an execution address gotten from
    /// the JIT engine. We ignore any previous address.
    pub fn with_addr(mut self, addr: LLVMOrcExecutorAddress) -> Self {
        self.addr = Some(addr);
        self
    }

    /// Mangle the symbol to get the name of it for the JIT engine or to find the symbol in the
    /// LLVM module. This is a heavy operation.
    ///
    /// If the symbol was not valid we panic. If the symbol was correctly constructed/parsed this
    /// should not happen.
    pub fn mangle(&self) -> CString {
        CString::new(self.to_string()).expect("invalid symbol")
    }
}

impl<'a> std::fmt::Display for VVRTSymbol<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let Self { sym, module, name, .. } = self;
        match sym {
            // Some shenanigans for builtin symbols...
            VVRTSymbolType::Method => {
                let mut module = unicode_segmentation::UnicodeSegmentation::graphemes(*module, true);
                let first = module.next().map(|str| str.to_uppercase()).unwrap_or_default();
                let tail = module.as_str();
                write!(f, "VV{first}{tail}_{}", name.to_lowercase())
            }

            // A simple rule for symbols generated from users' code.
            _ => write!(f, "_VV{sym}_{module}_{name}"),
        }
    }
}

impl<'a> TryFrom<&'a str> for VVRTSymbol<'a> {
    type Error = anyhow::Error;
    fn try_from(value: &'a str) -> Result<Self, Self::Error> {
        const CORRECT_SYMBOLS: &str = "`_VV{{sym}}_{{module}}_{{name}}` or `VV{{builtin_type}}_{{name}}`";

        let trimmed_value = value.trim();
        if let Some(pos) = trimmed_value.find('\0') {
            bail!("invalid symbol name `{trimmed_value}`, found null byte at {pos}")
        }

        let Some((sym, value)) = (match trimmed_value.strip_prefix("_VV") {
            Some(trimmed_value) => trimmed_value.split_once('_'),
            None => match trimmed_value.strip_prefix("VV") {
                Some(trimmed_value) => Some((VVRTSymbolType::Method.as_str(), trimmed_value)),
                None => bail!("invalid symbol name `{trimmed_value}`, didn't start with `_VV` or `VV`"),
            },
        }) else {
            bail!("invalid symbol name `{trimmed_value}`, should be {CORRECT_SYMBOLS}")
        };

        let Some((module, name)) = value.split_once('_') else {
            bail!("invalid symbol name `{trimmed_value}`, should be {CORRECT_SYMBOLS}")
        };

        Self::new(sym.parse()?, module, name)
    }
}

#[test]
fn test_vvrt_symbol() {
    use vvs_utils::*;

    assert_eq!(
        assert_ok!(VVRTSymbol::try_from("_VVConstant_retime_start")),
        VVRTSymbol { sym: VVRTSymbolType::Constant, module: "retime", name: "start", addr: None }
    );

    assert_eq!(
        "_VVFunction_tag_syl_modulo".to_string(),
        VVRTSymbol { sym: VVRTSymbolType::Function, module: "tag", name: "syl_modulo", addr: None }.to_string()
    );
}
