use std::{
    cmp::Ordering,
    hash::{Hash, Hasher},
    marker::PhantomData,
    mem::{align_of, size_of},
    ptr::NonNull,
};

/// A tagged pointer: a space-efficient representation of a pointer that smuggles an integer tag
/// into it.
///
/// `BITS` specifies how many bits are used for the tag. The alignment of `T` must be large enough
/// to store this many bits.
///
/// # Safety
/// The passed pointers, at construction time and each time you modify this struct, should all be
/// properly aligned, or your code may panic. The passed pointer must also be dereferenceable.
#[repr(transparent)]
pub struct TaggedPtr<T, const BITS: u8>(NonNull<u8>, PhantomData<NonNull<T>>);

impl<T, const BITS: u8> TaggedPtr<T, BITS> {
    const ASSERT: () = {
        assert!(size_of::<usize>() == size_of::<u64>(), "we only work with 64 bits architectures");
        assert!(align_of::<T>().is_power_of_two(), "expected alignment of `T` to be a power of 2");
        assert!(
            match size_of::<T>() {
                0 => true,
                n => n >= align_of::<T>(),
            },
            "expected size of non-ZST `T` to be at least alignment"
        );
        assert!(size_of::<T>() % align_of::<T>() == 0, "expected size of `T` to be multiple of alignment");
        assert!(1_usize.checked_shl(BITS as u32).is_some(), "`1 << BITS` doesn't fit in a `usize`");
        assert!(align_of::<T>().trailing_zeros() >= BITS as u32, "alignment of `T` must be at least `1 << BITS`");
    };

    const ALIGNMENT: usize = 1_usize.wrapping_shl(BITS as u32);
    const MASK: usize = Self::ALIGNMENT - 1;

    /// Creates a new tagged pointer. Only the lower `BITS` bits of `tag` are stored.
    ///
    /// If build in debug, a check is performed at compile time to ensure that the alignment of `T`
    /// is sufficient to smuggle the tag.
    #[allow(path_statements, clippy::no_effect)]
    #[inline]
    pub fn new(ptr: NonNull<T>, tag: usize) -> Self {
        Self::ASSERT; // Static assert!
        let ptr = ptr.as_ptr().cast::<u8>();

        let offset = ptr.align_offset(Self::ALIGNMENT);
        debug_assert!(offset != usize::MAX, "align offset failed");
        debug_assert!(offset & Self::MASK == 0, "`ptr` is not aligned enough");

        let ptr = ptr.wrapping_add(tag & Self::MASK);
        Self(unsafe { NonNull::new_unchecked(ptr) }, PhantomData)
    }

    /// Gets the pointer and tag stored by the tagged pointer.
    #[inline]
    pub fn get(self) -> (NonNull<T>, usize) {
        let ptr = self.0.as_ptr();
        let tag = Self::ALIGNMENT.wrapping_sub(ptr.align_offset(Self::ALIGNMENT)) & Self::MASK;
        let ptr = ptr.wrapping_sub(tag).cast::<T>();

        // SAFETY: ptr + 1 - 1 == ptr
        (unsafe { NonNull::new_unchecked(ptr) }, tag)
    }

    /// Get the address of the pointed thing as an integer. Here for moar reasons we only work with
    /// 64 bit architectures, thus the `usize == u64`.
    #[inline]
    pub fn addr(self) -> u64 {
        self.ptr().as_ptr() as u64
    }

    /// Gets the pointer stored by the tagged pointer, without the tag.
    #[inline]
    pub fn ptr(self) -> NonNull<T> {
        self.get().0
    }

    /// Gets the tag stored by the tagged pointer.
    #[inline]
    pub fn tag(self) -> usize {
        self.get().1
    }

    /// Sets the pointer without modifying the tag.
    #[inline]
    pub fn set_ptr(&mut self, ptr: NonNull<T>) {
        *self = Self::new(ptr, self.tag());
    }

    /// Sets the tag without modifying the pointer.
    #[inline]
    pub fn set_tag(&mut self, tag: usize) {
        *self = Self::new(self.ptr(), tag);
    }
}

impl<T, const BITS: u8> Copy for TaggedPtr<T, BITS> {}
impl<T, const BITS: u8> Clone for TaggedPtr<T, BITS> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<T, const BITS: u8> Eq for TaggedPtr<T, BITS> {}
impl<T, const BITS: u8> PartialEq for TaggedPtr<T, BITS> {
    fn eq(&self, other: &Self) -> bool {
        self.0 == other.0
    }
}

impl<T, const BITS: u8> PartialOrd for TaggedPtr<T, BITS> {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
impl<T, const BITS: u8> Ord for TaggedPtr<T, BITS> {
    fn cmp(&self, other: &Self) -> Ordering {
        self.0.cmp(&other.0)
    }
}

impl<T, const BITS: u8> Hash for TaggedPtr<T, BITS> {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.0.hash(state);
    }
}

impl<T, const BITS: u8> std::fmt::Debug for TaggedPtr<T, BITS> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (ptr, tag) = self.get();
        f.debug_struct("TaggedPtr")
            .field("ptr", &ptr)
            .field("tag", &tag)
            .finish()
    }
}
