use hashbrown::HashMap;

use crate::{
    types::{VVRTAny, VVRTType},
    TaggedPtr,
};
use core::{mem::size_of, ptr::NonNull};
use std::mem::align_of;

#[test]
fn test_ptr_size() {
    assert_eq!(size_of::<usize>(), size_of::<u64>());
    assert_eq!(size_of::<*const u64>(), size_of::<u64>());
}

#[test]
fn test_ptr_align() {
    assert_eq!(size_of::<NonNull<u64>>(), 8);
    assert_eq!(align_of::<NonNull<u64>>(), 8);
    assert_eq!(align_of::<NonNull<HashMap<Box<str>, VVRTAny>>>(), 8);
    assert_eq!(align_of::<NonNull<Vec<VVRTAny>>>(), 8);
}

#[test]
fn test_type_sizes() {
    use std::mem::size_of;
    assert_eq!(size_of::<VVRTAny>(), 8);
    assert_eq!(size_of::<VVRTType>(), 1);
}

#[repr(align(2))]
#[derive(Debug, Eq, PartialEq)]
struct Align2(pub u16);

#[repr(align(4))]
#[derive(Debug, Eq, PartialEq)]
struct Align4(pub u32);

#[repr(align(8))]
#[derive(Debug, Eq, PartialEq)]
struct Align8(pub u64);

#[test]
#[allow(clippy::let_unit_value)]
fn check_ptr_basic() {
    for i in 0..64 {
        let x = i as u8 * 3;
        let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), 0));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

        let x = Align2(i as u16 * 5);
        let tp = TaggedPtr::<_, 1>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), i & 0b1));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

        let mut x = Align4(i as u32 * 7);
        let ptr = NonNull::from(&mut x);
        let tp = TaggedPtr::<_, 2>::new(ptr, i);
        assert_eq!(tp.get(), (ptr, i & 0b11));
        unsafe { tp.ptr().as_mut() }.0 += 1;
        assert_eq!(x.0, i as u32 * 7 + 1);

        let x = Align8(i as u64 * 11);
        let tp = TaggedPtr::<_, 3>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), i & 0b111));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);
    }
}

#[test]
#[allow(clippy::let_unit_value)]
fn check_ptr_overaligned() {
    for i in 0..64 {
        let x = Align2(i as u16 * 3);
        let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), 0));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

        let x = Align4(i as u32 * 5);
        let tp = TaggedPtr::<_, 1>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), i & 0b1));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);

        let x = Align8(i as u64 * 7);
        let tp = TaggedPtr::<_, 2>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), i & 0b11));
        assert_eq!(*unsafe { tp.ptr().as_ref() }, x);
    }
}

#[test]
#[allow(clippy::let_unit_value)]
fn check_ptr_zst() {
    for i in 0..8 {
        let x = ();
        let tp = TaggedPtr::<_, 0>::new((&x).into(), i);
        assert_eq!(tp.get(), ((&x).into(), 0));
    }
}

#[test]
fn check_ptr_size() {
    assert_eq!(size_of::<TaggedPtr<u64, 3>>(), size_of::<core::ptr::NonNull<u64>>());
    assert_eq!(size_of::<TaggedPtr<u64, 3>>(), size_of::<Option<TaggedPtr<u64, 3>>>());
}

#[test]
fn check_ptr_not_entirely_dereferenceable() {
    #[repr(align(8))]
    #[allow(dead_code)]
    struct Type(u64, u64, u64);

    let a = Type(1, 2, 3);
    let mut tp = TaggedPtr::<_, 2>::new((&a).into(), 0b10);
    assert_eq!(tp.get(), ((&a).into(), 0b10));

    let mut b = Align4(0);
    let ptr = NonNull::from(&mut b).cast();
    tp.set_ptr(ptr);
    assert_eq!(tp.ptr(), ptr);
    tp.set_tag(0b11);
    assert_eq!(tp.tag(), 0b11);
    unsafe { tp.ptr().cast::<Align4>().as_mut() }.0 = 1234;
    assert_eq!(b.0, 1234);
}

#[test]
#[should_panic]
fn runtime_ptr_not_aligned_enough() {
    let ptr: *mut Align2 = &mut Align2(0);
    let ptr = unsafe { (ptr as *mut u8).add(1) };
    let ptr = ptr as *mut Align2;
    let ptr = unsafe { NonNull::new_unchecked(ptr) };
    TaggedPtr::<_, 1>::new(ptr, 0);
}
