use crate::types::*;
use std::{ptr::NonNull, rc::Rc};
use vvs_ass::ASSLine;

/// Represent a line, in an ffi-safe way. We represent if the line was create by using an Rc or was
/// built from a reference.
#[derive(Debug, Clone, Copy)]
#[repr(C)]
pub struct VVRTLine(NonNull<ASSLine<VVRTTable>>);
crate::types::any::impl_try_from_any! { ASSLine => VVRTLine }

impl VVRTLine {
    #[inline]
    pub fn dangling() -> Self {
        Self(NonNull::dangling())
    }

    #[inline]
    pub fn as_ptr(self) -> *mut u64 {
        self.0.as_ptr() as *mut _
    }

    #[inline]
    pub fn is_rc(&self) -> bool {
        !self.as_ref().aux.rt_infos().is_ass_borrowed
    }
}

impl From<ASSLine<VVRTTable>> for VVRTLine {
    fn from(mut value: ASSLine<VVRTTable>) -> Self {
        value.aux.rt_infos_mut().is_ass_borrowed = false;
        let ptr = Rc::into_raw(Rc::new(value)) as *mut _;
        VVRTLine(unsafe { NonNull::new_unchecked(ptr) })
    }
}

impl AsRef<ASSLine<VVRTTable>> for VVRTLine {
    fn as_ref(&self) -> &ASSLine<VVRTTable> {
        unsafe { self.0.as_ref() }
    }
}

impl AsMut<ASSLine<VVRTTable>> for VVRTLine {
    fn as_mut(&mut self) -> &mut ASSLine<VVRTTable> {
        unsafe { self.0.as_mut() }
    }
}

impl Eq for VVRTLine {}
impl PartialEq for VVRTLine {
    fn eq(&self, other: &Self) -> bool {
        self.as_ref().eq(other.as_ref())
    }
}

/// Module used to defined functions to be exported to the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTLine)]
mod ffi {
    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(this: VVRTLine) -> VVRTLine {
        this.as_ref().clone().into()
    }

    #[no_mangle]
    unsafe extern "C" fn clone(this @ VVRTLine(ptr): VVRTLine) -> VVRTLine {
        match this.is_rc() {
            false => VVRTLine_deep_clone(this),
            true => {
                Rc::increment_strong_count(ptr.as_ptr());
                this
            }
        }
    }

    #[no_mangle]
    unsafe extern "C" fn drop(this @ VVRTLine(ptr): VVRTLine) {
        if this.is_rc() {
            Rc::decrement_strong_count(ptr.as_ptr())
        }
    }

    #[no_mangle]
    unsafe extern "C" fn len(this: VVRTLine) -> i32 {
        this.as_ref().content.len().clamp(0, i32::MAX as usize) as i32
    }

    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn eq(this: VVRTLine, other: VVRTLine) -> bool { PartialEq::eq(&this, &other) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ne(this: VVRTLine, other: VVRTLine) -> bool { PartialEq::ne(&this, &other) }
}
