//! Contains utility to manipulate numbers from the runtime.

use vvs_utils::either;

/// Wrapper type for numbers that is intended to be used only for the [crate::types::VVRTAny]
/// and [crate::types::VVRTVariant] structs.
#[derive(Clone, Copy)]
#[repr(transparent)]
pub struct VVRTNumber(u64);

impl VVRTNumber {
    const FLT_TAG: u8 = 1;
    const INT_TAG: u8 = !Self::FLT_TAG;

    /// Cast the number as a pointer. We ensure by construction that this pointer is never null.
    /// You may never deref this pointer!
    #[inline]
    pub fn as_ptr(&self) -> *mut u64 {
        self.0 as *mut _
    }

    /// In debug mode, assert that the passed bytes are a valid number representation. In case of
    /// success re-return those bytes.
    ///
    /// In release build this function is a NoOp.
    #[inline]
    fn assert_is_number(bytes: [u8; 8]) -> [u8; 8] {
        let [.., flag, z1, z2, z3] = bytes;
        debug_assert!(z1 == z2 && z2 == z3 && z3 == 0, "invalid number");
        debug_assert!(flag == Self::INT_TAG || flag == Self::FLT_TAG, "invalid tag");
        bytes
    }

    /// Get the byte representation of the number we store.
    fn as_bytes(&self) -> (u8, [u8; 4]) {
        let [b1, b2, b3, b4, flag, _, _, _] = Self::assert_is_number(self.0.to_be_bytes());
        (flag, [b1, b2, b3, b4])
    }

    /// Create [VVRTAny] from bytes assumed to be a valid number representation.
    #[inline]
    pub fn from_number_be_bytes(bytes: [u8; 8]) -> Self {
        Self(u64::from_be_bytes(Self::assert_is_number(bytes)))
    }

    /// Tells whever the contained type is a floating number or not.
    #[inline]
    pub fn is_f32(&self) -> bool {
        self.as_bytes().0 == Self::FLT_TAG
    }

    /// Tells whever the contained type is an integer number or not.
    #[inline]
    pub fn is_i32(&self) -> bool {
        !self.is_f32()
    }
}

impl Default for VVRTNumber {
    fn default() -> Self {
        0_i32.into()
    }
}

impl From<VVRTNumber> for i32 {
    fn from(value: VVRTNumber) -> Self {
        let (flag, bytes) = value.as_bytes();
        either!(flag == VVRTNumber::INT_TAG
            => i32::from_be_bytes(bytes)
            ;  f32::from_be_bytes(bytes) as i32
        )
    }
}

impl From<VVRTNumber> for f32 {
    fn from(value: VVRTNumber) -> Self {
        let (flag, bytes) = value.as_bytes();
        either!(flag == VVRTNumber::INT_TAG
            => i32::from_be_bytes(bytes) as f32
            ;  f32::from_be_bytes(bytes)
        )
    }
}

impl From<i32> for VVRTNumber {
    fn from(value: i32) -> Self {
        let [b1, b2, b3, b4] = value.to_be_bytes();
        VVRTNumber::from_number_be_bytes([b1, b2, b3, b4, VVRTNumber::INT_TAG, 0, 0, 0])
    }
}

impl From<f32> for VVRTNumber {
    fn from(value: f32) -> Self {
        let [b1, b2, b3, b4] = value.to_be_bytes();
        VVRTNumber::from_number_be_bytes([b1, b2, b3, b4, VVRTNumber::FLT_TAG, 0, 0, 0])
    }
}

impl std::fmt::Debug for VVRTNumber {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (flag, bytes) = self.as_bytes();
        f.debug_struct("VVRTNumber")
            .field("flag", &flag)
            .field("bytes", &bytes)
            .finish()
    }
}

impl std::fmt::Display for VVRTNumber {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.is_f32() {
            true => write!(f, "{}", Into::<f32>::into(*self)),
            false => write!(f, "{}", Into::<i32>::into(*self)),
        }
    }
}
