use crate::types::*;
use std::{ptr::NonNull, rc::Rc};

#[derive(Debug, Clone, Copy)]
#[repr(transparent)]
pub struct VVRTSeq(NonNull<Vec<VVRTAny>>);
crate::types::any::impl_try_from_any! { Seq => VVRTSeq }

impl VVRTSeq {
    #[inline]
    pub fn dangling() -> Self {
        Self(NonNull::dangling())
    }

    #[inline]
    pub fn as_ptr(self) -> *mut u64 {
        self.0.as_ptr() as *mut _
    }

    #[inline]
    fn as_slice(&self) -> &[VVRTAny] {
        unsafe { self.0.as_ref() }.as_slice()
    }

    #[inline]
    fn from_vec(content: Vec<VVRTAny>) -> Self {
        unsafe { Self(NonNull::new_unchecked(Rc::into_raw(Rc::new(content)) as *mut _)) }
    }
}

impl Default for VVRTSeq {
    #[inline]
    fn default() -> Self {
        Self::from_vec(Default::default())
    }
}

impl Eq for VVRTSeq {}
impl PartialEq for VVRTSeq {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        self.as_slice() == other.as_slice()
    }
}

/// Module used to defined functions to be exported to the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTSeq)]
mod ffi {
    #[no_mangle]
    pub unsafe extern "C" fn new() -> VVRTSeq {
        Default::default()
    }

    /// Here we decrement the count of each child because we owned them. We don't support cyclic
    /// references...
    #[no_mangle]
    pub unsafe extern "C" fn drop(this @ VVRTSeq(ptr): VVRTSeq) {
        this.as_slice().iter().for_each(|child| VVRTAny_drop(*child));
        Rc::decrement_strong_count(ptr.as_ptr());
    }

    /// Here we increment the count of each child because we owned them. We don't support cyclic
    /// references...
    #[no_mangle]
    pub unsafe extern "C" fn clone(this @ VVRTSeq(ptr): VVRTSeq) -> VVRTSeq {
        this.as_slice().iter().for_each(|child| {
            let _ = VVRTAny_clone(*child);
        });
        Rc::increment_strong_count(ptr.as_ptr());
        VVRTSeq(ptr)
    }

    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(VVRTSeq(ptr): VVRTSeq) -> VVRTSeq {
        let mut content = Vec::with_capacity(ptr.as_ref().len());
        ptr.as_ref().iter().for_each(|e| content.push(VVRTAny_deep_clone(*e)));
        VVRTSeq::from_vec(content)
    }

    /// Get an item from the sequence, clone it.
    #[no_mangle]
    pub unsafe extern "C" fn get(this: VVRTSeq, idx: i32) -> VVRTAny {
        this.as_slice()
            .get(idx.clamp(0, i32::MAX) as usize)
            .map(|item| VVRTAny_clone(*item))
            .unwrap_or_default()
    }

    /// Push an item into the sequence, we take ownership of the item.
    #[no_mangle]
    pub unsafe extern "C" fn push(VVRTSeq(mut this): VVRTSeq, item: VVRTAny) {
        debug_assert!(
            !matches!(item.ty(), VVRTType::Table | VVRTType::Seq),
            "can't store container in containers"
        );
        this.as_mut().push(item);
    }

    #[no_mangle]
    pub unsafe extern "C" fn len(seq: VVRTSeq) -> i32 {
        seq.as_slice().len().try_into().expect("sequence to big")
    }

    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn eq(this: VVRTSeq, other: VVRTSeq) -> bool { PartialEq::eq(&this, &other) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn ne(this: VVRTSeq, other: VVRTSeq) -> bool { PartialEq::ne(&this, &other) }
}
