#![allow(non_snake_case)]

use hashbrown::{HashMap, HashSet};
use std::{ffi::CString, sync::OnceLock};
use vvs_lang::ast::ASTType;
use vvs_llvm::sys::*;

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct ManageFunctions {
    pub clone: &'static str,
    pub drop: &'static str,
}

/// Store properties about a type and the associated functions.
#[derive(Debug, Default)]
pub struct Properties {
    /// The drop/clone functions if present for this type. If [None], then we don't need to manage
    /// the memory for this type. The drop function takes the thing, and the clone a const pointer
    /// to the thing.
    pub manage: Option<ManageFunctions>,

    /// The function to call to check the equality. Must take only two arguments, const pointers to
    /// the same type, and return a boolean. This function doesn't consume the arguments.
    pub equality: Option<&'static str>,

    /// The function to call to get the length of the item. Must take one argument, a const pointer
    /// to the thing. This function doesn't consume the arguments. This function returns an
    /// integer.
    pub len: Option<&'static str>,

    /// Stores all the methods with some informations about them.
    pub methods: HashMap<&'static str, (unsafe extern "C" fn(), LLVMTypeRef, ASTType)>,
}

/// Trait used to give to the codegen the layout and functions associated with a type that is
/// present in the runtime and that must be handled in the codegen.
pub trait LLVMExported {
    /// Get the layout of the type.
    ///
    /// # Safety
    /// The [LLVMContextRef] must be valid.
    unsafe fn llvm_type(_: LLVMContextRef) -> LLVMTypeRef;

    /// Get all the functions associated with a type and some properties of this type. For example,
    /// integers and floating numbers can be copied without having to clone them, but strings, any,
    /// table, sequences, etc, must be cloned and dropped when no longer used.
    ///
    /// # Safety
    /// The [LLVMContextRef] must be valid.
    unsafe fn llvm_properties(_: LLVMContextRef) -> Properties;
}

/// Tells if a type is managed or not.
///
/// # Safety
/// - All the managed types must have been exported into at least one module
/// - This function is not thread-safe
pub unsafe fn LLVMTypeIsManaged(ty: LLVMTypeRef) -> bool {
    managed_types().contains(&ty)
}

unsafe fn managed_types() -> &'static mut HashSet<LLVMTypeRef> {
    static mut MANAGED_TYPES: OnceLock<HashSet<LLVMTypeRef>> = OnceLock::new();
    MANAGED_TYPES.get_or_init(HashSet::new);
    MANAGED_TYPES.get_mut().expect(" should have been created")
}

/// Register all the exported types and functions into a module.
///
/// # Safety
/// - The [LLVMContextRef] must be valid
/// - the [LLVMModuleRef] must be valid
/// - This function is not thread-safe
pub unsafe fn LLVMRegisterExportedIntoModule(
    c: LLVMContextRef,
    m: LLVMModuleRef,
) -> Vec<(&'static str, ASTType, LLVMValueRef, LLVMTypeRef)> {
    let mut ret = Vec::default();

    macro_rules! handle {
        ($ty: ty, $($tys: ty),+ $(,)?) => {
            handle!($ty);
            handle!($($tys),+);
        };
        ($ty: ty) => {
            let Properties { manage, equality: _, len: _, methods } = <$ty>::llvm_properties(c);
            if manage.is_some() {
                managed_types().insert(<$ty>::llvm_type(c));
            }
            for (name, (_, ty, asttype)) in methods {
                let cstr = CString::new(name).expect("invalid function name");
                let function = unsafe { LLVMAddFunction(m, cstr.as_ptr(), ty) };
                ret.push((name, asttype, function, ty));
            }
        };
    }

    use crate::types::*;
    handle!(VVRTType, VVRTAny, VVRTSeq, VVRTString, VVRTTable);
    ret
}
