//! Utilities to extend the [XDGFolder] thing with config files: take into account the
//! deserialization, merge of configs and others.

use crate::xdg::{MaybeFolderList, XDGError, XDGFindBehaviour, XDGFolder};
use serde::{Deserialize, Serialize};
use std::rc::Rc;

/// Search configurations in all config folders and merge them.
#[derive(Debug)]
pub struct XDGConfigMerged;

/// Only get the config file with higher priority, which means searching in order:
/// 1. $XDG_CONFIG_HOME <- which has a default value.
/// 2. $XDG_CONFIG_DIRS <- which has a default value.
#[derive(Debug)]
pub struct XDGConfigFirst;

/// Search configurations in all config folders and merge them. If the parsing for a configuration
/// file failed then log it and continue execution like normal, i.e. fail silently.
#[derive(Debug)]
pub struct XDGConfigMergedSilent;

pub type DeserializeFunctionPtr<'a, Format> =
    Rc<dyn Fn(String) -> Result<Format, Box<dyn std::error::Error>> + 'static>;

/// We will write one impl block for merged searches.
mod private_merged {
    use crate::xdg::*;
    use serde::Deserialize;
    use std::collections::VecDeque;

    /// Search with merging.
    pub trait Sealed: Sized {
        fn try_read<'a, Format>(config: &XDGConfig<'a, Format, Self>) -> Result<Format, XDGError>
        where
            Format: for<'de> Deserialize<'de> + Extend<Format>;
    }

    impl Sealed for XDGConfigMergedSilent {
        fn try_read<'a, Format>(config: &XDGConfig<'a, Format, Self>) -> Result<Format, XDGError>
        where
            Format: for<'de> Deserialize<'de> + Extend<Format>,
        {
            use XDGError::*;
            let result = config.search_files()?.into_iter().filter_map(|file| {
            let file = std::fs::read_to_string(file)
                .map_err(|err| log::error!(target: "xdg", "{}", ConfigIO(config.app.to_string(), config.get_file().to_string(), err)))
                .ok()?;
            config.deserialize.as_ref()(file)
                .map_err(|err| log::error!(target: "xdg", "{}", DeserializeError(config.app.to_string(), config.get_file().to_string(), err)))
                .ok()
        });

            let mut result: VecDeque<Format> = result.collect();
            let mut first = result
                .pop_front()
                .ok_or(ConfigNotFound(config.app.to_string(), config.get_file().to_string()))?;
            first.extend(result);
            Ok(first)
        }
    }

    impl Sealed for XDGConfigMerged {
        fn try_read<'a, Format>(config: &XDGConfig<'a, Format, Self>) -> Result<Format, XDGError>
        where
            Format: for<'de> Deserialize<'de> + Extend<Format>,
        {
            use XDGError::*;
            let mut result: VecDeque<Format> = Default::default();
            for file in config.search_files()?.into_iter() {
                let file = std::fs::read_to_string(file)
                    .map_err(|err| ConfigIO(config.app.to_string(), config.get_file().to_string(), err))?;
                let file = config.deserialize.as_ref()(file)
                    .map_err(|err| DeserializeError(config.app.to_string(), config.get_file().to_string(), err))?;
                result.push_back(file);
            }
            let mut first = result
                .pop_front()
                .ok_or(ConfigNotFound(config.app.to_string(), config.get_file().to_string()))?;
            first.extend(result);
            Ok(first)
        }
    }
}

/// A struct to contain informations about the app we are querying the config for. The behaviour of
/// the search can be changed.
/// ---
/// On the `Format` generic parameter:
/// - If the `Format` parameter implements [Serialize], then you can write or provide a default
///   that will be written to disk (or fail silently...)
/// - If the `Format` parameter implements [Serialize] and [Default], you can depend on the
///   default value if an error occured in the deserialization.
/// ---
/// On the `Merged` generic parameter:
/// - If the `Merged` parameter is [XDGConfigFirst], then only the file with the higher priority
///   is parsed.
/// - If the `Merged` parameter is [XDGConfigMerged] and `Format` implements [Extend] on itself,
///   then all the found files are parsed and then merged, an error is returned on the first
///   failure of the deserialization process.
pub struct XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de>,
{
    /// The application name.
    app: &'a str,

    /// The file name. If not present will be defaulted to `config` at resolution time.
    file: Option<&'a str>,

    /// Deserializer function.
    deserialize: DeserializeFunctionPtr<'a, Format>,

    /// Stores the format for deserialization.
    _type: std::marker::PhantomData<Format>,

    /// Stores the format for deserialization.
    _merged: std::marker::PhantomData<Merged>,
}

impl<'a, Format, Merged> std::fmt::Debug for XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de>,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("XDGConfig")
            .field("app", &self.app)
            .field("file", &self.get_file())
            .field("merged", &self._merged)
            .finish()
    }
}

impl<'a, Format, Merged> XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de>,
{
    /// By default we say that the default config file is ${FOLDER}/${APP}/config like many
    /// applications do.
    const DEFAULT_FILE: &'static str = "config";

    /// Create a new [XDGConfig] helper.
    #[inline]
    pub fn new(
        app: &'a str,
        deserialize: impl Fn(String) -> Result<Format, Box<dyn std::error::Error>> + 'static,
    ) -> Self {
        Self {
            app,
            deserialize: Rc::new(deserialize),
            file: Default::default(),
            _type: Default::default(),
            _merged: Default::default(),
        }
    }

    /// Change the file to resolve.
    #[inline]
    pub fn file(&mut self, file: &'a str) -> &mut Self {
        self.file = Some(file);
        self
    }

    /// Get the name of the config file that we will try to read. Returns the default if not set by
    /// the user.
    #[inline]
    pub fn get_file(&self) -> &str {
        match self.file.as_ref() {
            Some(file) => file,
            None => Self::DEFAULT_FILE,
        }
    }

    /// Search all config files in all locations...
    #[inline]
    fn search_files(&self) -> Result<MaybeFolderList, XDGError> {
        XDGFolder::ConfigDirs.find(self.app, self.get_file(), XDGFindBehaviour::ExistingOnly)
    }

    /// Prepare config folders.
    #[inline]
    pub fn prepare_folder(&self) -> impl IntoIterator<Item = <MaybeFolderList as IntoIterator>::Item> {
        XDGFolder::ConfigDirs.prepare_folder()
    }
}

impl<'a, Format> XDGConfig<'a, Format, XDGConfigFirst>
where
    Format: for<'de> Deserialize<'de>,
{
    /// Try to read the config file and deserialize it.
    pub fn try_read(&self) -> Result<Format, XDGError> {
        let Some(file) = self.search_files()?.into_first() else {
            return Err(XDGError::ConfigNotFound(self.app.to_string(), self.get_file().to_string()));
        };
        let file = std::fs::read_to_string(file)
            .map_err(|err| XDGError::ConfigIO(self.app.to_string(), self.get_file().to_string(), err))?;
        self.deserialize.as_ref()(file)
            .map_err(|err| XDGError::DeserializeError(self.app.to_string(), self.get_file().to_string(), err))
    }
}

impl<'a, Format> XDGConfig<'a, Format, XDGConfigFirst>
where
    Format: for<'de> Deserialize<'de> + Serialize,
{
    /// Try to read the config file and deserialize it. If an error is encountred at any point, log
    /// it and return the provided default. If needed the default value is written on the disk,
    /// note that this operation may fail silently.
    #[inline]
    pub fn read_or(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        default: Format,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.write_silent(serialize, default)
        })
    }

    /// Try to read the config file and deserialize it. If an error is encountred at any point, log
    /// it and return the provided default. If needed the default value is written on the disk,
    /// note that this operation may fail silently.
    #[inline]
    pub fn read_or_else(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        cb: impl FnOnce() -> Format,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.write_silent(serialize, cb())
        })
    }

    /// Write a value to the default config file location, the one with the higher priority
    /// (usually the user's one)
    fn write(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        value: &Format,
    ) -> Result<(), XDGError> {
        let content = serialize(value)
            .map_err(|err| XDGError::SerializeError(self.app.to_string(), self.get_file().to_string(), err))?;
        let Some(path) = XDGFolder::ConfigDirs
            .find(self.app, self.get_file(), XDGFindBehaviour::FirstOrCreate)?
            .into_first()
        else {
            unreachable!(
                "the user must have at least one location to place a config file, permission or fs quota problem?"
            )
        };

        match std::fs::create_dir_all(
            path.parent()
                .ok_or(XDGError::ConfigFileHasNoParentFolder(self.app.to_string(), self.get_file().to_string()))?,
        ) {
            Err(err) if !matches!(err.kind(), std::io::ErrorKind::AlreadyExists) => {
                return Err(XDGError::ConfigIO(self.app.to_string(), self.get_file().to_string(), err))
            }
            _ => {}
        }

        std::fs::write(path, content)
            .map_err(|err| XDGError::ConfigIO(self.app.to_string(), self.get_file().to_string(), err))?;

        Ok(())
    }

    /// Same as [XDGConfig::write] but log any error and fail silently, returning the value that
    /// was attempted to be written to disk.
    #[inline]
    fn write_silent(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        value: Format,
    ) -> Format {
        if let Err(err) = self.write(serialize, &value) {
            log::error!(target: "xdg", "failed to write default with err: {err}")
        }
        value
    }
}

impl<'a, Format> XDGConfig<'a, Format, XDGConfigFirst>
where
    Format: for<'de> Deserialize<'de> + Serialize + Default,
{
    /// Try to read the config file and deserialize it. If an error is encountred at any point, log
    /// it and return the default. If needed the default value is written on the disk, note that
    /// this operation may fail silently.
    #[inline]
    pub fn read_or_default(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.write_silent(serialize, Default::default())
        })
    }
}

// XDGConfigMerged + XDGConfigMergedSilent

impl<'a, Format, Merged> XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de> + Extend<Format>,
    Merged: private_merged::Sealed,
{
    /// When trying to read or write the default, we write the file with the same logic as the
    /// [XDGConfigFirst] variant, we add this function to reduce the code to write for the write
    /// logic...
    #[inline]
    fn to_config_first(&self) -> XDGConfig<Format, XDGConfigFirst> {
        XDGConfig::<Format, XDGConfigFirst> {
            app: self.app,
            file: self.file,
            deserialize: self.deserialize.clone(),
            _type: std::marker::PhantomData,
            _merged: std::marker::PhantomData,
        }
    }

    /// Try to read the config files and deserialize them. If one config file failed to be
    /// deserialized then returns an error if the merged was not silent, or just ignore the file if
    /// the merge was silent. If no config file where found this is an error. All the files are
    /// merged, this operation can't fail, implementations must fail silently or have sane defaults
    /// and prefer files with higher priority.
    #[inline]
    pub fn try_read(&self) -> Result<Format, XDGError> {
        Merged::try_read(self)
    }
}

impl<'a, Format, Merged> XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de> + Extend<Format> + Serialize,
    Merged: private_merged::Sealed,
{
    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        default: Format,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(serialize, default)
        })
    }

    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or_else(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
        cb: impl FnOnce() -> Format,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(serialize, cb())
        })
    }
}

impl<'a, Format, Merged> XDGConfig<'a, Format, Merged>
where
    Format: for<'de> Deserialize<'de> + Extend<Format> + Serialize + Default,
    Merged: private_merged::Sealed,
{
    /// Try to read the config files and deserialize them. If an error is encountred at any point,
    /// log it and return the provided default if the merge was not silent, skip the file if it was
    /// silent. If needed the default value is written on the disk, note that this operation may
    /// fail silently.
    #[inline]
    pub fn read_or_default(
        &self,
        serialize: impl FnOnce(&Format) -> Result<String, Box<dyn std::error::Error>>,
    ) -> Format {
        self.try_read().unwrap_or_else(|err| {
            log::error!(target: "xdg", "read error, return default value to user: {err}");
            self.to_config_first().write_silent(serialize, Default::default())
        })
    }
}
