//! Utility functions to follow the freedesktop specifications on config folders and such:
//! https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html

#[cfg(test)]
mod tests;

mod config;
mod folders;
mod options;
mod paths;

pub use config::*;
pub use folders::*;
pub use options::*;
pub use paths::*;

use std::{io::Error as IoError, path::PathBuf};
use thiserror::Error;

/// Get the user's home folder. Panics if the env variable was not found or it can't be
/// canonicalized.
#[inline]
pub fn home_folder() -> PathBuf {
    PathBuf::from(std::env::var("HOME").expect("failed to get the $HOME env variable"))
        .canonicalize()
        .expect("failed to canonicalize the $HOME folder path")
}

#[derive(Debug, Error)]
pub enum XDGError {
    #[error("failed to find file {2} for application {1} with folder family {0}")]
    NotFound(XDGFolder, String, String),

    #[error("failed to find config file {1} for application {0}")]
    ConfigNotFound(String, String),

    #[error("failed to find parent folder of config file {1} for application {0}")]
    ConfigFileHasNoParentFolder(String, String),

    #[error("io error for config file {1} for application {0}: {2}")]
    ConfigIO(String, String, IoError),

    #[error("deserialization failed on file {1} for application {0} with: {2}")]
    DeserializeError(String, String, Box<dyn std::error::Error>),

    #[error("serialization failed on file {1} for application {0} with: {2}")]
    SerializeError(String, String, Box<dyn std::error::Error>),
}
