#include "TimingScene.hh"

#include "TimingParams.hh"

using namespace Vivy;

//TimingScene::TimingScene(QWidget *parent) noexcept
//    : QGraphicsScene(parent)
//{
//}

TimingScene::TimingScene(QImage img_, AudioContext::StreamPtr stream, VivyDocumentView &rootView,
                         QWidget *parent) noexcept
    : QGraphicsScene(parent)
    , AbstractVivyDocumentNeeder(rootView)
    , img(img_)
    , audioStream(stream)
{
    params = new TimingParams();
    // We put a common origin for the items at the bottom of the axis
    params->setAxisHeight(30); // TODO: remove magic number

    rebuildScene();
}

void
TimingScene::updateScene(QImage img_, AudioContext::StreamPtr stream_)
{
    img         = img_;
    audioStream = stream_;
    rebuildScene();
}

void
TimingScene::rebuildScene()
{
    for (auto &item : items()) {
        removeItem(item);
        delete (item);
    }
    timingLines.clear();

    QPixmap pixmap(QPixmap::fromImage(img));
    backgroundImg = addPixmap(pixmap);
    backgroundImg->setZValue(Z_SPECTER);
    backgroundImg->setPos(0, params->axisHeight());

    params->setAudioHeight(pixmap.height());
    params->setAudioWidth(img.width());
    params->setAudioLength(int(audioStream->getLength()));

    ax = new TimingAxis();
    addItem(ax);
    ax->onParamsChanged();
    ax->setZValue(Z_AXIS);
    ax->setPos(0, params->axisHeight());

    cursor = new TimingCursor();
    addItem(cursor);
    cursor->onParamsChanged();
    cursor->setPos(0, params->axisHeight());

    // Freeze the scene boundaries
    QRectF sRect = sceneRect();
    setSceneRect(QRectF(0, sRect.top(), params->audioWidth(), sRect.height()).normalized());

    if (auto assDocument = rootVivyDocument.getAssSubDocument()) {
        QVector<Ass::LinePtr> lines = assDocument->getLines();
        for (int i = 0; i < lines.size(); ++i) {
            if (auto line = lines.at(i)) {
                TimingLine *l = new TimingLine(*line, i);
                addItem(l);
                l->onParamsChanged();
                l->setVisible(false);
                timingLines.append(l);
                if (auto assLinesModel = rootVivyDocumentView.getAssLinesModel()) {
                    connect(l, &TimingLine::lineChanged, assLinesModel, &AssLinesModel::updateLine);
                }
            }
        }
    }

    if (auto assLinesView = rootVivyDocumentView.getAssLinesView()) {
        connect(assLinesView, &AssLinesView::updateSelectedLines, this,
                &TimingScene::updateSelectedLines);
    }
}

void
TimingScene::mousePressEvent(QGraphicsSceneMouseEvent *event) noexcept
{
    QPointF pos        = event->scenePos();
    QGraphicsItem *got = itemAt(pos, QTransform());
    Ass::LinePtr p     = currentLine.lock();

    if (p && (got == nullptr || got == backgroundImg)) {
        // Handle the different cases
        switch (timingMode) {
        case TimingMode::Line: handleMousePressEventLine(event, p); break;
        case TimingMode::Syl: handleMousePressEventSyl(event, p); break;
        case TimingMode::Char: handleMousePressEventChar(event, p); break;
        }
    }

    QGraphicsScene::mousePressEvent(event);
}

void
TimingScene::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent) noexcept
{
    int x = int(mouseEvent->scenePos().rx());
    cursor->setPos(x, 30);
    cursor->setTime(params->printCursor(params->msFromPos(x)));
    QGraphicsScene::mouseMoveEvent(mouseEvent);
}

void
TimingScene::handleMousePressEventLine(QGraphicsSceneMouseEvent *event, Ass::LinePtr p) noexcept
{
    QPointF pos = event->scenePos();
    int time    = params->msFromPos(int(pos.x()));

    if (const auto &btn = event->button(); btn == Qt::LeftButton) {
        p->setStart(quint64(time));
    } else if (btn == Qt::RightButton) {
        p->setEnd(quint64(time));
    }
}

void
TimingScene::handleMousePressEventSyl(QGraphicsSceneMouseEvent *, Ass::LinePtr) noexcept
{
}

void
TimingScene::handleMousePressEventChar(QGraphicsSceneMouseEvent *, Ass::LinePtr) noexcept
{
}

QGraphicsPixmapItem *
TimingScene::bg() noexcept
{
    return backgroundImg;
}

TimingAxis *
TimingScene::getAxis()
{
    return ax;
}

TimingParams *
TimingScene::getParams()
{
    return params;
}

void
TimingScene::updateSelectedLines(const QItemSelection &selected, const QItemSelection &deselected)
{
    auto indexes = selected.indexes();
    for (int i = 0; i < indexes.size(); ++i) {
        timingLines[indexes[i].row()]->setVisible(true);
    }

    indexes = deselected.indexes();
    for (int i = 0; i < indexes.size(); ++i) {
        timingLines[indexes[i].row()]->setVisible(false);
    }
}
