#include "PreCompiledHeaders.hh"
#include "UI/LogConsole.hh"

using namespace Vivy;

ConsoleLogSinkDispatcher::ConsoleLogSinkDispatcher() noexcept
    : LogSinkDispatcher(std::string_view{ "console" })
{
}

void
ConsoleLogSinkDispatcher::handleLogMessage(const std::string_view strv,
                                           const LogMessage &msg) noexcept
{
    if (console)
        console->handleLogMessage(strv, msg);
}

void
ConsoleLogSinkDispatcher::attachLogConsole(LogConsole *widget) noexcept
{
    console = widget;
    if (widget)
        widget->setDispatcher(shared_from_this());
}

LogConsole::LogConsole(QWidget *parent) noexcept
    : QListWidget(parent)
{
    setSortingEnabled(false);
    setSelectionMode(QAbstractItemView::NoSelection);

    const QString style = QStringLiteral("* { font-family: \"FiraCode\"; font-size: 8pt; }");
    setStyleSheet(style);
}

LogConsole::~LogConsole() noexcept
{
    if (auto parentDispatcher = dispatcher.lock()) {
        qDebug() << "Reset parent LogConsole pointer";
        parentDispatcher->attachLogConsole(nullptr);
    } else {
        qCritical() << "Failed to reset LogConsole from parent";
    }
}

void
LogConsole::setDispatcher(std::shared_ptr<ConsoleLogSinkDispatcher> ptr) noexcept
{
    dispatcher = std::weak_ptr<ConsoleLogSinkDispatcher>(ptr);
}

void
LogConsole::setMessageCountLimit(int limit)
{
    if (limit <= 0)
        throw std::logic_error("Can't pass a negative count!");
    messageLimit = limit;
}

void
LogConsole::handleLogMessage(const std::string_view cat, const LogMessage &msg) noexcept
{
    while (count() >= messageLimit) {
        QListWidgetItem *itemWidget = item(0);
        if (itemWidget != nullptr)
            removeItemWidget(itemWidget);
    }

    QString label;

    // Magic number to reduce early allocations
    label.reserve(std::max(static_cast<int>(cat.size()) * 2, 10));

    label.append('[');
    for (const char c : cat)
        label.append(c);
    label.append("] ");

    for (const char c : LogLevel::toStdStringView(msg.getHeader().severity))
        label.append(c);
    label.append(" -> ");

    // Add the content of the message
    label.reserve(static_cast<int>(1 + label.size() +
                                   std::max(10, static_cast<int>(msg.getTextBuffer().size()))));
    for (const char c : msg.getTextBuffer())
        label.append(c);

    addItem(label);
    QListWidgetItem *lastItem = item(count() - 1);
    scrollToItem(lastItem);
}
